#ifndef OBJECTROB_HPP
#define OBJECTROB_HPP

#define OPEN_VALUE -100


#include <Eigen/Geometry>
#include "Planner-pkg.h"
#include "API/ConfigSpace/configuration.hpp"
#include "robot.hpp"
#include "GTP/GTPTools/geometricTools/geometricComputations.hpp"
#include "GTP/GTPTools/geometricTools/geometricForms-pkg.hpp"
#include "Collision-pkg.h"
#include "planner/planEnvironment.hpp"
#include "Logging/Logger.h"


//class Robot;
class GeometricForm;

extern std::vector<GeometricForm*> formToDraw;

class ObjectRob
{
    MOVE3D_STATIC_LOGGER;
public:
    ObjectRob();
    ObjectRob(Robot* rob);

    ~ObjectRob();

    //return the closest distance to a support from the object
    // for now implemented for rectangle only
    double distToPoint(p3d_point p);

    double distToPointZ(p3d_point p);

    void setManipulable (bool val) {_manipulable = val;}
    bool isManipulable() {return _manipulable;}

    void setSupport (bool val) {_support = val;}
    bool isSupport() {return _support;}

    void setContainer (bool val) {_container = val;}
    bool isContainer() {return _container;}

    void setIsObject (bool val) {_isObject = val;}
    bool isObject() {return _isObject;}

    void addStableRotation(configPt q) {_stableRotations.push_back(q);_manipulable=true;}
    std::vector<configPt> getStableRotations() {return _stableRotations;}
    configPt getStableRotation(unsigned int id, double rot);

    void addSupports(GeometricForm* gf) { _supportList.push_back(gf);_support = true;}
    std::vector<GeometricForm*> getSupports();
    void drawSupports(bool val);
    void cleanSupportList();

    /// true if the object has viman tags attributed
    bool hasTags() {return _tags.size();}
    /// add a viman tag in the form of a transform
    void addTag(Eigen::Affine3d tf) {_tags.push_back(tf);}
    void clearTags(){_tags.clear();}
    std::vector<Eigen::Affine3d> const & getTags(){return _tags;}

    int isInStableRotation();

    bool isVirtual() {return _isVirtual;}
    void getIsVirtual(bool iv) {_isVirtual = iv;}

    void setToVirtual(bool val);

    bool hasVerticalityConstraint() const;
    void setHasVerticalityConstraint(bool hasVerticalityConstraint);

    Eigen::Vector3d keepVerticalAxis() const;
    void setKeepVerticalAxis(const Eigen::Vector3d &keepVerticalAxis);

    double maxVerticalTolerance() const;
    void setMaxVerticalTolerance(double angle);

private:
    bool _manipulable;
    bool _support;
    bool _container;
    bool _isObject;
    bool _isVirtual;

    std::vector<configPt> _stableRotations;

    std::vector<GeometricForm*> _supportList;
    std::vector<GeometricForm*> _currentSupportList;

    bool _hasVerticalityConstraint;
    Eigen::Vector3d _keepVerticalAxis;
    double _maxVerticalTolerance;

    std::vector<Eigen::Affine3d> _tags; ///< list of viman tags. Applyng this tranform to a unity square in the x,y plane of the object frame gives the square representing the tag.



    Robot* _robot;

};

#endif // OBJECTROB_HPP
