/*
 * Romeo_Wrap.cpp : Wrapper between Move3D and HPP that provides an interface for handling Romeo
 *
 * Author : renaud Viry (renaud.viry@laas.fr)
 * Copyright LAAS-CNRS 2014
 *
 */

#include "Romeo_Wrap.hpp"

#ifdef HPP

RomeoWrap::RomeoWrap(Robot *move3dRobot)
{
	startConf = getStartConf ();

	_eef_joint = "";

	_leftAnkle = "LAnkleRoll";
	_rightAnkle = "RAnkleRoll";

	loadHumanoidRobot ("Romeo", "freeflyer", "romeo_description", "romeo", "", "", move3dRobot);
}

std::map < std::string, double > RomeoWrap::getStartConf()
{
	std::map < std::string, double > halfSittingConf;

	// halfSittingConf["base_joint_xyz"] = (0, 0, 0.840252) ;
	// halfSittingConf["base_joint_SO3"] = (1, 0, 0, 0)     ;
	halfSittingConf["LHipYaw"]        = 0          ;
	halfSittingConf["LHipRoll"]       = 0          ;
	halfSittingConf["LHipPitch"]      = -0.3490658 ;
	halfSittingConf["LKneePitch"]     = 0.6981317  ;
	halfSittingConf["LAnklePitch"]    = -0.3490658 ;
	halfSittingConf["LAnkleRoll"]     = 0          ;
	halfSittingConf["LToePitch"]      = 0          ;
	halfSittingConf["RHipYaw"]        = 0          ;
	halfSittingConf["RHipRoll"]       = 0          ;
	halfSittingConf["RHipPitch"]      = -0.3490658 ;
	halfSittingConf["RKneePitch"]     = 0.6981317  ;
	halfSittingConf["RAnklePitch"]    = -0.3490658 ;
	halfSittingConf["RAnkleRoll"]     = 0          ;
	halfSittingConf["RToePitch"]      = 0          ;
	halfSittingConf["TrunkYaw"]       = 0          ;
	halfSittingConf["LShoulderPitch"] = 1.5        ;
	halfSittingConf["LShoulderYaw"]   = 0.6        ;
	halfSittingConf["LElbowRoll"]     = -0.5       ;
	halfSittingConf["LElbowYaw"]      = -1.05      ;
	halfSittingConf["LWristRoll"]     = -0.4       ;
	halfSittingConf["LWristYaw"]      = -0.3       ;
	halfSittingConf["LWristPitch"]    = -0.2       ;
	halfSittingConf["LFinger11"]      = 0          ;
	halfSittingConf["LFinger12"]      = 0          ;
	halfSittingConf["LFinger13"]      = 0          ;
	halfSittingConf["LFinger21"]      = 0          ;
	halfSittingConf["LFinger22"]      = 0          ;
	halfSittingConf["LFinger23"]      = 0          ;
	halfSittingConf["LFinger31"]      = 0          ;
	halfSittingConf["LFinger32"]      = 0          ;
	halfSittingConf["LFinger33"]      = 0          ;
	halfSittingConf["LThumb1"]        = 0          ;
	halfSittingConf["LThumb2"]        = 0          ;
	halfSittingConf["LThumb3"]        = 0          ;
	halfSittingConf["NeckYaw"]        = 0          ;
	halfSittingConf["NeckPitch"]      = 0          ;
	halfSittingConf["HeadPitch"]      = 0          ;
	halfSittingConf["HeadRoll"]       = 0          ;
	halfSittingConf["LEyeYaw"]        = 0          ;
	halfSittingConf["LEyePitch"]      = 0          ;
	halfSittingConf["REyeYaw"]        = 0          ;
	halfSittingConf["REyePitch"]      = 0          ;
	halfSittingConf["RShoulderPitch"] = 1.5        ;
	halfSittingConf["RShoulderYaw"]   = -0.6       ;
	halfSittingConf["RElbowRoll"]     = 0.5        ;
	halfSittingConf["RElbowYaw"]      = 1.05       ;
	halfSittingConf["RWristRoll"]     = -0.4       ;
	halfSittingConf["RWristYaw"]      = -0.3       ;
	halfSittingConf["RWristPitch"]    = -0.2       ;
	halfSittingConf["RFinger11"]      = 0          ;
	halfSittingConf["RFinger12"]      = 0          ;
	halfSittingConf["RFinger13"]      = 0          ;
	halfSittingConf["RFinger21"]      = 0          ;
	halfSittingConf["RFinger22"]      = 0          ;
	halfSittingConf["RFinger23"]      = 0          ;
	halfSittingConf["RFinger31"]      = 0          ;
	halfSittingConf["RFinger32"]      = 0          ;
	halfSittingConf["RFinger33"]      = 0          ;
	halfSittingConf["RThumb1"]        = 0          ;
	halfSittingConf["RThumb2"]        = 0          ;
	halfSittingConf["RThumb3"]        = 0          ;

	return halfSittingConf;
}

void RomeoWrap::initializeConstraints()
{
	hpp::model::vector3_t tmp, footT;
	hpp::model::matrix3_t footRotation;
	p3d_vector4 quat;
	p3d_matrix3 footR;
	p3d_matrix4 Ma;
	std::string lA = "LeftAnkle";

	if (!_footFixed)
	{
		tmp[0] = 0;
		tmp[1] = 0;
		tmp[2] = 0;

		footT[0] = _manipulationSolver->robot()->getJointByName(_leftAnkle)->robot()->currentConfiguration()[0];
		footT[1] = _manipulationSolver->robot()->getJointByName(_leftAnkle)->robot()->currentConfiguration()[1];
		footT[2] = _manipulationSolver->robot()->getJointByName(_leftAnkle)->robot()->currentConfiguration()[2];
		quat[0] = _manipulationSolver->robot()->getJointByName(_leftAnkle)->robot()->currentConfiguration()[3];
		quat[1] = _manipulationSolver->robot()->getJointByName(_leftAnkle)->robot()->currentConfiguration()[4];
		quat[2] = _manipulationSolver->robot()->getJointByName(_leftAnkle)->robot()->currentConfiguration()[5];
		quat[3] = _manipulationSolver->robot()->getJointByName(_leftAnkle)->robot()->currentConfiguration()[6];
		p3d_quaternion_to_matrix3(quat, footR);
		p3d_mat4SetRotMatrix(footR, Ma);

		for (int i = 0; i < 3; i ++)
			for (int j = 0; j < 3; j ++)
				footRotation(i, j) = footR[i][j];

		// Lock Orientation and Position of feet
		std::vector<bool> true_vect (3, true);
		true_vect[2] = false;
		addPositionConstraint(lA + std::string ("/position"), std::string(_leftAnkle), tmp, footT, true_vect);
		std::vector<bool> false_vect (3, false);
		false_vect[2] = true;
		addOrientationConstraint(lA + std::string ("/orientation"), std::string (_leftAnkle), footRotation, false_vect);

		_footFixed = true;
	}

	// Lock some DoF to speed up IK
	addLockDof ("LFinger11", getJointDofValue ("LFinger11"));
	addLockDof ("LFinger12", getJointDofValue ("LFinger12"));
	addLockDof ("LFinger13", getJointDofValue ("LFinger13"));
	addLockDof ("LFinger21", getJointDofValue ("LFinger21"));
	addLockDof ("LFinger22", getJointDofValue ("LFinger22"));
	addLockDof ("LFinger23", getJointDofValue ("LFinger23"));
	addLockDof ("LFinger31", getJointDofValue ("LFinger31"));
	addLockDof ("LFinger32", getJointDofValue ("LFinger32"));
	addLockDof ("LFinger33", getJointDofValue ("LFinger33"));
	addLockDof ("LThumb1",   getJointDofValue ("LThumb1") );
	addLockDof ("LThumb2",   getJointDofValue ("LThumb2") );
	addLockDof ("LThumb3",   getJointDofValue ("LThumb3") );

	addLockDof ("RFinger11", getJointDofValue ("RFinger11"));
	addLockDof ("RFinger12", getJointDofValue ("RFinger12"));
	addLockDof ("RFinger13", getJointDofValue ("RFinger13"));
	addLockDof ("RFinger21", getJointDofValue ("RFinger21"));
	addLockDof ("RFinger22", getJointDofValue ("RFinger22"));
	addLockDof ("RFinger23", getJointDofValue ("RFinger23"));
	addLockDof ("RFinger31", getJointDofValue ("RFinger31"));
	addLockDof ("RFinger32", getJointDofValue ("RFinger32"));
	addLockDof ("RFinger33", getJointDofValue ("RFinger33"));
	addLockDof ("RThumb1",   getJointDofValue ("RThumb1")  );
	addLockDof ("RThumb2",   getJointDofValue ("RThumb2")  );
	addLockDof ("RThumb3",   getJointDofValue ("RThumb3")  );

	addLockDof ("LToePitch", getJointDofValue ("LToePitch"));
	addLockDof ("RToePitch", getJointDofValue ("RToePitch"));

	// addLockDof ("NeckYaw",   getJointDofValue ("NeckYaw")  );
	// addLockDof ("NeckPitch", getJointDofValue ("NeckPitch"));
	// addLockDof ("HeadPitch", getJointDofValue ("HeadPitch"));
	// addLockDof ("HeadRoll",  getJointDofValue ("HeadRoll") );

	addLockDof ("LEyeYaw",   getJointDofValue ("LEyeYaw")  );
	addLockDof ("LEyePitch", getJointDofValue ("LEyePitch"));
	addLockDof ("REyeYaw",   getJointDofValue ("REyeYaw")  );
	addLockDof ("REyePitch", getJointDofValue ("REyePitch"));
}

void RomeoWrap::randomizeConfig(hpp::core::ConfigurationPtr_t config, std::map<string, bool> randomizeParts)
{
	double vmin, vmax;
	for (std::map<string, bool>::iterator it = randomizeParts.begin(); it != randomizeParts.end(); ++it)
	{
		if ((it->first == "leftLeg") && (it->second == true))
		{
			for (int index = 0; index < 6; index ++)
			{
				vmin = _manipulationSolver->robot()->getJointVector()[index + 3]->lowerBound(0);
				vmax = _manipulationSolver->robot()->getJointVector()[index + 3]->upperBound(0);
				(*config)[index + 5] = p3d_random(vmin,vmax);
			}
		}
		if ((it->first == "rightLeg") && (it->second == true))
		{
			for (int index = 0; index < 6; index ++)
			{
				vmin = _manipulationSolver->robot()->getJointVector()[index + 11]->lowerBound(0);
				vmax = _manipulationSolver->robot()->getJointVector()[index + 11]->upperBound(0);
				(*config)[index + 13] = p3d_random(vmin,vmax);
			}
		}
		if ((it->first == "leftArm") && (it->second == true))
		{
			for (int index = 0; index < 7; index ++)
			{
				vmin = _manipulationSolver->robot()->getJointVector()[index + 20]->lowerBound(0);
				vmax = _manipulationSolver->robot()->getJointVector()[index + 20]->upperBound(0);
				(*config)[index + 22] = p3d_random(vmin,vmax);
			}
		}
		if ((it->first == "rightArm") && (it->second == true))
		{
			for (int index = 0; index < 7; index ++)
			{
				vmin = _manipulationSolver->robot()->getJointVector()[index + 51]->lowerBound(0);
				vmax = _manipulationSolver->robot()->getJointVector()[index + 51]->upperBound(0);
				(*config)[index + 53] = p3d_random(vmin,vmax);
			}
		}
		if ((it->first == "neck") && (it->second == true))
		{
			vmin = _manipulationSolver->robot()->getJointVector()[40]->lowerBound(0);
			vmax = _manipulationSolver->robot()->getJointVector()[40]->upperBound(0);
			(*config)[40] = p3d_random(vmin,vmax);
			vmin = _manipulationSolver->robot()->getJointVector()[41]->lowerBound(0);
			vmax = _manipulationSolver->robot()->getJointVector()[41]->upperBound(0);
			(*config)[41] = p3d_random(vmin,vmax);
		}
		if ((it->first == "head") && (it->second == true))
		{
			vmin = _manipulationSolver->robot()->getJointVector()[42]->lowerBound(0);
			vmax = _manipulationSolver->robot()->getJointVector()[42]->upperBound(0);
			(*config)[42] = p3d_random(vmin,vmax);
			vmin = _manipulationSolver->robot()->getJointVector()[43]->lowerBound(0);
			vmax = _manipulationSolver->robot()->getJointVector()[43]->upperBound(0);
			(*config)[43] = p3d_random(vmin,vmax);
		}
		if ((it->first == "torso") && (it->second == true))
		{
			vmin = _manipulationSolver->robot()->getJointVector()[19]->lowerBound(0);
			vmax = _manipulationSolver->robot()->getJointVector()[19]->upperBound(0);
			(*config)[21] = p3d_random(vmin,vmax);
		}
	}
}

bool RomeoWrap::generateConfig(confPtr_t &inputConf, confPtr_t &outputConf, p3d_matrix4 tragetPos, int armId, bool randomize)
{
	double Tx_eef, Ty_eef, Tz_eef, Rx_eef, Ry_eef, Rz_eef;
	hpp::model::vector3_t eef, tmp;
	hpp::model::matrix3_t rotation, footRotation;
	p3d_matrix3 rot;
	std::vector<bool> true_vect (3, true);
	std::string e ("eef");

	hpp::core::ConfigurationPtr_t config = createHppConfiguration(inputConf);
	_manipulationSolver->robot()->currentConfiguration(*config);

	if (armId == 1)
		_eef_joint = "LWristPitch";
	else
		_eef_joint = "RWristPitch";

	p3d_mat4ExtractPosReverseOrder(tragetPos, &Tx_eef, &Ty_eef, &Tz_eef, &Rx_eef, &Ry_eef, &Rz_eef);
	p3d_mat4ExtractRotMatrix (tragetPos, rot);
	eef[0] = Tx_eef; tmp[0] = 0;
	eef[1] = Ty_eef; tmp[1] = 0;
	eef[2] = Tz_eef; tmp[2] = 0;

	for (int i = 0; i < 3; i ++)
		for (int j = 0; j < 3; j ++)
			rotation(i, j) = rot[i][j];

	addPositionConstraint (e + std::string ("/position"), std::string (_eef_joint), tmp, eef, true_vect);
	addOrientationConstraint (e + std::string ("/orientation"), std::string (_eef_joint), rotation, true_vect);

	initializeConstraints();

	if (randomize)
	{
		std::map<string, bool> randomizedParts;
		if (armId == 1)
			randomizedParts["leftArm"] = true;
		else
			randomizedParts["rightArm"] = true;
		randomizedParts["leftLeg"] = false;
		randomizedParts["rightLeg"] = false;
		randomizedParts["torso"] = true;
		randomizedParts["neck"] = false;
		randomizedParts["head"] = true;
		randomizeConfig(config, randomizedParts);
	}

	_manipulationSolver->robot()->currentConfiguration(*config);

	double *timeOut;
	timeOut = new double(0);
	bool success = applyConstraints (config, timeOut);
	std::cout << "IK result : " << (success ? "SUCCESS" : "FAIL") << ", " << *timeOut << "s." << std::endl;
	if (!success)
		return false;

	confPtr_t sol = createConfiguration(config);
	outputConf = std::tr1::shared_ptr<Configuration>(new Configuration(*sol));
	_move3dRobot->setAndUpdate(*inputConf);
	return true;
}













/* Old Function to generate the 4 configs for a grasp
*/
std::vector<confPtr_t> RomeoWrap::generateGraspConfigs(confPtr_t &q, gpGrasp grasp, p3d_matrix4 objectPos,
													   p3d_vector3 approachDir, double approachOffset,
													   p3d_vector3 extractDir, double extractOffset)
{
	std::vector<confPtr_t> qVect;
	p3d_matrix4 trans;
	p3d_matrix3 rot;
	double Tx_eef, Ty_eef, Tz_eef, Rx_eef, Ry_eef, Rz_eef;

	// Set the robot in HPP as he (it ?) is in Move3d
	hpp::core::ConfigurationPtr_t config = createHppConfiguration(q);
	_manipulationSolver->robot()->currentConfiguration(*config);

	_grasp = grasp;
	if (grasp.hand_type == GP_ROMEO_LEFT_HAND)
		_eef_joint = "LWristPitch";
	else if (grasp.hand_type == GP_ROMEO_RIGHT_HAND)
		_eef_joint = "RWristPitch";
	else
	{
		std::cout << "ERROR : grasp does not contain Romeo hand type (" << grasp.hand_type << ")" << std::endl;
		return qVect;
	}

	p3d_mat4Mult(objectPos, grasp.frame, trans);
	p3d_mat4ExtractPosReverseOrder(trans, &Tx_eef, &Ty_eef, &Tz_eef, &Rx_eef, &Ry_eef, &Rz_eef);

	p3d_matrix4 correctionRotMat, M;
	// Matrix to correct grasp frame by -90 on X and -90 on Z axis
	correctionRotMat[0][0] = 0;
	correctionRotMat[0][1] = 1;
	correctionRotMat[0][2] = 0;
	correctionRotMat[0][3] = 0;
	correctionRotMat[1][0] = 0;
	correctionRotMat[1][1] = 0;
	correctionRotMat[1][2] = 1;
	correctionRotMat[1][3] = 0;
	correctionRotMat[2][0] = 1;
	correctionRotMat[2][1] = 0;
	correctionRotMat[2][2] = 0;
	correctionRotMat[2][3] = 0;
	correctionRotMat[3][0] = 0;
	correctionRotMat[3][1] = 0;
	correctionRotMat[3][2] = 0;
	correctionRotMat[3][3] = 1;

	p3d_mat4Mult(trans, correctionRotMat, M);
	p3d_mat4ExtractRotMatrix (M, rot);

	std::string e ("eef");
	hpp::model::vector3_t eef, tmp;
	eef[0] = Tx_eef; tmp[0] = 0;
	eef[1] = Ty_eef; tmp[1] = 0;
	eef[2] = Tz_eef; tmp[2] = 0;

	std::vector<bool> true_vect (3, true);
	hpp::model::matrix3_t rotation, footRotation;
	for (int i = 0; i < 3; i ++)
		for (int j = 0; j < 3; j ++)
			rotation(i, j) = rot[i][j];

	addPositionConstraint (e + std::string ("/position"), std::string (_eef_joint), tmp, eef, true_vect);
	addOrientationConstraint (e + std::string ("/orientation"), std::string (_eef_joint), rotation, true_vect);

	std::string lA = "LeftAnkle";
	std::string _leftAnkle_joint = "LAnkleRoll";
	hpp::model::vector3_t footT;
	p3d_vector4 quat;
	p3d_matrix3 footR;
	p3d_matrix4 Ma;

	footT[0] = _manipulationSolver->robot()->getJointByName(_leftAnkle_joint)->robot()->currentConfiguration()[0];
	footT[1] = _manipulationSolver->robot()->getJointByName(_leftAnkle_joint)->robot()->currentConfiguration()[1];
	footT[2] = _manipulationSolver->robot()->getJointByName(_leftAnkle_joint)->robot()->currentConfiguration()[2];
	quat[0] = _manipulationSolver->robot()->getJointByName(_leftAnkle_joint)->robot()->currentConfiguration()[3];
	quat[1] = _manipulationSolver->robot()->getJointByName(_leftAnkle_joint)->robot()->currentConfiguration()[4];
	quat[2] = _manipulationSolver->robot()->getJointByName(_leftAnkle_joint)->robot()->currentConfiguration()[5];
	quat[3] = _manipulationSolver->robot()->getJointByName(_leftAnkle_joint)->robot()->currentConfiguration()[6];
	p3d_quaternion_to_matrix3(quat, footR);
	p3d_mat4SetRotMatrix(footR, Ma);
	// p3d_mat4ExtractPosReverseOrder(Ma, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);
	for (int i = 0; i < 3; i ++)
		for (int j = 0; j < 3; j ++)
			footRotation(i, j) = footR[i][j];
			// footRotation(i, j) = ankleRot.getRow(i)[j];
	// true_vect[2] = false;
	// addPositionConstraint(lA + std::string ("/position"), std::string(_leftAnkle_joint), tmp, footT, true_vect);
	// std::vector<bool> false_vect (3, false);
	// false_vect[2] = true;
	// addOrientationConstraint(lA + std::string ("/orientation"), std::string (_leftAnkle_joint), footRotation, false_vect);

	// Add some LockedDof to accelerate IK
	addLockDof ("LFinger11", getJointDofValue ("LFinger11"));
	addLockDof ("LFinger12", getJointDofValue ("LFinger12"));
	addLockDof ("LFinger13", getJointDofValue ("LFinger13"));
	addLockDof ("LFinger21", getJointDofValue ("LFinger21"));
	addLockDof ("LFinger22", getJointDofValue ("LFinger22"));
	addLockDof ("LFinger23", getJointDofValue ("LFinger23"));
	addLockDof ("LFinger31", getJointDofValue ("LFinger31"));
	addLockDof ("LFinger32", getJointDofValue ("LFinger32"));
	addLockDof ("LFinger33", getJointDofValue ("LFinger33"));
	addLockDof ("LThumb1",   getJointDofValue ("LThumb1") );
	addLockDof ("LThumb2",   getJointDofValue ("LThumb2") );
	addLockDof ("LThumb3",   getJointDofValue ("LThumb3") );
  
	addLockDof ("RFinger11", getJointDofValue ("RFinger11"));
	addLockDof ("RFinger12", getJointDofValue ("RFinger12"));
	addLockDof ("RFinger13", getJointDofValue ("RFinger13"));
	addLockDof ("RFinger21", getJointDofValue ("RFinger21"));
	addLockDof ("RFinger22", getJointDofValue ("RFinger22"));
	addLockDof ("RFinger23", getJointDofValue ("RFinger23"));
	addLockDof ("RFinger31", getJointDofValue ("RFinger31"));
	addLockDof ("RFinger32", getJointDofValue ("RFinger32"));
	addLockDof ("RFinger33", getJointDofValue ("RFinger33"));
	addLockDof ("RThumb1",   getJointDofValue ("RThumb1")  );
	addLockDof ("RThumb2",   getJointDofValue ("RThumb2")  );
	addLockDof ("RThumb3",   getJointDofValue ("RThumb3")  );
  
	addLockDof ("LToePitch", getJointDofValue ("LToePitch"));
	addLockDof ("RToePitch", getJointDofValue ("RToePitch"));
  
	// addLockDof ("NeckYaw",   getJointDofValue ("NeckYaw")  );
	// addLockDof ("NeckPitch", getJointDofValue ("NeckPitch"));
	// addLockDof ("HeadPitch", getJointDofValue ("HeadPitch"));
	// addLockDof ("HeadRoll",  getJointDofValue ("HeadRoll") );
  
	addLockDof ("LEyeYaw",   getJointDofValue ("LEyeYaw")  );
	addLockDof ("LEyePitch", getJointDofValue ("LEyePitch"));
	addLockDof ("REyeYaw",   getJointDofValue ("REyeYaw")  );
	addLockDof ("REyePitch", getJointDofValue ("REyePitch"));


	double *timeOut;
	timeOut = new double;
	bool success = applyConstraints (config, timeOut);
	std::cout << "IK grasp result : " << (success ? "SUCCESS" : "FAIL") << ", " << *timeOut << "s." << std::endl;
	if (!success)
		return qVect;

	confPtr_t confSol = createConfiguration (config);
	confPtr_t confOpenGrasp = createConfiguration (config);
	confPtr_t confCloseGrasp = createConfiguration (config);

	if (!generateOpenGraspConf (confSol, confOpenGrasp))
		return qVect;
	if (!generateClosedGraspConf (confSol, confCloseGrasp))
		return qVect;

	hpp::model::vector3_t approach_eef = hpp::model::vector3_t (Tx_eef, Ty_eef, Tz_eef);
	for (int i = 0; i < 3; i++)
		approach_eef[i] += approachDir[i] * approachOffset;
	std::cout << std::endl;
	updatePositionConstraintRef (e + std::string ("/position"), approach_eef);
	success = applyConstraints (config, timeOut);
	std::cout << "IK approach result : " << (success ? "SUCCESS" : "FAIL") << ", " << *timeOut << "s." << std::endl;
	if (!success)
		return qVect;
	confSol = createConfiguration (config);
	// Open the hand of the robot for the approach conf
	confPtr_t approachConf = createConfiguration (config);
	if (!generateOpenGraspConf (confSol, approachConf))
		return qVect;

	hpp::model::vector3_t extract_eef = hpp::model::vector3_t (Tx_eef, Ty_eef, Tz_eef);
	for (int i = 0; i < 3; i++)
		extract_eef[i] += extractDir[i] * extractOffset;
	updatePositionConstraintRef (e + std::string ("/position"), extract_eef);
	success = applyConstraints (config, timeOut);
	std::cout << "IK extract result : " << (success ? "SUCCESS" : "FAIL") << ", " << *timeOut << "s." << std::endl;
	if (!success)
		return qVect;
	confSol = createConfiguration (config);
	confPtr_t extractConf = createConfiguration (config);
	if (!generateClosedGraspConf (confSol, extractConf))
		return qVect;

	qVect.push_back (approachConf);
	qVect.push_back (confOpenGrasp);
	qVect.push_back (confCloseGrasp);
	qVect.push_back (extractConf);

	return qVect;
}

#endif /* HPP */