/*
 *  compco.cpp
 *  BioMove3D
 *
 *  Created by Jim Mainprice on 27/04/10.
 *  Copyright 2010 LAAS/CNRS. All rights reserved.
 *
 */

#include "API/Roadmap/compco.hpp"
#include "API/Roadmap/graph.hpp"

#include "../p3d/env.hpp"

#include "Planner-pkg.h"

#include <iterator>

using namespace std;
using namespace tr1;


/**
 * Out of use!!!
 */
ConnectedComponent::ConnectedComponent(Graph * G, p3d_compco * C) :
	m_Compco(C), m_Graph(G), m_Id(-1)
{
	cout << "Warning: constructor out of use ConnectedComponent(Graph* G, p3d_compco* C)" << endl;
}


/**
 * Constructor.
 */
ConnectedComponent::ConnectedComponent(Graph * G, Node * N) :
	m_Graph(G), m_Id(-1)
{
	m_Compco = MY_ALLOC(p3d_compco, 1);

	m_Compco->num = G->getNumberOfCompco();
	m_Compco->nbRefinNodes = 0;
	m_Compco->minCost = P3D_HUGE;
	m_Compco->maxCost = 0.;
	m_Compco->maxUrmsonCost = 0.;
	//  m_Compco->AnaSuccessTab = MY_ALLOC(int,nbAnaSuccesSamples);
	m_Compco->nbTests = 0;
	// Warning the compco are handled in the boost graph
	// p3d_create_compco(m_Graph->getGraphStruct(), N->getNodeStruct());

	m_Nodes.push_back(N);
	N->setConnectedComponent(this);

	// initialize the bounding box
	double * conf = N->getConfiguration()->getConfigStruct();
	for (unsigned i(0); i < G->getRobot()->getNumberOfJoints(); ++i) {
		p3d_jnt * jntPt = G->getRobot()->getRobotStruct()->joints[i];
		for (unsigned j(0); j < jntPt->dof_equiv_nbr; ++j) {
			unsigned k = jntPt->index_dof + j;
			bBoxMin.push_back(conf[k]);
			bBoxMax.push_back(conf[k]);
		}
	}
	bBoxVolume = 0.;
}


/**
 * Get the id of the connected component.
 */
unsigned ConnectedComponent::getId() const
{
	return m_Compco->num;
}


/**
 * Get the number of refinement nodes.
 */
unsigned ConnectedComponent::getNumberOfRefinementNodes() const
{
    return m_Compco->nbRefinNodes;
}


/**
 * Add the given number to the current number of refinement nodes.
 */
void ConnectedComponent::updateNumberOfRefinementNodes()
{
    m_Compco->nbRefinNodes += 1;
}


/**
 * Add a node to the connected component.
 * @param N the node to be added
 */
void ConnectedComponent::addNode(Node * N)
{
	m_Nodes.push_back(N);
	N->setConnectedComponent(this);

	// update the bounding box
	double * conf = N->getConfiguration()->getConfigStruct();
	unsigned counter(0);
	bool change(false);
	unsigned changeID;
	for (unsigned i(0); i < m_Graph->getRobot()->getNumberOfJoints(); ++i)
	{
		p3d_jnt * jntPt = m_Graph->getRobot()->getRobotStruct()->joints[i];
		for (unsigned j(0); j < jntPt->dof_equiv_nbr; ++j)
		{
			unsigned k = jntPt->index_dof + j;
			if (conf[k] < bBoxMin[counter]) {
				bBoxMin[counter] = conf[k];
				change = true;
				changeID = counter;
			}
			if (conf[k] > bBoxMax[counter]) {
				bBoxMax[counter] = conf[k];
				change = true;
				changeID = counter;
			}
			++counter;
		}
	}
	// if the bounding box has changed, compute the new volume
	if (change)
	{
		bBoxVolume = bBoxMax[changeID] - bBoxMin[changeID];
		for (unsigned i(0); i < bBoxMin.size(); ++i) {
			double length = bBoxMax[i] - bBoxMin[i];
			if (i != changeID && length > 0.)
				bBoxVolume *= length;
		}
	}
}


void ConnectedComponent::addToReachableList(ConnectedComponent* Comp)
{
	// This function is a copy of
	// p3d_add_compco_to_reachable_list
	m_CanReach.push_back(Comp);
}

//! This function is a copy of p3d_compco_linked_to_compco
bool ConnectedComponent::isLinkedToCompco(ConnectedComponent* Comp)
{
	if( find (m_CanReach.begin(), m_CanReach.end(), Comp ) == m_CanReach.end() )
		return false;
	else
		return true;
}

void ConnectedComponent::addToReachableListAndUpdatePredecessors(ConnectedComponent* compco_add)
{
	// This function is a copy of
	// p3d_add_compco_to_reachable_list_and_update_predecessors
	
	m_CanReach.push_back(compco_add);
	
	vector<ConnectedComponent*> v = m_Graph->getConnectedComponents();
	vector<ConnectedComponent*>::iterator compco;
	
	for (compco = v.begin(); compco!=v.end(); ++compco) 
	{
		if( (*compco)->isLinkedToCompco( this ) )
		{
			(*compco)->addToReachableList( compco_add );
		}
	}
}

//! This function is a copy of p3d_merge_comp which updates the 
//! graph and the connected components
void ConnectedComponent::mergeWith(ConnectedComponent* compco)
{
	// The nodes of C2 are now in C1 
	for (int i = 0; i < int(compco->m_Nodes.size()); i++)
	{
		Node* tmpNode = compco->m_Nodes[i];
		tmpNode->setConnectedComponent(this);
		addNode(tmpNode);
	}

	// update the bounding box of this connected component, based on the bounding box of compco
	const std::vector<double> & bBoxMin2 = compco->getLowerBBox();
	const std::vector<double> & bBoxMax2 = compco->getUpperBBox();
	for (unsigned i(0); i < bBoxMin.size(); ++i)
	{
		if (bBoxMin2[i] < bBoxMin[i])
			bBoxMin[i] = bBoxMin2[i];
		if (bBoxMax2[i] > bBoxMax[i])
			bBoxMax[i] = bBoxMax2[i];
	}
	if (bBoxVolume != 0. || compco->getBBoxVolume() != 0.)
	{
		bBoxVolume = 1.;
		for (unsigned i(0); i < bBoxMin.size(); ++i)
		{
			double length = bBoxMax[i] - bBoxMin[i];
			if (length > 0.)
				bBoxVolume *= length;
		}
	}

	// All the compcos that can reach C2 can now reach C1 
	if (m_Graph->getGraphStruct()->oriented)
	{
		cout << "Oriented" << endl;
		p3d_list_compco * ListCompcoScan;
		ListCompcoScan = compco->getCompcoStruct()->canreach;
		while (ListCompcoScan != NULL)
		{
			p3d_add_compco_to_reachable_list_and_update_predecessors(m_Graph->getGraphStruct(),
					getCompcoStruct(), ListCompcoScan->comp);
			ListCompcoScan = ListCompcoScan->next;
		}
	}
	
	// C2 is deleted from the graph 
	// and throw an exception if the p3d and C++ are not the same
	m_Graph->removeCompco(compco);
	getNumberOfNodes();

	if (m_Graph->getGraphStruct()->oriented)
		m_Graph->mergeCheck();
}


/**
 * Return the node whose configuration is the closest (in the C-space) to the given configuration.
 */
Node * ConnectedComponent::nearestNeighbor(confPtr_t conf,bool backward)
{
	double distance;
	double minDistance = numeric_limits<double>::max();
	Node * nearestNode(NULL);
    Configuration *qA=conf.get();
    if(!qA)
        cout<<"error: ConnectedComponent::nearestNeighbor: qA is null pointer"<<endl;

	for (vector<Node *>::iterator node = m_Nodes.begin(); node != m_Nodes.end(); ++node)
	{
		// only take into account the nodes that have not been discarded
		if (!(*node)->getNodeStruct()->IsDiscarded)
		{
            if(!(*node))
                cout<<"error: ConnectedComponent::nearestNeighbor: (*node) is null pointer"<<endl;
            p3d_rob *robotPt=(*node)->getRobot()->getRobotStruct();
            Configuration *qB=(*node)->getConfiguration().get();
            if(!qB)
                cout<<"error: ConnectedComponent::nearestNeighbor: qB is null pointer"<<endl;
            if(p3d_has_a_quad_joint(robotPt) && !ENV.getBool(Env::armOnly)){
                if(ENV.getBool(Env::isHybrid) || !ENV.getBool(Env::useMetric))
                    distance=p3d_square_dist_config_quadspline(robotPt,qA->getConfigStruct(),qB->getConfigStruct());
                else{
                    if(backward)
                        distance=p3d_quadspline_kino_metric(robotPt,qA->getConfigStruct(),qB->getConfigStruct());
                    else
                        distance=p3d_quadspline_kino_metric(robotPt,qB->getConfigStruct(),qA->getConfigStruct());
                }
            }
            else
                distance=conf->squareDistance(qB);

			if (distance < minDistance) {
				minDistance = distance;
				nearestNode = *node;
			}
		}
	}

	return nearestNode;
}


/**
 * Return the k nodes whose configurations are the closest (in the C-space) to the given configuration.
 */
vector<Node *> ConnectedComponent::kNearestNeighbors(unsigned k, confPtr_t conf)
{
	double distance;
	vector<Node *> nearestNodes;
	vector<pair<double, Node *> > nodeMap;

	// compute the square distance between each node and the given conformation
	for (vector<Node *>::iterator node = m_Nodes.begin(); node != m_Nodes.end(); ++node)
	{
		// only take into account the nodes that have not been discarded
		if (!(*node)->getNodeStruct()->IsDiscarded)
		{
			distance = conf->squareDistance((*node)->getConfiguration().get());
			nodeMap.push_back(make_pair(distance, *node));
		}
	}

    // sort the nodes based on their distance to the given conformation
    sort(nodeMap.begin(), nodeMap.end());

    // return the first k nodes in a vector
    unsigned bound = MIN(k, nodeMap.size());
    for (unsigned i(0); i < bound; ++i)
    	nearestNodes.push_back(nodeMap[i].second);
    return nearestNodes;
}


/**
 * Return the nodes whose configurations are within a ball (in the C-space) having the given radius
 * and centered on the given configuration.
 */
vector<Node *> ConnectedComponent::neighborsInBall(confPtr_t conf, double radius)
{
	vector<Node *> nearestNodes;
	double sqrRadius = radius * radius;

	for (vector<Node *>::iterator node = m_Nodes.begin(); node != m_Nodes.end(); ++node)
	{
		// only take into account the nodes that have not been discarded
		if (!(*node)->getNodeStruct()->IsDiscarded &&
				(conf->squareDistance((*node)->getConfiguration().get()) < sqrRadius))
			nearestNodes.push_back(*node);
	}

	return nearestNodes;
}


Node * ConnectedComponent::nearestWeightNeighbour(confPtr_t q, bool weighted, int distConfigChoice)
{
	double current_dist, current_score;
	double best_score = numeric_limits<double>::max();
	Node * bestNode(NULL);
	Node * node(NULL);

	// go through all the nodes of the connected component
	for (unsigned i(0); i < m_Nodes.size(); ++i)
	{
		node = m_Nodes[i];

		// only take into account the nodes that have not been discarded
		if (!node->getNodeStruct()->IsDiscarded)
		{
			current_dist = q->dist(*(node->getConfiguration()), distConfigChoice);
			current_score = current_dist * (weighted ? p3d_GetNodeWeight(node->getNodeStruct()) : 1.0);

			if (current_score < best_score) {
				best_score = current_score;
				bestNode = node;
			}
		}
	}

	return bestNode;
}


/**
 * K Nearest Weighted Neighbors in connected component.
 * Return the K Nearest Neighbors of the configuration q in the connected component within a minimum radius.
 */
vector<Node*> ConnectedComponent::KNearestWeightNeighbour(confPtr_t q, int K, double radius, bool weighted,
        int distConfigChoice)
{
    double score;
    vector<pair<double, Node*> > nearNodesQueue;
    vector<Node*> nearNodes;

    // for each node in the graph
    for (vector<Node*>::iterator it = m_Nodes.begin(); it != m_Nodes.end(); ++it)
    {
        // compute its distance to the configuration q
        score = q->dist(*(*it)->getConfiguration(), distConfigChoice);
        //*(weighted ? p3d_GetNodeWeight((*it)->getNodeStruct()) : 1.0);

        // add it to the queue only if it is within radius
        if (score <= radius)
            nearNodesQueue.push_back(make_pair(score, *it));
    }

    // sort the queue
    sort(nearNodesQueue.begin(), nearNodesQueue.end());

    // put the first K nodes of the queue in a vector
    unsigned bound = MIN(K, nearNodesQueue.size());
    for (unsigned i = 0; i < bound; ++i) {
        Node* node = nearNodesQueue[i].second;
        if (node)
            nearNodes.push_back(node);
    }

    return nearNodes;
}


Node* ConnectedComponent::searchConf(Configuration& q) {
  
  for(int i=0;i<int(m_Nodes.size());i++)
  {
    if(m_Nodes[i]->getConfiguration()->equal(q))
    {
      return m_Nodes[i];
    }
  }
  return NULL;
}


/**
 * Return a node, randomly chosen in the connected component.
 */
Node* ConnectedComponent::randomNode()
{
    int RandId = (int) floor(p3d_random(0.0, (double) m_Nodes.size() - EPS6));
    return m_Nodes[RandId];
}



/**
 * Get the temperature of the component.
 */
double ConnectedComponent::getTemperature()
{
    return m_Compco->temperature;
}

/**
 * Set the temperature of the component.
 */
void ConnectedComponent::setTemperature(double temp)
{
    m_Compco->temperature = temp;
}
