#include "facts.hpp"
#include "P3d-pkg.h"
#include "Collision-pkg.h"
#include "planEnvironment.hpp"
#include "GTP/GTPTools/geometricTools/geometricComputations.hpp"
#include "GTP/GTPTools/geometricTools/geometricForms-pkg.hpp"

#ifdef HRI_PLANNER
#include <libmove3d/hri/hri.h>
#include <libmove3d/hri/hri_visibility_proto.h>
#include <libmove3d/hri/hri_affordance_proto.h>
#endif

using namespace std;

vector<fact_id> FactsManager::factList;

double isNextToDist = 0.2;//meters
double isOnSteepness = 0.25;//rad
double isOnthreshold = 0.04;//meters
double isInStepdiv = 20; //ratio to devide with
double isInEpsilonForIngnoringObj = 0.005;
double clutterUpLimit = 0.3;

double percentOfSeenObj = 0.7;

vector<string> objToIgnore;



bool inIgnoreList(string robName)
{
    for (unsigned int i = 0; i < objToIgnore.size(); i++)
    {
        if (robName.find(objToIgnore.at(i)) != string::npos)
        {
            return true;
        }
    }
    return false;
}

bool inOrigin(Robot* entity)
{
    confPtr_t q = entity->getCurrentPos();
    if(entity->getHriAgent()) {
      if(entity->getHriAgent()->type == HRI_HERAKLES)
      {
        int firstIndexOfHumanDof = entity->getJoint("Pelvis")->getIndexOfFirstDof();
        if ( ( (*q)[firstIndexOfHumanDof] == 0.0) && ( (*q)[firstIndexOfHumanDof + 1] == 0.0) )
          return true;
      }
    }else if ( ( (*q)[6] == 0.0) && ( (*q)[7] == 0.0)
        && ( (*q)[8] == 0.0) ){
      return true;
    }
    return false;
}   


FactsManager::FactsManager()
{

    fact_id FI;
    FI.factType = "IsVisible";
    FI.factSubType = "";
    FI.isMainAnAgent = true;
    factList.push_back(FI);

    FI.factType = "IsVisible";
    FI.factSubType = "HeadMotion";
    FI.isMainAnAgent = true;
    factList.push_back(FI);

    FI.factType = "IsReachable";
    FI.factSubType = "Distance";
    FI.isMainAnAgent = true;
    factList.push_back(FI);

    FI.factType = "IsReachable";
    FI.factSubType = "CollisionFree";
    FI.isMainAnAgent = true;
    factList.push_back(FI);

    FI.factType = "IsReachable";
    FI.factSubType = "GIK";
    FI.isMainAnAgent = true;
    factList.push_back(FI);


    FI.factType = "IsOn";
    FI.factSubType = "";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    FI.factType = "IsOn";
    FI.factSubType = "IsAbove";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    FI.factType = "IsIn";
    FI.factSubType = "";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    FI.factType = "IsIn";
    FI.factSubType = "IsCovered";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    FI.factType = "IsIn";
    FI.factSubType = "IsContainedBy";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    FI.factType = "IsNextTo";
    FI.factSubType = "";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    FI.factType = "Size";
    FI.factSubType = "IsSmaller";
    FI.isMainAnAgent = false;
    factList.push_back(FI);


    FI.factType = "Size";
    FI.factSubType = "IsBigger";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    FI.factType = "Cluttered";
    FI.factSubType = "IsMoreCluttered";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    FI.factType = "Cluttered";
    FI.factSubType = "IsLessCluttered";
    FI.isMainAnAgent = false;
    factList.push_back(FI);



    objToIgnore.push_back("SAHandRight");
    objToIgnore.push_back("HUM_BAR");
    objToIgnore.push_back("VISBALL_MIGHTABILITY");
    objToIgnore.push_back("PR2_GRIPPER_LEFT");
    objToIgnore.push_back("PR2_GRIPPER");
    objToIgnore.push_back("SAHandRight2");
    objToIgnore.push_back("PR_2CYLINDER");
    objToIgnore.push_back("HUMCYLINDER");
    objToIgnore.push_back("BROWN_BOX");


}

FactsManager::FactsManager(Scene *scen): sce(scen)
{
    FactsManager();
    initAll();
}

FactsManager::~FactsManager()
{
    cleanWSList();
}

void FactsManager::initAll()
{
    WSlistId = 0;
}

bool FactsManager::disableFacts(string s)
{
    bool test = false;
    for (unsigned int i = 0; i < factList.size(); i++)
    {
        fact_id fi = factList.at(i);
        if (fi.factType == s)
        {
            test = true;
            break;
        }
    }

    if (test)
    {
        disabledFacts.push_back(s);
        cout << "disable fact " << s << endl;
    }
    else
        cout << "fact " << s << " does not exist" << endl;

    return test;
}

bool FactsManager::enableFacts(string s)
{
    int id = -1;
    for (unsigned int i = 0; i < disabledFacts.size(); i++)
    {
        if (s == disabledFacts.at(i))
        {
            id = i;
            break;
        }
    }
    if (id != -1)
    {
        disabledFacts.erase(disabledFacts.begin() + id);
        enableFacts(s);
        cout << "enable fact " << s << endl;
        return true;
    }
    else
        return false;
//    else
//    {
//        cout << "no such fact " << s << " to disable" << endl;
//        return false;
//    }
}

bool FactsManager::isFactEnabled(string s)
{
    for (unsigned int i = 0; i < disabledFacts.size(); i++)
    {
//        cout << "compare " << s << " and " << disabledFacts.at(i) << endl;
        if (s == disabledFacts.at(i))
        {
            cout << "do no compute " << s << endl;
            return false;
        }
    }
    cout << "compute " << s << endl;
    return true;
}



//base fcts
/** computing visibility fact
  * for now, it is done roughly
  * a first loop check if the center of mass of the target object is in the fov
  * second check the oclusions, and keep the percent of oject seen
  * repeat same procedure for a larger fov (representing head motion
  */
bool FactsManager::computeVisibility(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute)
{
    if (!mainEntity->getHriAgent() || (targetEntity->getName() == mainEntity->getName()) )
        return false;

    fact_id FI1;
    FI1.factType = "IsVisible";
    FI1.factSubType = "";
    FI1.isMainAnAgent = true;

    fact_id FI2;
    FI2.factType = "IsVisible";
    FI2.factSubType = "HeadMotion";
    FI2.isMainAnAgent = true;

    if (WS->getFact(FI1, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;


    //actual visibility computation
    HRI_AGENT* agent = mainEntity->getHriAgent();
//     angle computation between camjoint and target obj

    double a[3];
    double b[3];
    a[0] = agent->perspective->camjoint->abs_pos[0][0];
    a[1] = agent->perspective->camjoint->abs_pos[1][0];
    a[2] = agent->perspective->camjoint->abs_pos[2][0];

    p3d_oriented_BB *target_OBB = p3d_get_oriented_BB_obj(targetEntity->getJoint(0)->getJointStruct()->o);

    p3d_point targetCenter;
    targetCenter.x = (target_OBB->UpBackLeft.x +target_OBB->BottomFrontRight.x)/2;
    targetCenter.y = (target_OBB->UpBackLeft.y +target_OBB->BottomFrontRight.y)/2;
    targetCenter.z = (target_OBB->UpBackLeft.z +target_OBB->BottomFrontRight.z)/2;
    p3d_destroy_oriented_BB(target_OBB);


    b[0] = targetCenter.x - agent->perspective->camjoint->abs_pos[0][3];
    b[1] = targetCenter.y - agent->perspective->camjoint->abs_pos[1][3];
    b[2] = targetCenter.z - agent->perspective->camjoint->abs_pos[2][3];


    double first = (a[0]*b[0] + a[1]*b[1] + a[2]*b[2])/ (sqrt(a[0]*a[0] + a[1]*a[1] + a[2]*a[2])*sqrt(b[0]*b[0] + b[1]*b[1] + b[2]*b[2]));

    double angle = acos(first) ;

    if (angle > agent->perspective->fov/2)
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
    }
    else
    {
        // object in FOV, test if obstruction?
        double vis;

        if (PlanEnv->getBool(PlanParam::usingSpark))
        {
            //hri_initialize_visibility();
            g3d_is_object_visible_from_viewpoint(agent->perspective->camjoint->abs_pos, agent->perspective->fov, targetEntity->getRobotStruct(), &vis, 0);
            //int res = is_object_visible_for_agent(mainEntity->getHriAgent(), targetEntity->getRobotStruct(), 0.01, 0, 0);
            //printf("[GTP] Vis result: %f\n", vis);



/* hri_initialize_visibility();
            //g3d_is_object_visible_from_viewpoint(agent->perspective->camjoint->abs_pos, agent->perspective->fov, targetEntity->getRobotStruct(), &vis, 0);
            //int res = is_object_visible_for_agent(mainEntity->getHriAgent(), targetEntity->getRobotStruct(), 0.01, 0, 0);

            //HRI_VISIBILITY* vis_result = MY_ALLOC(HRI_VISIBILITY, 1);
            //HRI_ENTITY** ents = MY_ALLOC(HRI_ENTITY*, 1);
            HRI_VISIBILITY vis_result[1];
            HRI_ENTITY* ents[1];
            HRI_ENTITY* ent = new HRI_ENTITY();
            ent->robotPt = targetEntity->getRobotStruct();
            ent->disappeared = false;

            ents[0] = ent;

            g3d_compute_visibility_for_given_entities(ents, agent, vis_result, 1);
            printf("[GTP] Vis result: %d\n", vis_result[0]);
            free(ents[0]);
*/
        }
        else
        {
            vis= percentOfSeenObj +0.01;
        }

        //cout << "main = " << mainEntity->getName() << " target " << targetEntity->getName() << " vis " << vis << endl;


        if (vis > percentOfSeenObj) // more then half the object is visible
        {
            Fact f(FI1);
            f.setIsComputed(true);
            f.setMainEntity(mainEntity);
            f.setTargetEntity(targetEntity);
            f.setValueType(1);
            f.setBoolValue(true);
            f.setSpecValue(vis);

            WS->setFact(f);

            Fact f2(FI2);
            f2.setIsComputed(true);
            f2.setMainEntity(mainEntity);
            f2.setTargetEntity(targetEntity);
            f2.setValueType(1);
            f2.setBoolValue(true);
            f2.setSpecValue(vis);

            WS->setFact(f2);
        }
        return true;

    }

    if (WS->getFact(FI2, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;

    //actual visibility computation with head moving > using a wider fov
    if (angle > agent->perspective->fov)
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
    }
    else
    {
        // object in FOV, test if obstruction?
        double vis;

        if (PlanEnv->getBool(PlanParam::usingSpark))
        {
            hri_initialize_visibility();
            g3d_is_object_visible_from_viewpoint(agent->perspective->camjoint->abs_pos, agent->perspective->fov*2, targetEntity->getRobotStruct(), &vis, 0);
        }
        else
        {
            vis= 0.6;
        }

        if (vis > 0.5) // more then half the object is visible
        {
            Fact f(FI2);
            f.setIsComputed(true);
            f.setMainEntity(mainEntity);
            f.setTargetEntity(targetEntity);
            f.setValueType(1);
            f.setBoolValue(true);
            f.setSpecValue(vis);

            WS->setFact(f);
        }

    }
    return true;
}


/**
  *
  * the hand will be stocked in the sepecific value
  * 2 hand considered
  */
bool FactsManager::computeReachability(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute)
{

//    cout << "main = " << mainEntity->getName() << endl;
    GeometricComputations  GC = GeometricComputations();
    if (!mainEntity->getHriAgent() || targetEntity->getHriAgent() || !targetEntity->getObjectRob())
    {
        return true;
    }


    fact_id FI1;
    FI1.factType = "IsReachable";
    FI1.factSubType = "Distance";
    FI1.isMainAnAgent = true;

    fact_id FI2;
    FI2.factType = "IsReachable";
    FI2.factSubType = "CollisionFree";
    FI2.isMainAnAgent = true;

    fact_id FI3;
    FI3.factType = "IsReachable";
    FI3.factSubType = "GIK";
    FI3.isMainAnAgent = true;

    if (WS->getFact(FI1, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;

    //actual reacheability computation using only distances
//    HRI_AGENT* agent = hri_create_agent(mainEntity->getRobotStruct());
    HRI_AGENT* agent = mainEntity->getHriAgent();
//    cout << "main = " << mainEntity->getName() << " target = " << targetEntity->getName() << " dist = " << agent->manip->reach_tolerance << endl ;
//    cout << "end also : jnt are : " << mainEntity->getJoint(agent->manip->shoulder_joint_ids[0])->getName() << " end " << mainEntity->getJoint(agent->manip->shoulder_joint_ids[1])->getName() << endl;




//    p3d_point targetCenter = GC.getCenterOfObj(targetEntity);

    vector<double> dists;
    for (int i = 0; i < agent->manip->shoulder_joint_ids_no; i ++)
    {
        p3d_point shoulderP;
        shoulderP.x = mainEntity->getJoint(agent->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[0][3];
        shoulderP.y = mainEntity->getJoint(agent->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[1][3];
        shoulderP.z = mainEntity->getJoint(agent->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[2][3];


        Robot* visBall = sce->getRobotByName("VISBALL_MIGHTABILITY");
        double d = P3D_HUGE;
        if (!visBall)
        {
            p3d_point targetClose = GC.getClosestBBPointFromPoint(targetEntity, shoulderP);
            d = GC.distBetweenPoints(targetClose, shoulderP);
        }
        else
        {
            confPtr_t qv = visBall->getCurrentPos();
            (*qv)[6] = shoulderP.x;
            (*qv)[7] = shoulderP.y;
            (*qv)[8] = shoulderP.z;
            visBall->setAndUpdate(*qv);
            double vs;
            d = p3d_BB_obj_obj_extern_dist ( visBall->getJoint(0)->getJointStruct()->o , targetEntity->getJoint(0)->getJointStruct()->o, &vs );
            (*qv)[6] = 0;
            (*qv)[7] = 0;
            (*qv)[8] = 0;
            visBall->setAndUpdate(*qv);
        }

        dists.push_back(d);
//
//        p3d_point targetClose = GC.getAClosePointInBBFromPoint(targetEntity,shoulderP);




    }

    if (dists.size() < 2)
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI3,mainEntity,targetEntity);
        return true;
    }

    double res = 0;
    if (dists.at(0) - agent->manip->reach_tolerance < 0 && dists.at(1) - agent->manip->reach_tolerance < 0)
    {
        res = 3;
        // reachable by both hand
    }
    else if (dists.at(0) - agent->manip->reach_tolerance < 0)
    {
        res = 1;
        // reachable by right hand
    }
    else if (dists.at(1) - agent->manip->reach_tolerance < 0)
    {
        res = 2;
        // reachable by left hand
    }


    if (res == 0)
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
//        setFactToFalse(WS,FI2,mainEntity,targetEntity);
//        setFactToFalse(WS,FI3,mainEntity,targetEntity);
//        return true;
    }
    else
    {
        Fact f(FI1);
        f.setIsComputed(true);
        f.setMainEntity(mainEntity);
        f.setTargetEntity(targetEntity);
        f.setValueType(0);
        f.setBoolValue(true);
        f.setSpecValue(res);

        WS->setFact(f);
    }


    //actual reacheability computation using GIK
    if (WS->getFact(FI3, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;

    if (!targetEntity->getObjectRob())
    {
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI3,mainEntity,targetEntity);
        return true;
    }

    if(!targetEntity->getObjectRob()->isManipulable())
    {
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI3,mainEntity,targetEntity);
        return true;
    }


    double distance_tolerance = 0.3;

    configPt q = p3d_get_robot_config(mainEntity->getRobotStruct());
    configPt qCurr = p3d_copy_config(mainEntity->getRobotStruct(),q);

    res = 0;
    for (int i = 0; i < agent->manip->shoulder_joint_ids_no; i ++)
    {
        HRI_GIK_TASK_TYPE task;
        if (i == 0)
            task = GIK_RATREACH;
        else if (i == 1)
            task = GIK_LATREACH;

        p3d_point shoulderP;
        shoulderP.x = mainEntity->getJoint(agent->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[0][3];
        shoulderP.y = mainEntity->getJoint(agent->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[1][3];
        shoulderP.z = mainEntity->getJoint(agent->manip->shoulder_joint_ids[i])->getJointStruct()->abs_pos[2][3];

        p3d_point clPoint = GC.getClosestBBPointFromPoint(targetEntity, shoulderP);
        p3d_vector3 Tcoord;

        Tcoord[0] = clPoint.x;
        Tcoord[1] = clPoint.y;
        Tcoord[2] = clPoint.z;

        int IKSucceded = hri_agent_single_task_manip_move(mainEntity->getHriAgent(),
                                                          task,
                                                          &Tcoord,
                                                          distance_tolerance,
                                                          &q);


        if (IKSucceded == 1)
        {
            p3d_set_and_update_this_robot_conf(mainEntity->getRobotStruct(),q);
//            g3d_draw_allwin_active();
            p3d_col_deactivate_robot_robot(mainEntity->getRobotStruct(), targetEntity->getRobotStruct());

            for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
            {
                if ( sce->getRobot(i)->isAgent() &&  sce->getRobot(i)->getName() != mainEntity->getName())
                {
                    p3d_col_deactivate_robot_robot(mainEntity->getRobotStruct(), sce->getRobot(i)->getRobotStruct());
                }
            }

            int colA = p3d_col_test_robot_other_only(mainEntity->getRobotStruct(),0);
            if (colA == 1)
            {
                p3d_print_col_pair();
                p3d_set_and_update_this_robot_conf(mainEntity->getRobotStruct(),qCurr);
                continue;
            }
            p3d_col_activate_robot_robot(mainEntity->getRobotStruct(), targetEntity->getRobotStruct());
            for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
            {
                if ( sce->getRobot(i)->isAgent() &&  sce->getRobot(i)->getName() != mainEntity->getName())
                {
                    p3d_col_activate_robot_robot(mainEntity->getRobotStruct(), sce->getRobot(i)->getRobotStruct());
                }
            }

            if (task == GIK_RATREACH)
                res += 1;
            else if (task == GIK_LATREACH)
                res += 2;

//            cout << "res = " << res << endl;
//            g3d_draw_allwin_active();
//            sleep(3);
        }
        p3d_set_and_update_this_robot_conf(mainEntity->getRobotStruct(),qCurr);
    }


    if (res != 0)
    {
        Fact f(FI3);
        f.setIsComputed(true);
        f.setMainEntity(mainEntity);
        f.setTargetEntity(targetEntity);
        f.setValueType(0);
        f.setBoolValue(true);
        f.setSpecValue(res);

        WS->setFact(f);
    }
    else
    {
        setFactToFalse(WS,FI3,mainEntity,targetEntity);
    }

        //actual reacheability computation using distances and collision free space
    if (WS->getFact(FI2, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;


    setFactToFalse(WS,FI2,mainEntity,targetEntity);

    return true;

}


/**
  * computation of is On fact
  * an object A is On another one B in the next conditions:
  * The center of A is within the boundaries of the BB of B
  * B is not too tilted (less then 15deg)
  * TODO: Is On needs to go all the way to the ground (passing through multiple objects;
  * Is On is not computed for object with multiple joints (BB computation problems)
  * is above is a special case where there is no contact
  */
bool FactsManager::computeIsOn(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute)
{
    fact_id FI1;
    FI1.factType = "IsOn";
    FI1.factSubType = "";
    FI1.isMainAnAgent = false;

    fact_id FI2;
    FI2.factType = "IsOn";
    FI2.factSubType = "IsAbove";
    FI2.isMainAnAgent = false;

    if (!mainEntity->getObjectRob() || !targetEntity->getObjectRob() || targetEntity->getObjectRob()->getSupports().size() <= 0)
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        return true;
    }

    if (WS->getFact(FI1, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;

    //actual ison computation

    bool ok = false;
    bool isOn = false; //ison and above

    double dist= 1000;
    GeometricComputations  GC = GeometricComputations();

    unsigned int i = 0;
    while (i<targetEntity->getObjectRob()->getSupports().size() && !ok)
    {
        GeometricForm* gf = targetEntity->getObjectRob()->getSupports().at(i);
        p3d_point p = GC.getBotomCenterOfObj(mainEntity,0.03);
        dist = gf->distHToPoint(p);

        if (gf->isPointInForm(p) && (dist > 0 || fabs(dist) < isOnthreshold))
        {
            ok = true;
            if (fabs(dist) < isOnthreshold)
            {
                isOn =true;
            }
        }

        i++;
    }

    if (!ok)
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        return true;
    }

    if (isOn)
    {
        Fact f(FI1);
        f.setIsComputed(true);
        f.setMainEntity(mainEntity);
        f.setTargetEntity(targetEntity);
        f.setValueType(0);
        f.setBoolValue(true);
        f.setSpecValue(dist);

        WS->setFact(f);
        setFactToFalse(WS,FI1,targetEntity,mainEntity);
    }
    else
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
    }



    if (WS->getFact(FI2, mainEntity, targetEntity).isItComputed() && !forceRecompute)
    {
        return true;
    }

    //actual isOn computation with specification of object above another //reaching here mean that it is above!!

    if (ok && !isOn)
    {
        Fact f(FI2);
        f.setIsComputed(true);
        f.setMainEntity(mainEntity);
        f.setTargetEntity(targetEntity);
        f.setValueType(0);
        f.setBoolValue(true);
        f.setSpecValue(dist);

        WS->setFact(f);
        setFactToFalse(WS,FI2,targetEntity,mainEntity);
    }
    else
    {
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
    }

    return true;
}




/**
  * Compute is insinde fact
  * is in means at least half of the mainobject is in the OBB of the target object
  * no agents
  *
  */
bool FactsManager::computeIsIn(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute)
{

    fact_id FI1;
    FI1.factType = "IsIn";
    FI1.factSubType = "";
    FI1.isMainAnAgent = false;

    fact_id FI2;
    FI2.factType = "IsIn";
    FI2.factSubType = "IsCovered";
    FI2.isMainAnAgent = false;

    fact_id FI3;
    FI3.factType = "IsIn";
    FI3.factSubType = "IsContainedBy";
    FI3.isMainAnAgent = false;

    if (mainEntity->getName() == targetEntity->getName())
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI3,mainEntity,targetEntity);
        return true;
    }

    if (WS->getFact(FI1, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;

    //actual is in computation

    if (mainEntity->getNumberOfJoints() > 2 || targetEntity->getNumberOfJoints() > 2)
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI3,mainEntity,targetEntity);
        return true;
    }

    p3d_oriented_BB *main_OBB = p3d_get_oriented_BB_obj(mainEntity->getJoint(0)->getJointStruct()->o);
    p3d_oriented_BB *target_OBB = p3d_get_oriented_BB_obj(targetEntity->getJoint(0)->getJointStruct()->o);

    GeometricComputations GC = GeometricComputations();


    if (GC.distBetweenPoints(main_OBB->UpBackLeft, main_OBB->BottomBackLeft) < isInEpsilonForIngnoringObj
            || GC.distBetweenPoints(main_OBB->UpBackLeft, main_OBB->UpBackRight) < isInEpsilonForIngnoringObj
            || GC.distBetweenPoints(main_OBB->UpBackLeft, main_OBB->UpFrontLeft) < isInEpsilonForIngnoringObj
            || GC.distBetweenPoints(target_OBB->UpBackLeft, target_OBB->BottomBackLeft) < isInEpsilonForIngnoringObj
            || GC.distBetweenPoints(target_OBB->UpBackLeft, target_OBB->UpBackRight) < isInEpsilonForIngnoringObj
            || GC.distBetweenPoints(target_OBB->UpBackLeft, target_OBB->UpFrontLeft) < isInEpsilonForIngnoringObj)
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI3,mainEntity,targetEntity);
        return true;
    }

    p3d_point mainCenter;
    mainCenter.x = (main_OBB->UpBackLeft.x +main_OBB->BottomFrontRight.x)/2;
    mainCenter.y = (main_OBB->UpBackLeft.y +main_OBB->BottomFrontRight.y)/2;
    mainCenter.z = (main_OBB->UpBackLeft.z +main_OBB->BottomFrontRight.z)/2;

    if (GC.point_in_hexahedron(target_OBB->UpBackLeft,target_OBB->UpBackRight, target_OBB->UpFrontLeft, target_OBB->UpFrontRight,
                            target_OBB->BottomBackLeft,target_OBB->BottomBackRight, target_OBB->BottomFrontLeft, target_OBB->BottomFrontRight,
                            mainCenter) &&
        GC.hexahedron_in_hexahedron(target_OBB->UpBackLeft,target_OBB->UpBackRight, target_OBB->UpFrontLeft, target_OBB->UpFrontRight,
                                 target_OBB->BottomBackLeft,target_OBB->BottomBackRight, target_OBB->BottomFrontLeft, target_OBB->BottomFrontRight,
                                 main_OBB->UpBackLeft,main_OBB->UpBackRight, main_OBB->UpFrontLeft, main_OBB->UpFrontRight,
                                 main_OBB->BottomBackLeft,main_OBB->BottomBackRight, main_OBB->BottomFrontLeft, main_OBB->BottomFrontRight))
    {
        Fact f(FI1);
        f.setIsComputed(true);
        f.setMainEntity(mainEntity);
        f.setTargetEntity(targetEntity);
        f.setValueType(0);
        f.setBoolValue(true);
        f.setSpecValue(-1);

        WS->setFact(f);

        setFactToFalse(WS,FI1,targetEntity,mainEntity);
    }
    else
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI3,mainEntity,targetEntity);

        p3d_destroy_oriented_BB(main_OBB);
        p3d_destroy_oriented_BB(target_OBB);

        return true;
    }



    // for the two following facts, a step is needed, this step will be computed from the ratio of the smallest lenght of the main obb over
    // the biggest lenght of target OBB

    double mainSL = sqrt(pow(main_OBB->UpBackLeft.x - main_OBB->UpBackRight.x,2)
                       + pow(main_OBB->UpBackLeft.y - main_OBB->UpBackRight.y,2)
                       + pow(main_OBB->UpBackLeft.z - main_OBB->UpBackRight.z,2));

    double other = sqrt(pow(main_OBB->UpBackLeft.x - main_OBB->UpFrontLeft.x,2)
                      + pow(main_OBB->UpBackLeft.y - main_OBB->UpFrontLeft.y,2)
                      + pow(main_OBB->UpBackLeft.z - main_OBB->UpFrontLeft.z,2));

    if (mainSL > other)
        mainSL = other;

    double targetBL =  sqrt(pow(target_OBB->UpBackLeft.x - target_OBB->UpBackRight.x,2)
                          + pow(target_OBB->UpBackLeft.y - target_OBB->UpBackRight.y,2)
                          + pow(target_OBB->UpBackLeft.z - target_OBB->UpBackRight.z,2));

    other =            sqrt(pow(target_OBB->UpBackLeft.x - target_OBB->UpFrontLeft.x,2)
                          + pow(target_OBB->UpBackLeft.y - target_OBB->UpFrontLeft.y,2)
                          + pow(target_OBB->UpBackLeft.z - target_OBB->UpFrontLeft.z,2));

    if (targetBL < other)
        targetBL = other;

    double step = (mainSL/targetBL)/isInStepdiv;

    if (WS->getFact(FI2, mainEntity, targetEntity).isItComputed() && !forceRecompute)
    {
        p3d_destroy_oriented_BB(main_OBB);
        p3d_destroy_oriented_BB(target_OBB);
        return true;
    }

    //actual is covered computation

    confPtr_t q_init = mainEntity->getCurrentPos();
    confPtr_t q = mainEntity->getCurrentPos();

    // this is dirty and should be recoded proporly
    double onStep = 0;
    bool colTest = false;
    for (unsigned int i = 0; i < 3 ;i++)
    {
        colTest = false;
        q = mainEntity->getCurrentPos();
        for(int j = -1 ; j <2; j +=2)
        {
            if (i == 2 && j == -1)
                continue;
            colTest = false;
            onStep = 0;
            q = mainEntity->getCurrentPos();
            while (targetBL - fabs(onStep) > 0)
            {
                (*q)[6+i] = (*q)[6+i] + onStep;
                mainEntity->setAndUpdate(*q);
                if (pqp_collision_test(mainEntity->getJoint(0)->getJointStruct()->o, targetEntity->getJoint(0)->getJointStruct()->o) == 1) //collision
                {
                    colTest = true;
                    mainEntity->setAndUpdate(*q_init);
                    break;
                }
                onStep += j*step;
                mainEntity->setAndUpdate(*q_init);
            }
            mainEntity->setAndUpdate(*q_init);
            if (!colTest)
                break;
        }
        mainEntity->setAndUpdate(*q_init);
        if (!colTest)
            break;
    }

    if (colTest)
    {
        Fact f(FI2);
        f.setIsComputed(true);
        f.setMainEntity(mainEntity);
        f.setTargetEntity(targetEntity);
        f.setValueType(0);
        f.setBoolValue(true);
        f.setSpecValue(-1);

        WS->setFact(f);

        setFactToFalse(WS,FI2,targetEntity,mainEntity);

        setFactToFalse(WS,FI3,mainEntity,targetEntity);

        p3d_destroy_oriented_BB(main_OBB);
        p3d_destroy_oriented_BB(target_OBB);

        return true;
    }
    else
    {
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
    }

    if (WS->getFact(FI3, mainEntity, targetEntity).isItComputed() && !forceRecompute)
    {
        p3d_destroy_oriented_BB(main_OBB);
        p3d_destroy_oriented_BB(target_OBB);
        return true;
    }

    //actual is inside computation

    onStep = 0;
    colTest = false;
    for (unsigned int i = 0; i < 3 ;i++)
    {
        colTest = false;
        q = mainEntity->getCurrentPos();
        for(int j = -1 ; j <2; j +=2)
        {
            if (i == 2 && j == 1)
                continue;
            colTest = false;
            onStep = 0;
            q = mainEntity->getCurrentPos();
            while (targetBL - fabs(onStep) > 0)
            {
                (*q)[6+i] = (*q)[6+i] + onStep;
                mainEntity->setAndUpdate(*q);
                if (pqp_collision_test(mainEntity->getJoint(0)->getJointStruct()->o, targetEntity->getJoint(0)->getJointStruct()->o) == 1) //collision
                {
                    colTest = true;
                    mainEntity->setAndUpdate(*q_init);
                    break;
                }
                onStep += j*step;
                mainEntity->setAndUpdate(*q_init);
            }
            mainEntity->setAndUpdate(*q_init);
            if (!colTest)
                break;
        }
        mainEntity->setAndUpdate(*q_init);
        if (!colTest)
            break;
    }
    if (colTest)
    {
        Fact f(FI3);
        f.setIsComputed(true);
        f.setMainEntity(mainEntity);
        f.setTargetEntity(targetEntity);
        f.setValueType(0);
        f.setBoolValue(true);
        f.setSpecValue(-1);

        WS->setFact(f);

        setFactToFalse(WS,FI3,targetEntity,mainEntity);

        setFactToFalse(WS,FI2,mainEntity,targetEntity);
    }
    else
    {
        setFactToFalse(WS,FI3,mainEntity,targetEntity);
    }


    p3d_destroy_oriented_BB(main_OBB);
    p3d_destroy_oriented_BB(target_OBB);

    return true;
}


/**
  *  computation of is next to fact
  * an object is next to another if the closest distance between those two is smaller than isNextToDist (0.2m)
  * NOTE: if an object is next to another, the inverse is also true!
  */
bool FactsManager::computeIsNextTo(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute)
{

    if (mainEntity->getName() == targetEntity->getName())
        return true;

    fact_id FI;
    FI.factType = "IsNextTo";
    FI.factSubType = "";
    FI.isMainAnAgent = false;

    if (WS->getFact(FI, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;

    //actual is next to computation
    int nof_bodies = mainEntity->getRobotStruct()->no;
    p3d_vector3 *body = new p3d_vector3[nof_bodies];
    p3d_vector3 *other = new p3d_vector3[nof_bodies];

    double minDist = pqp_robot_robot_distance(mainEntity->getRobotStruct(),targetEntity->getRobotStruct(), body[0], other[0]);

    Fact f(FI);
    f.setIsComputed(true);
    f.setMainEntity(mainEntity);
    f.setTargetEntity(targetEntity);
    f.setValueType(0);
    f.setBoolValue(minDist < isNextToDist);
    f.setSpecValue(minDist);

    WS->setFact(f);


    f.setIsComputed(true);
    f.setMainEntity(targetEntity);
    f.setTargetEntity(mainEntity);
    f.setValueType(0);
    f.setBoolValue(minDist < isNextToDist);
    f.setSpecValue(minDist);

    WS->setFact(f);


    return true;
}

bool FactsManager::computeSize(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute)
{
    fact_id FI1;
    FI1.factType = "Size";
    FI1.factSubType = "IsSmaller";
    FI1.isMainAnAgent = false;

    fact_id FI2;
    FI2.factType = "Size";
    FI2.factSubType = "IsBigger";
    FI2.isMainAnAgent = false;

    if (mainEntity->getName() == targetEntity->getName())
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        return true;
    }

    if (WS->getFact(FI1, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;

    if (mainEntity->isAgent() || targetEntity->isAgent())
    {
        setFactToFalse(WS,FI1,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        return true;
    }

    //compute size of BB of main:
    double sizeMain;

    GeometricComputations GC = GeometricComputations();
    p3d_oriented_BB *main_OBB = p3d_get_oriented_BB_obj(mainEntity->getJoint(0)->getJointStruct()->o);

    sizeMain = GC.distBetweenPoints(main_OBB->BottomBackLeft, main_OBB->UpBackLeft)*
            GC.distBetweenPoints(main_OBB->BottomBackLeft, main_OBB->BottomBackRight)*
            GC.distBetweenPoints(main_OBB->BottomBackLeft, main_OBB->BottomFrontLeft);
    p3d_destroy_oriented_BB(main_OBB);


    //compute size of BB of target
    double sizeTarget;

    p3d_oriented_BB *target_OBB = p3d_get_oriented_BB_obj(targetEntity->getJoint(0)->getJointStruct()->o);

    sizeTarget = GC.distBetweenPoints(target_OBB->BottomBackLeft, target_OBB->UpBackLeft)*
            GC.distBetweenPoints(target_OBB->BottomBackLeft, target_OBB->BottomBackRight)*
            GC.distBetweenPoints(target_OBB->BottomBackLeft, target_OBB->BottomFrontLeft);

    p3d_destroy_oriented_BB(target_OBB);

    Fact f1(FI1);
    f1.setIsComputed(true);
    f1.setMainEntity(mainEntity);
    f1.setTargetEntity(targetEntity);
    f1.setValueType(0);
    f1.setBoolValue(sizeMain < sizeTarget);
    f1.setSpecValue(sizeMain);

    WS->setFact(f1);

    Fact f2(FI2);
    f2.setIsComputed(true);
    f2.setMainEntity(mainEntity);
    f2.setTargetEntity(targetEntity);
    f2.setValueType(0);
    f2.setBoolValue(sizeMain > sizeTarget);
    f2.setSpecValue(sizeMain);

    WS->setFact(f2);

    return true;
}


bool FactsManager::computeClutter(WorldStateFacts* WS,Robot* mainEntity, Robot* targetEntity, bool forceRecompute)
{
    fact_id FI;
    FI.factType = "Cluttered";
    FI.factSubType = "IsMoreCluttered";
    FI.isMainAnAgent = false;
    factList.push_back(FI);

    fact_id FI2;
    FI2.factType = "Cluttered";
    FI2.factSubType = "IsLessCluttered";
    FI2.isMainAnAgent = false;
    factList.push_back(FI);


    if (mainEntity->getName() == targetEntity->getName())
    {
        setFactToFalse(WS,FI,mainEntity,targetEntity);
        setFactToFalse(WS,FI,targetEntity,mainEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,targetEntity,mainEntity);
        return true;
    }

    if (WS->getFact(FI, mainEntity, targetEntity).isItComputed() && !forceRecompute)
        return true;

    if (mainEntity->isAgent() || targetEntity->isAgent())
    {
        setFactToFalse(WS,FI,mainEntity,targetEntity);
        setFactToFalse(WS,FI,targetEntity,mainEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,targetEntity,mainEntity);
        return true;
    }

    if (!mainEntity->getObjectRob() || !targetEntity->getObjectRob())
    {
        setFactToFalse(WS,FI,mainEntity,targetEntity);
        setFactToFalse(WS,FI,targetEntity,mainEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,targetEntity,mainEntity);
        return true;
    }

    if (mainEntity->getObjectRob()->getSupports().size() == 0 || targetEntity->getObjectRob()->getSupports().size() == 0
            || !mainEntity->getObjectRob()->isSupport() || !targetEntity->getObjectRob()->isSupport())
    {
        setFactToFalse(WS,FI,mainEntity,targetEntity);
        setFactToFalse(WS,FI,targetEntity,mainEntity);
        setFactToFalse(WS,FI2,mainEntity,targetEntity);
        setFactToFalse(WS,FI2,targetEntity,mainEntity);
        return true;
    }


//    cout << "main = " << mainEntity->getName() << " target = "  << targetEntity->getName() << endl;

    GeometricComputations GC = GeometricComputations();

    vector<GeometricForm*> vg = mainEntity->getObjectRob()->getSupports();

    int convertTo = 100000;

    double mainVol = 0;
    for (unsigned int i = 0; i < vg.size(); i++)
    {
        if (vg.at(i)->type == "rectangle")
        {
            RectangleForm* rf = (RectangleForm*)vg.at(i);
            p3d_point a,b,c,d,e,f,g,h;
            e = rf->getA();
            f = rf->getB();
            g = rf->getC();
            h = rf->getD();
            a = e;
            a.z = a.z + clutterUpLimit;
            b = f;
            b.z = b.z + clutterUpLimit;
            c = g;
            c.z = c.z + clutterUpLimit;
            d = h;
            d.z = d.z + clutterUpLimit;

            int nb = (int)(GC.distBetweenPoints(a,b) * GC.distBetweenPoints(a,c) * GC.distBetweenPoints(a,e) *convertTo);
//            nb = 100;


//            cout << "nb of point = " << nb << endl;
            if (nb == 0)
                continue;
            vector<p3d_point> vp = GC.getRandomPointsInHexahedron(nb,a,b,c,d,e,f,g,h);
            int nbIn = 0;
            for (unsigned int j = 0; j < vp.size(); j++)
            {
//                cout << "point nb = " << j << endl;
                for (unsigned int k = 0; k < sce->getNumberOfRobots(); k++)
                {

                    Robot* rtmp = sce->getRobot(k);
                    if (rtmp->getNumberOfJoints() > 2)
                        continue;

                    p3d_oriented_BB *tmp_OBB = p3d_get_oriented_BB_obj(rtmp->getJoint(0)->getJointStruct()->o);
                    if (GC.point_in_hexahedron(vp.at(j),tmp_OBB->UpBackLeft,tmp_OBB->UpFrontLeft,
                                         tmp_OBB->UpBackRight, tmp_OBB->UpFrontRight,
                                         tmp_OBB->BottomBackLeft, tmp_OBB->BottomFrontLeft,
                                         tmp_OBB->BottomBackRight, tmp_OBB->BottomFrontRight))
                    {
                        nbIn++;
                    }
                    p3d_destroy_oriented_BB(tmp_OBB);
                }
            }
            mainVol += 1 - (double)nbIn/(double)nb;
        }
        else
        {
            cout << "only rectangles computed" << endl;
        }
    }
    vg = targetEntity->getObjectRob()->getSupports();

    double targetVol = 0;
    for (unsigned int i = 0; i < vg.size(); i++)
    {
        if (vg.at(i)->type == "rectangle")
        {
            RectangleForm* rf = (RectangleForm*)vg.at(i);
            p3d_point a,b,c,d,e,f,g,h;
            e = rf->getA();
            f = rf->getB();
            g = rf->getC();
            h = rf->getD();
            a = e;
            a.z = a.z + clutterUpLimit;
            b = f;
            b.z = b.z + clutterUpLimit;
            c = g;
            c.z = c.z + clutterUpLimit;
            d = h;
            d.z = d.z + clutterUpLimit;

            int nb = (int)(GC.distBetweenPoints(a,b)*GC.distBetweenPoints(a,c)*GC.distBetweenPoints(a,e) *convertTo);

//            nb = 100;

            if (nb == 0)
            {
                continue;
            }


            vector<p3d_point> vp = GC.getRandomPointsInHexahedron(nb,a,b,c,d,e,f,g,h);
            int nbIn = 0;
            for (unsigned int j = 0; j < vp.size(); j++)
            {
                for (unsigned int k = 0; k < sce->getNumberOfRobots(); k++)
                {
                    Robot* rtmp = sce->getRobot(k);
                    if (rtmp->getNumberOfJoints() > 2)
                        continue;

                    p3d_oriented_BB *tmp_OBB = p3d_get_oriented_BB_obj(rtmp->getJoint(0)->getJointStruct()->o);
                    if (GC.point_in_hexahedron(vp.at(j),tmp_OBB->UpBackLeft,tmp_OBB->UpFrontLeft,
                                               tmp_OBB->UpBackRight, tmp_OBB->UpFrontRight,
                                               tmp_OBB->BottomBackLeft, tmp_OBB->BottomFrontLeft,
                                               tmp_OBB->BottomBackRight, tmp_OBB->BottomFrontRight))
                    {
                        nbIn++;
                    }
                    p3d_destroy_oriented_BB(tmp_OBB);
                }
            }
            targetVol += 1 - (double)nbIn/(double)nb;
        }
        else
        {
            cout << "only rectangles computed" << endl;
        }
    }

//    cout << "size main = " << mainVol << " size target = " << targetVol << endl;

    //



    Fact f1(FI);
    f1.setIsComputed(true);
    f1.setMainEntity(mainEntity);
    f1.setTargetEntity(targetEntity);
    f1.setValueType(0);
    f1.setBoolValue(mainVol <= targetVol);
    f1.setSpecValue(mainVol);

    WS->setFact(f1);

    Fact f2(FI);
    f2.setIsComputed(true);
    f2.setMainEntity(targetEntity);
    f2.setTargetEntity(mainEntity);
    f2.setValueType(0);
    f2.setBoolValue(mainVol > targetVol);
    f2.setSpecValue(targetVol);

    WS->setFact(f2);

    Fact f3(FI2);
    f3.setIsComputed(true);
    f3.setMainEntity(mainEntity);
    f3.setTargetEntity(targetEntity);
    f3.setValueType(0);
    f3.setBoolValue(mainVol > targetVol);
    f3.setSpecValue(mainVol);

    WS->setFact(f3);

    Fact f4(FI2);
    f4.setIsComputed(true);
    f4.setMainEntity(targetEntity);
    f4.setTargetEntity(mainEntity);
    f4.setValueType(0);
    f4.setBoolValue(mainVol <= targetVol);
    f4.setSpecValue(targetVol);

    WS->setFact(f4);

    return true;
}




//functions for each facts concerning robots
bool FactsManager::computeRobotVisibility(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i)) )
            t = computeVisibility(WS, mainEntity, sce->getRobot(i), forceRecompute);
    }
    return t;
}

bool FactsManager::computeRobotReachability(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
        {
            t = computeReachability(WS, mainEntity, sce->getRobot(i), forceRecompute);
        }
    }
    return t;
}

bool FactsManager::computeRobotIsOn(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
            t = computeIsOn(WS, mainEntity, sce->getRobot(i), forceRecompute);
    }
    return t;
}

bool FactsManager::computeRobotIsIn(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
            t = computeIsIn(WS, mainEntity, sce->getRobot(i), forceRecompute);
    }
    return t;
}

bool FactsManager::computeRobotIsNextTo(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
        {
            t = computeIsNextTo(WS, mainEntity, sce->getRobot(i), forceRecompute);
        }
    }
    return t;
}

bool FactsManager::computeRobotSize(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName())
          && !inOrigin(sce->getRobot(i))  )
        {
            t = computeSize(WS, mainEntity, sce->getRobot(i), forceRecompute);
        }
    }
    return t;
}

bool FactsManager::computeRobotClutter(WorldStateFacts* WS, Robot* mainEntity, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName())
          && !inOrigin(sce->getRobot(i))  )
        {
            t = computeClutter(WS, mainEntity, sce->getRobot(i), forceRecompute);
        }
    }
    return t;
}


// fct for each facts
bool FactsManager::computeAllVisibilities(WorldStateFacts* WS, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
            t = computeRobotVisibility(WS,sce->getRobot(i), forceRecompute);
    }
    return t;
}

bool FactsManager::computeAllReachabilities(WorldStateFacts* WS, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
        {
            t = computeRobotReachability(WS,sce->getRobot(i), forceRecompute);
        }
    }
    return t;
}

bool FactsManager::computeAllIsOn(WorldStateFacts* WS, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
            t = computeRobotIsOn(WS,sce->getRobot(i), forceRecompute);
    }
    return t;
}

bool FactsManager::computeAllIsIn(WorldStateFacts* WS, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
            t = computeRobotIsIn(WS,sce->getRobot(i), forceRecompute);
    }
    return t;
}

bool FactsManager::computeAllIsNextTo(WorldStateFacts* WS, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) 
          && !inOrigin(sce->getRobot(i))  )
            t = computeRobotIsNextTo(WS,sce->getRobot(i), forceRecompute);
    }
    return t;
}

bool FactsManager::computeAllSize(WorldStateFacts* WS, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName())
          && !inOrigin(sce->getRobot(i))  )
            t = computeRobotSize(WS,sce->getRobot(i), forceRecompute);
    }
    return t;
}


bool FactsManager::computeAllClutter(WorldStateFacts* WS, bool forceRecompute)
{
    bool t = true;
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName())
          && !inOrigin(sce->getRobot(i))  )
            t = computeRobotClutter(WS,sce->getRobot(i), forceRecompute);
    }
    return t;
}


// general fct
bool FactsManager::computeFacts(WorldStateFacts* WS, bool forceRecompute)
{
    cout << "#################################" << endl;
    cout << "begin compute facts" << endl;
    clock_t begin = clock();

    bool t = true;
    if (isFactEnabled("IsVisible"))
        t = computeAllVisibilities(WS,forceRecompute);

    if (isFactEnabled("IsReachable"))
        t = computeAllReachabilities(WS,forceRecompute);

    if (isFactEnabled("IsOn"))
        t = computeAllIsOn(WS,forceRecompute);

    if (isFactEnabled("IsIn"))
        t = computeAllIsIn(WS,forceRecompute);

    if (isFactEnabled("IsNextTo"))
        t = computeAllIsNextTo(WS,forceRecompute);

    if (isFactEnabled("Size"))
        t = computeAllSize(WS,forceRecompute);

    if (isFactEnabled("Cluttered"))
        t = computeAllClutter(WS,forceRecompute);

//    WS->printAllFacts();
//    WS->printAllComputedFacts();

    cout << "facts computed successfully within : " << (double)(clock() - begin)/CLOCKS_PER_SEC << " s" << endl;
    cout << "#################################" << endl;
    return t;

}


bool FactsManager::computeFacts(int WSid, bool forceRecompute)
{
    int id =-1;
    for (unsigned int i = 0; i < WSlist.size(); i++)
    {
        if(WSid == WSlist.at(i).first)
        {
            id = i;
            break;
        }
    }
    if (id == -1)
    {
        cout <<"lllllllllll" << endl;
        return false;
    }
    return computeFacts(WSlist.at(id).second, forceRecompute);
}

int FactsManager::computeFacts(bool forceRecompute)
{

    WorldStateFacts* WS = new WorldStateFacts(sce);
    int id = findOrAddWorldStateFacts(WS);
//    WS->initFacts(factList);
//    int id = findOrAddWorldStateFacts(WS);

    cout << "ws id = " << id << endl;
    if (computeFacts(id, forceRecompute))
        return id;
    else
        return -1;
}


bool FactsManager::computeOneFact(WorldStateFacts* WSF, string fact, Robot *mainEntity)
{

    bool t = true;
    if (fact == "IsVisible")
        t = computeRobotVisibility(WSF, mainEntity, true);

    if (fact == "IsReachable")
        t = computeRobotReachability(WSF, mainEntity,true);

    if (fact == "IsOn")
        t = computeRobotIsOn(WSF, mainEntity,true);

    if (fact == "IsIn")
        t = computeRobotIsIn(WSF, mainEntity,true);

    if (fact == "IsNextTo")
        t = computeRobotIsNextTo(WSF, mainEntity,true);

    if (fact == "Size")
        t = computeRobotSize(WSF, mainEntity,true);

    if (fact == "Cluttered")
        t = computeRobotClutter(WSF, mainEntity,true);

    return t;
}

vector<Fact> FactsManager::computeAndGetFact(string fact, Robot* mainEntity)
{
    WorldStateFacts* WS = new WorldStateFacts(sce);
    WS->initFacts(factList);
    computeOneFact(WS,fact,mainEntity);

    vector<Fact> vf;
    vf = WS->getFactFromMain(fact, mainEntity->getName());
    delete WS;
     vector<Fact> vf2;
    for (unsigned int i= 0; i < vf.size(); i++)
    {
        Fact f = vf.at(i);
        if(f.getValueType() != 0 || f.getBoolValue())
        {
            vf2.push_back(f);
        }
    }
    return vf2;

}

Fact FactsManager::computeAndGetFact(string fact, string factSubType, Robot *mainEntity, Robot* targetEntity)
{
    WorldStateFacts* WS = new WorldStateFacts(sce);
    WS->initFacts(factList);
    computeOneFact(WS,fact,mainEntity);

    vector<Fact> vf;
    vf = WS->getFactFromMain(fact, mainEntity->getName());
    delete WS;
    Fact f;
    for (unsigned int i= 0; i < vf.size(); i++)
    {
        f = vf.at(i);
        if (f.getType().factType == fact
                && f.getType().factSubType == factSubType
                && f.getTargetEntity()->getName() == targetEntity->getName())
        {
            return f;
        }
    }
    f.setMainEntity(NULL);
    f.setTargetEntity(NULL);
    return f;

}

int FactsManager::findWorldStateFacts(WorldStateFacts* WS)
{
    int id = -1;
    for (unsigned int i = 0; i < WSlist.size(); i++)
    {
        if(WSlist.at(i).second->compareWS(WS).size() == 0)
        {
            id = i;
            break;
        }
    }
    return id;
}

int FactsManager::findCurWorldStateFacts()
{
    WorldStateFacts* WS = new WorldStateFacts(sce);
    WS->initFacts(factList);
    return findWorldStateFacts(WS);
}

void FactsManager::setFactToFalse(WorldStateFacts* WS, fact_id ft, Robot* mainEntity, Robot* targetEntity)
{
    Fact f(ft);
    f.setIsComputed(true);
    f.setMainEntity(mainEntity);
    f.setTargetEntity(targetEntity);
    f.setValueType(0);
    f.setBoolValue(false);
    f.setSpecValue(-1);

    WS->setFact(f);
}

int FactsManager::findOrAddWorldStateFacts(WorldStateFacts* WS)
{
    for (unsigned int i = 0; i < WSlist.size(); i++)
    {
        if(WSlist.at(i).second->compareWS(WS).size() == 0)
        {
            delete WS;
            return i;
        }
    }

    pair<int,WorldStateFacts*> p;
    p.first = WSlistId;
    WSlistId++;
    p.second = WS;
    WSlist.push_back(p);
    WS->initFacts(factList);
    return p.first;
}

WorldStateFacts* FactsManager::getWSFromId(int WSid)
{
    for(unsigned int i = 0; i < WSlist.size(); i++)
    {
        pair<int,WorldStateFacts*> p = WSlist.at(i);
        if (p.first == WSid)
            return p.second;
    }
    return NULL;
}

bool FactsManager::cleanWSList()
{
    for (unsigned int i = 0; i < WSlist.size(); i++)
    {
//        delete WSlist.at(i).second;
    }
    WSlist.clear();
    WSlistId = 0;
    return true;
}


//########################################## class Fact ##########################################

Fact::Fact()
{

}

Fact::Fact(fact_id fact_type)
{
    fType = fact_type;
    isComputed = false;

    mainEntity = NULL;
    targetEntity = NULL;

    value_type = -1;

    boolValue = false;
    specificValue = -1;
}

bool Fact::compareFactsDescr(Fact f)
{
    if (f.fType.factType == fType.factType
            && f.fType.factSubType == fType.factSubType
            && f.fType.isMainAnAgent == fType.isMainAnAgent
            && f.mainEntity->getName() == mainEntity->getName()
            && f.targetEntity->getName() == targetEntity->getName())
    {
        return true;
    }
    return false;
}

bool Fact::compareFactsDescr(fact_id FI, Robot* main, Robot* target)
{
    if (FI.factType == fType.factType
            && FI.factSubType == fType.factSubType
            && FI.isMainAnAgent == fType.isMainAnAgent
            && main->getName() == mainEntity->getName()
            && target->getName() == targetEntity->getName())
    {
        return true;
    }
    return false;
}


//return true if equal
bool Fact::compareFactsValue(Fact f)
{
     if ( f.compareFactsDescr(fType, mainEntity, targetEntity) ){
         if (value_type)  //Case of specific value
             if (f.getSpecValue() == specificValue)
                return true;
             else
                return false;
         else  //Case of bool
             if (f.getBoolValue() == boolValue)
                return true;
             else
                return false;
     }
     else
         return false;
}

fact_id Fact::getType()
{
    return fType;
}

Robot* Fact::getMainEntity()
{
    return mainEntity;
}

Robot* Fact::getTargetEntity()
{
    return targetEntity;
}

void Fact::setMainEntity(Robot* r)
{
    mainEntity = r;
}

void Fact::setTargetEntity(Robot* r)
{
    targetEntity = r;
}

bool Fact::isItComputed()
{
    return isComputed;
}

void Fact::setIsComputed(bool t)
{
    isComputed = t;
}

int Fact::getValueType()
{
    return value_type;
}

void Fact::setValueType(int v)
{
    value_type = v;
}

bool Fact::getBoolValue()
{
    return boolValue;
}

void Fact::setBoolValue(bool v)
{
    boolValue =v;
}

double Fact::getSpecValue()
{
    return specificValue;
}

void Fact::setSpecValue(double v)
{
    specificValue = v;
}

void Fact::printFact()
{
    cout << "----------------- print fact ------------------" << endl;
    cout << "fact type: " << fType.factType << endl;
    cout << "fact sub type: " << fType.factSubType << endl;
    string str = "False";
    if (fType.isMainAnAgent)
        str = "True";
    cout << "is main entity an agent?: " << str << endl;
    isComputed?str="True":str="False";
    cout << "is it computed?: " << str << endl;
    if (mainEntity != NULL && targetEntity != NULL)
    {
        cout << "main Entity: " << mainEntity->getName() << endl;
        cout << "target Entity: " << targetEntity->getName() << endl;
    }
    //    if (value_type == 0)
    //    {
    boolValue?str="True":str="False";
    cout << "value: " <<str << endl;
    //    }
    //    else
    cout << "value: " << specificValue << endl;

    cout << "-------------- end print fact ---------------" << endl;
}

string Fact::getFactDescr()
{
    string s = fType.factType;
    s+= " ";
    s+=fType.factSubType;
    s+= " ";
    if (mainEntity != NULL && targetEntity != NULL)
    {
        s+= mainEntity->getName();
        s+= " ";
        s+= targetEntity->getName();
        s+= " ";
    }
    boolValue?s+="True":s+="False";

    return s;
}

//########################################## class WorldState ##########################################

WorldStateFacts::WorldStateFacts()
{

}


WorldStateFacts::WorldStateFacts(Scene* scen): sce(scen)
{
    for (unsigned int i = 0; i < sce->getNumberOfRobots(); i++)
    {
        if ( !inIgnoreList(sce->getRobot(i)->getName()) )
            robotsConfs.push_back(sce->getRobot(i)->getCurrentPos());
    }
}

WorldStateFacts::WorldStateFacts(std::vector<confPtr_t> confs)
{
    for (unsigned int i = 0; i < confs.size(); i++)
    {
        if ( !inIgnoreList(confs.at(i)->getRobot()->getName()) )
            robotsConfs.push_back(confs.at(i));
    }
}

bool WorldStateFacts::initFacts(vector<fact_id> factList)
{
    //initialiazing all the Facts:
    for (unsigned int i = 0; i < factList.size(); i++)
    {
        for (unsigned int j = 0; j < sce->getNumberOfRobots(); j++)
        {
            if ( inIgnoreList(sce->getRobot(j)->getName()) 
              || inOrigin(sce->getRobot(j)) )
                continue;

            for (unsigned int k = 0; k < sce->getNumberOfRobots(); k++)
            {
                if ( inIgnoreList(sce->getRobot(k)->getName()) 
                  || inOrigin(sce->getRobot(k)) )
                    continue;


                //checking if it's the same robot
                if (sce->getRobot(j)->getName() == sce->getRobot(k)->getName())
                    continue;

                //checking of the facts involving agents involve them correctly
                if (factList.at(i).isMainAnAgent
                        && sce->getRobot(k)->getHriAgent()
                        && !sce->getRobot(j)->getHriAgent())
                    continue;

                Fact F(factList.at(i));
                F.setMainEntity(sce->getRobot(j));
                F.setTargetEntity(sce->getRobot(k));
                facts.push_back(F);
            }
        }
    }
    return true;

}

void WorldStateFacts::printAllFacts()
{
    for (unsigned int i = 0; i < facts.size(); i++)
    {
        facts.at(i).printFact();
    }
}

void  WorldStateFacts::printAllComputedFacts()
{
    for (unsigned int i = 0; i < facts.size(); i++)
    {
        if (facts.at(i).isItComputed() && facts.at(i).getBoolValue())
            facts.at(i).printFact();
    }
}

void  WorldStateFacts::printFacts(vector<Fact> fp)
{
    for (unsigned int i = 0; i < fp.size(); i++)
    {
        fp.at(i).printFact();
    }
    cout<< "nb of facts = " << fp.size() << endl;
}

vector<Fact> WorldStateFacts::getFacts()
{
    return facts;
}

vector<Fact> WorldStateFacts::getMeaningfulFacts()
{
    vector<Fact> vf;
    for(unsigned int i = 0; i < facts.size(); i++)
    {
        Fact f = facts.at(i);
        if (f.isItComputed())
        {
            if(f.getValueType() == 0 && !f.getBoolValue())
            {
                continue;
            }
            if(f.getType().isMainAnAgent && !f.getMainEntity()->isAgent())
            {
                continue;
            }
            vf.push_back(f);
        }
    }
    return vf;
}


vector<Fact> WorldStateFacts::getUpdatedFacts(WorldStateFacts* WS)
{
    bool newFact;
    vector<Fact> vectorOldFacts = WS->getFacts();
    vector<Fact> vf;
    //Look for new facts
    for(unsigned int i = 0; i < facts.size(); i++)
    {
        Fact f = facts.at(i);
        if (f.isItComputed())
        {
			if(f.getValueType() == 0 && !f.getBoolValue())
            {
                continue;
            }
            if(f.getType().isMainAnAgent && !f.getMainEntity()->isAgent())
            {
                continue;
            }

            for(unsigned int j = 0; j < vectorOldFacts.size(); j++)
            {
	          newFact = true;
              Fact oldF = vectorOldFacts.at(j);

              // If the facts are different, we keep looking
              if (!f.compareFactsValue(oldF))
                continue;

              //If we found the fact in the previous WS we skip it
              else {
                newFact = false;
                break;
              }
            }
	    if (newFact)
              vf.push_back(f);
        }
    }
    //Look for old facts that disappeared
    for(unsigned int i = 0; i < vectorOldFacts.size(); i++)
    {
        Fact f = vectorOldFacts.at(i);
        if (f.isItComputed())
        {
			//Only for boolean
			if(f.getValueType() != 0)
            {
                continue;
            }
            if(f.getType().isMainAnAgent && !f.getMainEntity()->isAgent())
            {
                continue;
            }
            for(unsigned int j = 0; j < facts.size(); j++)
            {
	          newFact = true;
              Fact newF = facts.at(j);

              // If the facts are different, we keep looking
              if (!f.compareFactsValue(newF))
                continue;

              //If we found the fact in the next WS we skip it
              else {
                newFact = false;
                break;
              }
            }
            // The fact is not in the new WS, we put it with opposite value
	        if (newFact)
	        {
				f.setBoolValue(!f.getBoolValue());
				vf.push_back(f);
			}
		  }
    }
    return vf;
}


void WorldStateFacts::setFacts(vector<Fact> Facts)
{
    facts = Facts;
}

vector<Fact> WorldStateFacts::getFactFromMain(string factType, string main)
{
    vector<Fact> ff;
    for (unsigned int i = 0; i < facts.size(); i++)
    {
        Fact f = facts.at(i);
//        cout << "f1 = " << f.getType().factType << " f2 " << factType << " m1 " << main << " m2 " << f.getMainEntity()->getName() << endl;
        if (f.getType().factType == factType && main == f.getMainEntity()->getName() )
        {
            if (f.isItComputed())
            {
                ff.push_back(f);
            }
        }
    }
    return ff;
}

vector<Fact> WorldStateFacts::getFactFromTarget(string factType, string target)
{
    vector<Fact> ff;
    for (unsigned int i = 0; i < facts.size(); i++)
    {
        Fact f = facts.at(i);
//        cout << "f1 = " << f.getType().factType << " f2 " << factType << " m1 " << main << " m2 " << f.getMainEntity()->getName() << endl;
        if (f.getType().factType == factType && target == f.getTargetEntity()->getName() )
        {
            if (f.isItComputed())
            {
                ff.push_back(f);
            }
        }
    }
    return ff;
}

Fact WorldStateFacts::getFact(fact_id Fact_Id, Robot* main, Robot* target)
{
    Fact f;
    for (unsigned int i = 0; i < facts.size(); i++)
    {
        f = facts.at(i);
        if (f.compareFactsDescr(Fact_Id, main, target))
        {
            return f;
        }
    }
    f.setMainEntity(NULL);
    f.setTargetEntity(NULL);
    return f;
}


Fact WorldStateFacts::getFact(string factType, string main, string target)
{
    Fact f;
    for (unsigned int i = 0; i < facts.size(); i++)
    {
        f = facts.at(i);
        if (f.getType().factType == factType && main == f.getMainEntity()->getName() && target == f.getTargetEntity()->getName() )
        {
            return f;
        }
    }
    return f;
}

void WorldStateFacts::setFact(Fact fact)
{
    int id = -1;
    for (unsigned int i = 0; i < facts.size(); i++)
    {
        if (facts.at(i).compareFactsDescr(fact.getType(), fact.getMainEntity(), fact.getTargetEntity()))
        {
            id =i;
            break;
        }
    }
    if (id !=-1)
    {
        facts.at(id).setIsComputed(fact.isItComputed());
        facts.at(id).setValueType(fact.getValueType());
        facts.at(id).setBoolValue(fact.getBoolValue());
        facts.at(id).setSpecValue(fact.getSpecValue());
    }
    else
    {
        facts.push_back(fact);
    }
}

vector<confPtr_t> WorldStateFacts::getRobotsConfs()
{
    return robotsConfs;
}

confPtr_t WorldStateFacts::getRobotConf(Robot* robot)
{
    for (unsigned int i = 0; i < robotsConfs.size(); i++)
    {
        if ( inIgnoreList(sce->getRobot(i)->getName()) )
            continue;
        if (robotsConfs.at(i)->getRobot()->getName() == robot->getName())
        {
            return robotsConfs.at(i);
        }
    }
    confPtr_t q;
    return q;

}

confPtr_t WorldStateFacts::getRobotConfById(int robot_id)
{

    Robot* robot = sce->getRobot(robot_id);
    for (unsigned int i = 0; i < robotsConfs.size(); i++)
    {
        if ( inIgnoreList(sce->getRobot(i)->getName()) )
            continue;
        if (robotsConfs.at(i)->getRobot()->getName() == robot->getName())
        {
            return robotsConfs.at(i);
        }
    }
    confPtr_t q;
    return q;
}


vector<Robot*> WorldStateFacts::compareWS(WorldStateFacts* WS)
{
    vector<confPtr_t > outsideConfs = WS->getRobotsConfs();

    vector<Robot*> v;

    if(outsideConfs.size() != robotsConfs.size())
    {
        cout<< "ERROR: not same nb of robots in two WS" <<endl;
        return v;
    }

    for (unsigned int i = 0; i < robotsConfs.size(); i++)
    {
        confPtr_t c1 = robotsConfs.at(i);
        for (unsigned int j = 0; j < robotsConfs.size(); j++)
        {
            confPtr_t c2 = outsideConfs.at(j);
            if ( inIgnoreList(c1->getRobot()->getName()) 
              || inIgnoreList(c2->getRobot()->getName()) )
                continue;
            if (c2->getRobot()->getName() == c1->getRobot()->getName())
            {
                if(!c1->equal(*c2))
                {
                    v.push_back(c1->getRobot());
                }
            }
        }
    }
    return v;
}

void WorldStateFacts::setConfiguration(confPtr_t conf)
{
    for (unsigned int i = 0; i < robotsConfs.size(); i++)
    {
        if (robotsConfs.at(i)->getRobot()->getName() == conf->getRobot()->getName())
        {
            robotsConfs.at(i) = conf;
            break;
        }
    }
}
