#include "grid.hpp"

#include "include/P3d-pkg.h"
#include "Collision-pkg.h"
#include "API/Grids/gridsAPI.hpp"
using namespace std;

extern std::vector<p3d_point> currNavTraj;


GTPGrid::GTPGrid()
{
}

GTPCell::GTPCell()
{
}



GTP2DGrid::GTP2DGrid() :
        API::TwoDGrid()
{
}

GTP2DGrid::GTP2DGrid(double pace, vector<double> envSize) :
        API::TwoDGrid(pace,envSize)
{
    createAllCells();
    cout << "Number total of cells = " << _nbCellsX*_nbCellsY << endl;
    initCylinders();
}

bool GTP2DGrid::initCylinders()
{
    _humCyl = NULL;
    _robotCyl = NULL;
    for (unsigned int i=0; i<global_Project->getActiveScene()->getNumberOfRobots(); i++)
    {
        Robot* r = global_Project->getActiveScene()->getRobot(i);
        if (r->getName() == "HUMCYLINDER")
        {
            _humCyl = r;
        }
        else if (r->getName() == "PR_2CYLINDER")
        {
            _robotCyl = r;
        }
    }
    if (!_humCyl || !_robotCyl)
    {
        cout << "no cylinder found" << endl;
        return false;
    }
    return true;
}

void GTP2DGrid::addAgentAndComputeNav(Robot* robot)
{
    _robotList.push_back(robot);
    propagateDist(robot);
}



void GTP2DGrid::propagateDist(Robot* robot)
{
    initAllCells();
    Robot* cyl;
    if (!robot->isAgent())
    {
        return;
    }
    if (robot->getHriAgent()->is_human != 0)
    {
        cyl = _humCyl;
    }
    else
    {
        cyl = _robotCyl;
    }

    ENV.setBool(Env::printMightabilities, true);
    vector< GTP2DCell*> cellList;
    confPtr_t q_cur = robot->getCurrentPos();
    Eigen::Vector2d pos;
    pos[0] = (*q_cur)[6];
    pos[1] = (*q_cur)[7];

    GTP2DCell* init = dynamic_cast<GTP2DCell*>(getCell(pos));

    cellList.push_back(init);
    init->setDist(robot,0);
    init->setPrev(robot, NULL);
    init->setState(Closed);
    p3d_col_deactivate_robot_robot(robot->getRobotStruct(), cyl->getRobotStruct());
    while (cellList.size() > 0)
    {
//        GTP2DCell* last = cellList.at(cellList.size()-1);
//        cellList.pop_back();


        GTP2DCell* last = cellList.at(0);
        cellList.erase(cellList.begin());


        confPtr_t q = cyl->getCurrentPos();
        (*q)[6] = last->getCenter()[0];
        (*q)[7] = last->getCenter()[1];
        cyl->setAndUpdate(*q);

//        cout << "last = " << last->getCoord()[0] << " , " << last->getCoord()[1] << endl;
//        p3d_print_col_pair();



//        int nb_cols = pqp_robot_environment_collision_test(robot->getRobotStruct());
//        if (nb_cols > 0 )
//            continue;
//        for (int i=0; i<global_Project->getActiveScene()->getNumberOfRobots(); i++)
//        {
//            Robot* r = global_Project->getActiveScene()->getRobot(i);
//            if (r->getName() == robot->getName() || r->getName() == cyl->getName())
//                continue;
//            nb_cols = pqp_robot_robot_collision_test(r->getRobotStruct(), cyl->getRobotStruct());
//            if (nb_cols > 0 )
//                break;
//        }


//        if (nb_cols > 0 )
//            continue;

        //        if(!cyl->isInCollisionWithOthersAndEnv())
        if (p3d_col_test_robot(cyl->getRobotStruct(),0) == 0) //no col
        {
            last->addToNavigability(robot);

            vector<GTP2DCell*> neighbors = last->getNeighboors();
            for (unsigned int i = 0 ; i< neighbors.size() ; i++)
            {
//                cout << "neighbor " << neighbors.at(i)->getCoord()[0] << " , " << neighbors.at(i)->getCoord()[1] << endl;
                if (neighbors.at(i)->getState() == NotTested)
                {
                    neighbors.at(i)->setState(Open);
                    cellList.push_back(neighbors.at(i));

                    neighbors.at(i)->setDist(robot, last->getDist(robot) + neighbors.at(i)->computeDistTo(last));
                    neighbors.at(i)->setPrev(robot, last);
                }
//                else if (neighbors.at(i)->getState() == Closed || neighbors.at(i)->getState() == Open)
//                {
//                    double distFromNeighbor = neighbors.at(i)->getDist(robot) + neighbors.at(i)->computeDistTo(last);
//                    if (last->getDist(robot) > distFromNeighbor)
//                    {
//                        last->setDist(robot, distFromNeighbor);
//                        last->setPrev(robot, neighbors.at(i));
//                    }
//                }
//                cout << "distance computedd = " << last->getDist(robot) << endl;
            }


        }
        last->setState(Closed);

        if (init->getPrev(robot))
        {
            cout << "------------------------------------------------" <<endl;
         }

    }
    p3d_col_activate_robot_robot(robot->getRobotStruct(), cyl->getRobotStruct());

    API_activeGrid = this;
    ENV.setBool(Env::drawGrid, false);

    confPtr_t q = cyl->getCurrentPos();
    (*q)[6] = 0;
    (*q)[7] = 0;
    cyl->setAndUpdate(*q);





}


void GTP2DGrid::initAllCells()
{
    for (unsigned int x =0;x<_nbCellsX;++x)
    {
        for (unsigned int y =0;y<_nbCellsY;++y)
        {
            GTP2DCell* cell = dynamic_cast<GTP2DCell*>(getCell(x,y));
            cell->initialize();
        }
    }
}

vector<GTP2DCell*> GTP2DGrid::getCellsWithinReach(double xCenter, double yCenter, double reach, Robot* agent)
{
    vector<GTP2DCell*> res;
    Eigen::Vector2d pos;
    pos[0] = xCenter;
    pos[1] = yCenter;

    GTP2DCell* init = dynamic_cast<GTP2DCell*>(getCell(pos));
    for (unsigned int x =0;x<_nbCellsX;++x)
    {
        for (unsigned int y =0;y<_nbCellsY;++y)
        {
            GTP2DCell* cell = dynamic_cast<GTP2DCell*>(getCell(x,y));
            if (!cell->getNavigability(agent))
                continue;
            double dist = sqrt( pow(init->getCenter()[0] - cell->getCenter()[0],2)  + pow(init->getCenter()[1] - cell->getCenter()[1],2));
            if(dist > reach)
                continue;
            res.push_back(cell);
            cell->setDrawn(true);

        }
    }
    return res;

}




void GTP2DGrid::draw()
{



//    double colorvector[4];

//    colorvector[0] = 1.0;       //red
//    colorvector[1] = 0.5;       //green
//    colorvector[2] = 0.0;       //blue
//    colorvector[3] = 0.05;       //transparency


    glEnable(GL_BLEND);
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    //    glBlendFunc(GL_SRC_ALPHA, GL_ONE);

    glDisable(GL_LIGHTING);
    glDisable(GL_LIGHT0);

    glEnable(GL_CULL_FACE);
    glBegin(GL_QUADS);


    double color[3];

    double cMax= 0, cMin = 1000;

    for (unsigned int x =0;x<_nbCellsX;++x)
    {
        for (unsigned int y =0;y<_nbCellsY;++y)
        {
            GTP2DCell* Cell = dynamic_cast<GTP2DCell*>(getCell(x,y));

            if (Cell->toBeDrawn())
            {
                if (cMax < Cell->Ecolor)
                    cMax = Cell->Ecolor;
                else if (cMin > Cell->Ecolor)
                    cMin = Cell->Ecolor;
            }
        }
    }


    for (unsigned int x =0;x<_nbCellsX;++x)
    {
        for (unsigned int y =0;y<_nbCellsY;++y)
        {

            GTP2DCell* Cell = dynamic_cast<GTP2DCell*>(getCell(x,y));


            if (Cell->toBeDrawn())
            {
                double colorRation = (Cell->Ecolor - cMin)/(cMax-cMin);
                Vector2d center = Cell->getCenter();

                //            double colorRation = (((double)x*(double)_nbCellsY)+(double)y)/(nbCells);
                //            cout << " X = "  << _nbCellsX << " , Y = "  << _nbCellsY << endl;
                //            cout << "ColorRation[" << x*_nbCellsY+y << "]  =  "  << colorRation << endl;

                //            colorRation = colorRation*ENV.getDouble(Env::colorThreshold2)*1000;
                GroundColorMixGreenToRed(color,colorRation);
                //                colorRation = depth;

                double depth;
                //                    depth =colorRation;
                depth =0.1;
                glColor3d(color[0],color[1],color[2]);
                glVertex3d( (double)(center[0] - _cellSize[0]/2) + 0.01 , (double)(center[1] - _cellSize[1]/2) + 0.01, depth);//depth );
                glVertex3d( (double)(center[0] + _cellSize[0]/2) - 0.01 , (double)(center[1] - _cellSize[1]/2) + 0.01, depth);//depth );
                glVertex3d( (double)(center[0] + _cellSize[0]/2) - 0.01 , (double)(center[1] + _cellSize[1]/2) - 0.01, depth);//depth );
                glVertex3d( (double)(center[0] - _cellSize[0]/2) + 0.01 , (double)(center[1] + _cellSize[1]/2) - 0.01, depth);//depth );

            }
        }
    }



    glEnd();

    glDisable(GL_CULL_FACE);
    glDisable(GL_BLEND);

}

/*!
 * \brief Creates All Cells
 *
 * \param vector envSize XMin Xmax YMin YMax
 */
void GTP2DGrid::createAllCells()
{
  unsigned int nbCells = _nbCellsX * _nbCellsY ;

  _cells.resize(nbCells);

  unsigned int x=0;
  unsigned int y=0;

  for(unsigned int i = 0; i<nbCells; i++)
  {
    // cout << "("<< x << "," << y << ")" << endl;

    GTP2DCell* ptrCell = createNewCell2(i,x,y);
    _cells[i] = ptrCell;

    x++;
    if( x >= _nbCellsX )
    {
      y++;
      x=0;
      if( y >= _nbCellsY )
      {
        //cout << "TwoDGrid : Error Size of TwoDGrid " << endl;
        return;
      }
    }
  }
  //    cout << "Finished"<< endl;
}

GTP2DCell* GTP2DGrid::createNewCell2(unsigned int index,unsigned  int x,unsigned  int y )
{
  if (index == 0)
  {
      Vector2i coord;
      coord[0] = 0;
      coord[1] = 0;
    return new GTP2DCell( 0, coord,  _originCorner , this );
  }

  Vector2i coord;
  coord[0] = x;
  coord[1] = y;
  GTP2DCell* newCell = new GTP2DCell( index, coord, computeCellCorner(x,y) , this );
//  Vector2d corner = newCell->getCorner();
  //    cout << " = (" << corner[0] <<"," << corner[1] << ")" << endl;
  return newCell;
}

vector<GTP2DCell*> GTP2DGrid::getPath(Robot* r, GTP2DCell* targetCell)
{
    vector<GTP2DCell*> vg;

    //A* for nav
    vector<GTP2DCell*> closedSet;
    vector<GTP2DCell*> openSet;

    confPtr_t q = r->getCurrentPos();
    Vector2d v2; v2[0] = (*q)[6]; v2[1] = (*q)[7];
    GTP2DCell* start = dynamic_cast<GTP2DCell*>(getCell(v2));

    openSet.push_back(start);
    start->setDist(r,0);
    start->setScore(start->getScore()+ start->computeDistTo(targetCell));


//    std::remove()
    nbExploredCells = 1;
    while (openSet.size() > 0)
    {
        double fscore = 1000;
        int id;
        for (unsigned int i = 0; i< openSet.size(); i++)
        {
            if (fscore > openSet.at(i)->getScore())
            {
                id = i;
                fscore = openSet.at(i)->getScore();
            }
        }
        GTP2DCell* curr = openSet.at(id);
        if (curr->getIndex() == targetCell->getIndex())
            break;
        openSet.erase(openSet.begin()+id);
        nbExploredCells++;

        closedSet.push_back(curr);
        vector<GTP2DCell*> neighbs = curr->getNeighboors();
        for (unsigned int i = 0; i < neighbs.size(); i++)
        {

            if (!neighbs.at(i)->getNavigability(r))
                continue;
            bool isClosed = false;
            for (unsigned int j = 0; j < closedSet.size(); j++)
            {
                if (closedSet.at(j)->getIndex() == neighbs.at(i)->getIndex())
                {
                    isClosed = true;
                    break;
                }
            }
            if (isClosed)
                continue;

            double t_dist = curr->getDist(r) + curr->computeDistTo(neighbs.at(i));
            bool isOpen = false;
            for (unsigned int j = 0; j < openSet.size(); j++)
            {
                if (openSet.at(j)->getIndex() == neighbs.at(i)->getIndex())
                {
                    isOpen = true;
                    break;
                }
            }

            if (!isOpen || t_dist < neighbs.at(i)->getDist(r) )
            {
                neighbs.at(i)->setPrev(r,curr);
                neighbs.at(i)->setDist(r,t_dist);
                neighbs.at(i)->setScore(t_dist+neighbs.at(i)->computeDistTo(targetCell));
                if (!isOpen)
                    openSet.push_back(neighbs.at(i));
            }
        }
    }


    GTP2DCell* tmp = targetCell;
    vg.push_back(tmp);
    while (tmp->getPrev(r))
    {
        tmp= tmp->getPrev(r);
        vg.push_back(tmp);
    }

    return vg;
}

vector<GTP2DCell*> GTP2DGrid::getHumanAwarePath(Robot* r, GTP2DCell* targetCell, double distWeight,
                                                double Hweight, double angleWeight, double distHWeight, double distWeight2)
{
    vector<GTP2DCell*> vg;

    currNavTraj.clear();
    //A* for nav
    vector<GTP2DCell*> closedSet;
    vector<GTP2DCell*> openSet;

    confPtr_t q = r->getCurrentPos();
    Vector2d v2; v2[0] = (*q)[6]; v2[1] = (*q)[7];
    GTP2DCell* start = dynamic_cast<GTP2DCell*>(getCell(v2));

    openSet.push_back(start);
    start->setDist(r,0);
    start->setScore(start->getScore()+ start->computeDistTo(targetCell));

    vector<p3d_point> vHP;
    for (unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
    {
        if (global_Project->getActiveScene()->getRobot(i)->isAgent())
        {
            if (global_Project->getActiveScene()->getRobot(i)->getHriAgent()->is_human == 1)
            {
                p3d_point HP;
                confPtr_t q = global_Project->getActiveScene()->getRobot(i)->getCurrentPos();
                HP.x = (*q)[6];
                HP.y = (*q)[7];
                HP.z = (*q)[11];
//                cout << "HP ( " << HP.x << " , " << HP.y << " , " << HP.z << " )" << endl;
                vHP.push_back(HP);
            }
        }
    }

    GeometricComputations GC = GeometricComputations();



//    std::remove()
    while (openSet.size() > 0)
    {
        double fscore = 1000;
        int id;
        for (unsigned int i = 0; i< openSet.size(); i++)
        {
            if (fscore > openSet.at(i)->getScore())
            {
                id = i;
                fscore = openSet.at(i)->getScore();
            }
        }
        GTP2DCell* curr = openSet.at(id);

        if (ENV.getBool(Env::drawExploration))
        {
            currNavTraj.clear();
            p3d_point ew;
            ew.x = curr->getCenter()[0];
            ew.y = curr->getCenter()[1];
            ew.z = 0;
            currNavTraj.push_back(ew);
            GTP2DCell* tmpTrajCell = curr;

            while (tmpTrajCell->getPrev(r))
            {
                tmpTrajCell= tmpTrajCell->getPrev(r);
                p3d_point ew;
                ew.x = tmpTrajCell->getCenter()[0];
                ew.y = tmpTrajCell->getCenter()[1];
                ew.z = 0;
                currNavTraj.push_back(ew);
            }
//            g3d_draw_allwin_active();
        }

        if (curr->getIndex() == targetCell->getIndex())
            break;
        openSet.erase(openSet.begin()+id);

        closedSet.push_back(curr);
        vector<GTP2DCell*> neighbs = curr->getNeighboors();
        for (unsigned int i = 0; i < neighbs.size(); i++)
        {

            if (!neighbs.at(i)->getNavigability(r))
                continue;
            bool isClosed = false;
            for (unsigned int j = 0; j < closedSet.size(); j++)
            {
                if (closedSet.at(j)->getIndex() == neighbs.at(i)->getIndex())
                {
                    isClosed = true;
                    break;
                }
            }
            if (isClosed)
                continue;

            vector<p3d_point> vp;
            p3d_point po;
            po.x = neighbs.at(i)->getCenter()[0];
            po.y = neighbs.at(i)->getCenter()[1];
            po.z = 0;
            vp.push_back(po);

            double t_dist = curr->getDist(r) + ((curr->computeDistTo(neighbs.at(i))*distWeight
                                                 + GC.computeCost(vHP,vp,angleWeight,distHWeight,distWeight2)*Hweight)
                                                /(distWeight+Hweight));
//            neighbs.at(i)
            bool isOpen = false;
            for (unsigned int j = 0; j < openSet.size(); j++)
            {
                if (openSet.at(j)->getIndex() == neighbs.at(i)->getIndex())
                {
                    isOpen = true;
                    break;
                }
            }

            if (!isOpen || t_dist < neighbs.at(i)->getDist(r) )
            {
                neighbs.at(i)->setPrev(r,curr);
                neighbs.at(i)->setDist(r,t_dist);
                neighbs.at(i)->setScore(t_dist+neighbs.at(i)->computeDistTo(targetCell));
                if (!isOpen)
                    openSet.push_back(neighbs.at(i));
            }
        }
    }



    GTP2DCell* tmp = targetCell;
    vg.push_back(tmp);
    while (tmp->getPrev(r))
    {
        tmp= tmp->getPrev(r);
        vg.push_back(tmp);
    }

    return vg;
}

vector<GTP2DCell*> GTP2DGrid::getCellsArroundSupport(Robot* agent, Robot* support)
{
    vector<GTP2DCell*> tmpCells;


    Robot* cyl;
    if (!agent->isAgent())
    {
        return tmpCells;
    }
    if (agent->getHriAgent()->is_human != 0)
    {
        cyl = _humCyl;
    }
    else
    {
        cyl = _robotCyl;
    }


    configPt q_curr = cyl->getCurrentPos()->getConfigStruct();
    configPt q = p3d_copy_config(cyl->getRobotStruct(), q_curr);


    for (unsigned int x =0;x<_nbCellsX;++x)
    {
        for (unsigned int y =0;y<_nbCellsY;++y)
        {
            GTP2DCell* cell = dynamic_cast<GTP2DCell*>(getCell(x,y));
            q[6] = cell->getCenter()[0];
            q[7] = cell->getCenter()[1];
            q[8] = 0;

            p3d_set_and_update_this_robot_conf(cyl->getRobotStruct(),q);

            if (p3d_col_test_robot(cyl->getRobotStruct(), 0)  != 1 ) //no col
                cell->setIsInCol(false);
            else
            {
                p3d_col_activate_robot_robot(cyl->getRobotStruct(), support->getRobotStruct());
                cell->setIsInCol(true);
                if (pqp_robot_robot_collision_test(cyl->getRobotStruct(), support->getRobotStruct()) == 1) //collision
                {
                    vector<GTP2DCell*> neighbors = cell->getNeighboors();
                    for (unsigned int i = 0; i < neighbors.size(); i++)
                    {
                        tmpCells.push_back(neighbors.at(i));
                    }
                }
            }
        }
    }


    vector<GTP2DCell*> res;
    for (unsigned int j = 0; j < tmpCells.size(); j++)
    {
        if(!tmpCells.at(j)->isInCol() && tmpCells.at(j)->isReachable(agent))
        {
            res.push_back(tmpCells.at(j));
        }
    }

    vector<GTP2DCell*> res2;
    for (unsigned int j = 0; j < res.size(); j++)
    {
        bool test = false;
        for(unsigned int k = 0; k < res2.size(); k++)
        {
            if (res.at(j)->getIndex() == res2.at(k)->getIndex())
            {
                test = true;
                break;
            }
        }
        if (!test)
            res2.push_back(res.at(j));
    }


    return res2;
}

vector<GTP2DCell*> GTP2DGrid::getCellsArroundObject(Robot* agent, Robot* manipulable)
{
    vector<GTP2DCell*> tmpCells;


    Robot* cyl;
    if (!agent->isAgent())
    {
        return tmpCells;
    }
    if (agent->getHriAgent()->is_human != 0)
    {
        cyl = _humCyl;
    }
    else
    {
        cyl = _robotCyl;
    }


    configPt q_curr = cyl->getCurrentPos()->getConfigStruct();
    configPt q = p3d_copy_config(cyl->getRobotStruct(), q_curr);


    GeometricComputations GC = GeometricComputations();

    double agReach = agent->getHriAgent()->manip->reach_tolerance - 0.1;

    p3d_oriented_BB *obj_OBB = p3d_get_oriented_BB_obj(manipulable->getJoint(0)->getJointStruct()->o);

    p3d_point pr;
    pr.x = (obj_OBB->UpBackLeft.x + obj_OBB->BottomFrontRight.x)/2;
    pr.y = (obj_OBB->UpBackLeft.y + obj_OBB->BottomFrontRight.y)/2;
    pr.z = (obj_OBB->UpBackLeft.z + obj_OBB->BottomFrontRight.z)/2;

    p3d_destroy_oriented_BB(obj_OBB);

    for (unsigned int x =0;x<_nbCellsX;++x)
    {
        for (unsigned int y =0;y<_nbCellsY;++y)
        {
            GTP2DCell* cell = dynamic_cast<GTP2DCell*>(getCell(x,y));
            q[6] = cell->getCenter()[0];
            q[7] = cell->getCenter()[1];

            p3d_set_and_update_this_robot_conf(cyl->getRobotStruct(),q);


            if (p3d_col_test_robot(cyl->getRobotStruct(), 0)  != 1 ) //no col
                cell->setIsInCol(false);
            else
                cell->setIsInCol(true);



            p3d_point cellCenter;
            cellCenter.x = cell->getCenter()[0];
            cellCenter.y = cell->getCenter()[1];
            cellCenter.z = 0;


            if (GC.distBetweenPoints2D(pr,cellCenter)< agReach && cell->isReachable(agent))
            {
                tmpCells.push_back(cell);
            }
        }
    }


    vector<GTP2DCell*> res;
    for (unsigned int j = 0; j < tmpCells.size(); j++)
    {
        if(!tmpCells.at(j)->isInCol())
        {
            res.push_back(tmpCells.at(j));
        }
    }

    vector<GTP2DCell*> res2;
    for (unsigned int j = 0; j < res.size(); j++)
    {
        bool test = false;
        for(unsigned int k = 0; k < res2.size(); k++)
        {
            if (res.at(j)->getIndex() == res2.at(k)->getIndex())
            {
                test = true;
                break;
            }
        }
        if (!test)
            res2.push_back(res.at(j));
    }


    return res2;
}

vector<GTP2DCell*> GTP2DGrid::getCellsArroundAgent(Robot* agent, Robot* targetAgent)
{
    vector<GTP2DCell*> tmpCells;


    Robot* cyl;
    if (!agent->isAgent() || !targetAgent->isAgent())
    {
        return tmpCells;
    }
    if (agent->getHriAgent()->is_human != 0)
    {
        cyl = _humCyl;
    }
    else
    {
        cyl = _robotCyl;
    }



    configPt q_curr = cyl->getCurrentPos()->getConfigStruct();
    configPt q = p3d_copy_config(cyl->getRobotStruct(), q_curr);


    GeometricComputations GC = GeometricComputations();

    double agReach = agent->getHriAgent()->manip->reach_tolerance + targetAgent->getHriAgent()->manip->reach_tolerance - 0.35;

    confPtr_t q_tar = targetAgent->getCurrentPos();

    p3d_point pr;
    pr.x = (*q_tar)[6];
    pr.y = (*q_tar)[7];
    pr.z = (*q_tar)[8];


    for (unsigned int x =0;x<_nbCellsX;++x)
    {
        for (unsigned int y =0;y<_nbCellsY;++y)
        {
            GTP2DCell* cell = dynamic_cast<GTP2DCell*>(getCell(x,y));
            q[6] = cell->getCenter()[0];
            q[7] = cell->getCenter()[1];

            p3d_set_and_update_this_robot_conf(cyl->getRobotStruct(),q);


            if (p3d_col_test_robot(cyl->getRobotStruct(), 0)  != 1 ) //no col
                cell->setIsInCol(false);
            else
                cell->setIsInCol(true);

            p3d_point cellCenter;
            cellCenter.x = cell->getCenter()[0];
            cellCenter.y = cell->getCenter()[1];
            cellCenter.z = 0;

            if (GC.distBetweenPoints2D(pr,cellCenter)< agReach && cell->isReachable(agent))
            {
                tmpCells.push_back(cell);
            }
        }
    }


    vector<GTP2DCell*> res;
    for (unsigned int j = 0; j < tmpCells.size(); j++)
    {
        if(!tmpCells.at(j)->isInCol())
        {
            res.push_back(tmpCells.at(j));
        }
    }


    vector<GTP2DCell*> res2;
    for (unsigned int j = 0; j < res.size(); j++)
    {
        bool test = false;
        for(unsigned int k = 0; k < res2.size(); k++)
        {
            if (res.at(j)->getIndex() == res2.at(k)->getIndex())
            {
                test = true;
                break;
            }
        }
        if (!test)
            res2.push_back(res.at(j));
    }

    return res2;
}






void GTP2DGrid::setCurrentRobot(Robot* rob)
{
    for (unsigned int x =0;x<_nbCellsX;++x)
    {
        for (unsigned int y =0;y<_nbCellsY;++y)
        {
            GTP2DCell* cell = dynamic_cast<GTP2DCell*>(getCell(x,y));
            cell->setCurrentRobot(rob);
        }
    }
}


GTP2DCell::GTP2DCell():
    _state(NotTested)
{
    _collisionCheck = false;
    _currentRobot = NULL;
}

GTP2DCell::GTP2DCell(int i, Vector2i coord, Vector2d corner, GTP2DGrid* grid) :
        API::TwoDCell(i,corner,grid),
        _Coord(coord),
        _state(NotTested),
        _toBeDrawn(false)
{
    _currentRobot = NULL;
}

vector<GTP2DCell*> GTP2DCell::getNeighboors()
{
    std::vector<GTP2DCell*> neighbors;
    Eigen::Vector2i coord = getCoord();
    double nbCellX = dynamic_cast<GTP2DGrid*>(_grid)->getNbCellX();
    double nbCellY = dynamic_cast<GTP2DGrid*>(_grid)->getNbCellY();

//    EnvCell* cell =dynamic_cast<EnvCell*>(dynamic_cast<EnvGrid*>(_grid)->getCell(coord[0]+1,coord[1]));
    for (int i = -1; i <= 1; i++)
    {
        for (int j = -1; j <= 1; j++)
        {
            if (i != 0 || j != 0)
            {

                if (coord[0]+i < nbCellX && coord[0]+i >= 0 && coord[1]+j < nbCellY && coord[1]+j >= 0)
                {
                    GTP2DCell* cell = dynamic_cast<GTP2DCell*>(dynamic_cast<GTP2DGrid*>(_grid)->getCell(coord[0]+i,coord[1]+j));
                    neighbors.push_back(cell);
                }
            }
        }
    }

    return neighbors;
}

void GTP2DCell::setState(cellState s)
{
    _state = s;
}

void GTP2DCell::initialize()
{
    _state = NotTested;
}

void GTP2DCell::addToNavigability(Robot* robot)
{
    pair<Robot*,bool> p;
    p.first = robot;
    p.second = true;
    navigability.push_back(p);
}

double GTP2DCell::dist(GTP2DCell* c)
{
    return sqrt( pow(getCenter()[0] - c->getCenter()[0],2)  + pow(getCenter()[1] - c->getCenter()[1],2));
}
void GTP2DCell::print()
{
    cout << "cell with id = " << getIndex() << " coord = ( " << getCoord()[0] << " , " <<
            getCoord()[1] << " ) and pos = ( " << getCenter()[0] << " , " << getCenter()[1] << ")" << endl;
}


bool GTP2DCell::getNavigability(Robot* r)
{
    for (unsigned int i = 0; i < navigability.size(); i++)
    {
        if (navigability.at(i).first->getName() == r->getName())
        {
            return navigability.at(i).second;
        }
    }
    return false;
}

bool GTP2DCell::toBeDrawn()
{

    return _toBeDrawn;
    for (unsigned int i = 0 ; i < navigability.size(); i++)
    {
        if (navigability.at(i).second)
            return true;
    }
    return false;
}

void GTP2DCell::setPrev(Robot* r, GTP2DCell* c)
{
    previous_to_init[r] = c;
}

GTP2DCell* GTP2DCell::getPrev(Robot* r)
{
    return previous_to_init[r];
}

void GTP2DCell::setDist(Robot* r, double d)
{
    distances_to_init[r] = d;
}

double GTP2DCell::getDist(Robot* r)
{
    return distances_to_init[r];
}

bool GTP2DCell::isReachable(Robot* r)
{
    for(map<Robot*,double>::iterator it = distances_to_init.begin(); it != distances_to_init.end(); ++it) {
        if (it->first->getName() == r->getName())
            return true;
    }
    return false;
}

double GTP2DCell::computeDistTo(GTP2DCell* c)
{

//    cout  << "cell 1 = ( " << getCoord()[0] << " , " << getCoord()[1] << ") cell 2 = (" << c->getCoord()[0] << " , " << c->getCoord()[1] << ")" << endl;
    double dist = sqrt(pow(getCenter()[0] - c->getCenter()[0],2) + pow(getCenter()[1] - c->getCenter()[1],2));
//    cout << "dist between cells = " << dist << endl ;

    return dist;
}


double GTP2DCell::getDistCurrentRobot()
{
    return distances_to_init.at(_currentRobot);
}
