
#include "hatpConnection.hpp"
#include <jsoncpp/json/json.h>
#include <iostream>
#ifdef HATP_CONNECTION
#include <msgClient.hh>
#endif


using namespace std;

HATPConnection::HATPConnection()
{
}


//! @ingroup KINECT
//! Computes a condfiguration
//! from a set of points
configPt HATPConnection::getConfFromJointsPos( p3d_rob* robot, HumanJoints& data )
{
    //p3d_rob* m_human = robot;

    // Set pelvis
    p3d_jnt* joint = p3d_get_robot_jnt_by_name(robot, (char*) "Pelvis");
    int index_dof = joint->index_dof;

    configPt q = p3d_get_robot_config(robot);

    p3d_vector3 pos;

    if( data.TORSO.confidence > 0 && data.HIP_LEFT.confidence > 0 && data.HIP_RIGHT.confidence > 0 )
    {
        q[index_dof+0] = data.TORSO.pos[0];
        q[index_dof+1] = data.TORSO.pos[1];
        q[index_dof+2] = data.TORSO.pos[2] - 0.20 ; // Hack 1 meter

        // calcul de l'orientation du pelvis
        p3d_vector3 sum,midP;
        p3d_vectAdd( data.HIP_LEFT.pos, data.HIP_RIGHT.pos , sum );
        p3d_vectScale(  sum , midP , 0.5 );

        q[index_dof+5] = atan2( data.HIP_LEFT.pos[1]-midP[1] , data.HIP_LEFT.pos[0]-midP[0]  );
        q[index_dof+5] += -M_PI/2; // Hack +  Pi / 2
    }

    //--------------------------------------------------------------
    p3d_matrix4 Tinv,Tpelv;
    p3d_vector3 shoulder;
    p3d_vector3 Xaxis = { 1 , 0 , 0 };
    p3d_matrix4 TrotX,TrotXtmp;
    p3d_vector3 Yaxis = { 0 , 1 , 0 };
    p3d_matrix4 TrotY,TrotYtmp;

    if( data.NECK.confidence > 0 && data.SHOULDER_LEFT.confidence > 0 )
    {
        p3d_set_and_update_this_robot_conf( robot, q );
        joint = p3d_get_robot_jnt_by_name(robot, (char*) "Pelvis");


        p3d_mat4Copy( joint->abs_pos , Tpelv );
        //p3d_mat4Copy( joint->abs_pos , m_absPos );
        p3d_matInvertXform( Tpelv , Tinv  );
        p3d_xformPoint( Tinv , data.NECK.pos , pos );

        double TorsoX = atan2( -pos[1] , pos[2] );
        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "TorsoX")->index_dof;
        q[index_dof] = TorsoX;


        p3d_mat4TransRot( TrotX , 0 , 0 , 0 , Xaxis , TorsoX  );
        p3d_matMultXform( Tpelv , TrotX , TrotXtmp );
        p3d_matInvertXform( TrotXtmp , Tinv  );
        p3d_xformPoint( Tinv , data.NECK.pos , pos );

        double TorsoY = atan2( pos[0] , pos[2] );
        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "TorsoY")->index_dof;
        q[index_dof] = TorsoY;


        p3d_mat4TransRot( TrotY , 0 , 0 , 0 , Yaxis , TorsoY  );
        p3d_matMultXform( TrotXtmp , TrotY , TrotYtmp );
        p3d_matInvertXform( TrotYtmp , Tinv  );
        p3d_xformPoint( Tinv , data.SHOULDER_LEFT.pos , pos );

        double TorsoZ = atan2( -pos[0] , pos[1] );
        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "TorsoZ")->index_dof;
        q[index_dof] = TorsoZ;
    }

    // -----------------------------------------------------------------
    // Set and update robot to
    // new position
    // -----------------------------------------------------------------
    p3d_matrix4 Trot;
    p3d_matrix4 Trot2,Trot3;
    p3d_vector3 Zaxis = { 0 , 0 , 1 };
    p3d_matrix4 Trot4;
    //p3d_vector3 vect3;
    p3d_vector3 vect1,vect2;

    if( data.ELBOW_RIGHT.confidence > 0 && data.HAND_RIGHT.confidence > 0 )
    {
        p3d_set_and_update_this_robot_conf( robot, q );

        joint = p3d_get_robot_jnt_by_name(robot, (char*) "rShoulderX");
        p3d_jnt_get_cur_vect_point( joint , pos );
        p3d_jnt_get_cur_vect_point( joint , shoulder );

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "rArmTrans")->index_dof;
        q[index_dof] = p3d_vectDistance( data.ELBOW_RIGHT.pos , pos ) - 0.2066 ;


        joint = p3d_get_robot_jnt_by_name(robot, (char*) "TorsoZ");
        p3d_mat4Copy( joint->abs_pos , Trot );
        Trot[0][3] = pos[0];
        Trot[1][3] = pos[1];
        Trot[2][3] = pos[2];

        //  p3d_mat4Copy( Trot , m_absPos );

        p3d_matInvertXform( Trot , Tinv  );
        p3d_xformPoint( Tinv , data.ELBOW_RIGHT.pos , pos );

        // calcul de la direction pour le bras droit
        //p3d_vector3 sub;
        p3d_vector3 dir;
        p3d_vectNormalize( pos , dir );

        // JP
        // selon x
        double alpha1r = atan2( -pos[2] , -pos[1] );

        p3d_mat4TransRot( Trot2 , 0 , 0 , 0 , Xaxis , alpha1r );
        p3d_matMultXform( Trot , Trot2 , Trot3 );
        p3d_matInvertXform( Trot3 , Tinv  );
        p3d_xformPoint( Tinv , data.ELBOW_RIGHT.pos , pos );

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "rShoulderX")->index_dof;
        q[index_dof] = alpha1r;

        // selon z
        double alpha2r = atan2( pos[0] , -pos[1] );

        //  printf("alpha1 = %f\n", alpha1r );
        //  printf("alpha2 = %f\n", alpha2r );
        //  printf("dir = ( %f , %f , %f )\n", pos[0] , pos[1] , pos[2] );


        p3d_mat4TransRot( Trot2 , 0 , 0 , 0 , Zaxis , alpha2r );
        p3d_matMultXform( Trot3 , Trot2 , Trot4 );
        p3d_matInvertXform( Trot4 , Tinv  );
        p3d_xformPoint( Tinv , data.HAND_RIGHT.pos , pos );

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "rShoulderZ")->index_dof;
        q[index_dof] = alpha2r;

        // selon y
        double alpha3r = atan2( pos[2], pos[0] );

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "rShoulderY")->index_dof;
        q[index_dof] = alpha3r;


        p3d_vectSub( shoulder ,  data.ELBOW_RIGHT.pos , vect1 );
        p3d_vectNormalize( vect1 , vect1 );

        p3d_vectSub( data.HAND_RIGHT.pos , data.ELBOW_RIGHT.pos , vect2 );
        p3d_vectNormalize( vect2 , vect2 );

        // Elbow
        double alpha4r = M_PI - acos( p3d_vectDotProd( vect1 , vect2) ) ;

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "rElbowZ")->index_dof;
        q[index_dof] = alpha4r;
    }

    //---------------------------------------------------------------------------
    //---------------------------------------------------------------------------
    if( data.ELBOW_LEFT.confidence > 0 && data.HAND_LEFT.confidence > 0 )
    {
        joint = p3d_get_robot_jnt_by_name(robot, (char*) "lShoulderX");
        p3d_jnt_get_cur_vect_point( joint , pos );
        p3d_jnt_get_cur_vect_point( joint , shoulder );

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "lArmTrans")->index_dof;
        q[index_dof] = p3d_vectDistance( data.ELBOW_LEFT.pos , pos ) - 0.2066 ;

        joint = p3d_get_robot_jnt_by_name(robot, (char*) "TorsoZ");
        p3d_mat4Copy( joint->abs_pos , Trot );
        Trot[0][3] = pos[0];
        Trot[1][3] = pos[1];
        Trot[2][3] = pos[2];

        //  p3d_mat4Copy( Trot , m_absPos );


        p3d_matInvertXform( Trot , Tinv  );
        p3d_xformPoint( Tinv , data.ELBOW_LEFT.pos , pos );

        // JP
        // selon x
        double alpha1l = atan2( pos[2] , pos[1] );

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "lShoulderX")->index_dof;
        q[index_dof] = alpha1l;

        p3d_mat4TransRot( Trot2 , 0 , 0 , 0 , Xaxis , alpha1l );
        p3d_matMultXform( Trot , Trot2 , Trot3 );
        p3d_matInvertXform( Trot3 , Tinv  );
        p3d_xformPoint( Tinv , data.ELBOW_LEFT.pos , pos );

        // selon z
        double alpha2l = atan2( -pos[0] , pos[1] );

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "lShoulderZ")->index_dof;
        q[index_dof] = alpha2l;

        //  printf("alpha1 = %f\n", alpha1l );
        //  printf("alpha2 = %f\n", alpha2l );
        //  printf("dir = ( %f , %f , %f )\n", pos[0] , pos[1] , pos[2] );

        p3d_mat4TransRot( Trot2 , 0 , 0 , 0 , Zaxis , alpha2l );
        p3d_matMultXform( Trot3 , Trot2 , Trot4 );
        p3d_matInvertXform( Trot4 , Tinv  );
        p3d_xformPoint( Tinv , data.HAND_LEFT.pos , pos );

        // selon y
        double alpha3l = -atan2( -pos[2], pos[0] );

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "lShoulderY")->index_dof;
        q[index_dof] = alpha3l;

        p3d_vectSub( shoulder ,  data.ELBOW_LEFT.pos , vect1 );
        p3d_vectNormalize( vect1 , vect1 );

        p3d_vectSub( data.HAND_LEFT.pos , data.ELBOW_LEFT.pos , vect2 );
        p3d_vectNormalize( vect2 , vect2 );

        double alpha4l = -M_PI + acos( p3d_vectDotProd( vect1 , vect2) ) ;

        index_dof = p3d_get_robot_jnt_by_name(robot, (char*) "lElbowZ")->index_dof;
        q[index_dof] = alpha4l;
    }

    return q;
}




//int main(int argc, char ** argv){
//    //For data/Tests/test.dom
//    std::string hatpRequest("{\"HATP-REQ\": { \"REQ-TYPE\": \"plan\", \"REQ-ID\": 1, \"TASK-NAME\": \"Transfer_two_container\", \"PARAMETERS\": [\"C1\", \"C2\", \"L1\", \"L2\", \"R1\"]}}");

//    msgClient client;
//    client.connect("Tester", "localhost", 5500);

//    if(client.isConnected()){
//        client.sendMessage("HATP", hatpRequest);
//        std::pair<std::string, std::string> result = client.getBlockingMessage();
//        std::cout<<"#### Answer : \n"<<result.second;
//    }

//    exit(0);
//}


bool HATPConnection::connectMsgConnector()
{
#ifdef HATP_CONNECTION
    msgClient client;
    client.connect("move3d", "localhost", 5500);

    if(client.isConnected())
    {
        cout << "move3d is now connected" << endl;
        while (PlanEnv->getBool(PlanParam::GTP_msgconnector))
        {
            g3d_draw_allwin_active();
            if (client.isMessageWaiting())
            {
                std::pair<std::string, std::string> result = client.getMessage();
                std::cout<<"#### Answer : \n"<<result.second << endl;
                string s = interpreteJson(result.second);
                if (s != "command")
                    client.sendMessage(result.first, s);
            }
            else
            {
                usleep(100000);
            }
        }
        return true;
    }
    else
    {
        return false;
    }
#else
    cout << "HATP_CONNECTION not set" << endl;
#endif
}


const string &HATPConnection::interpreteJson(const string &u)
{
    _result = "";
    Json::Value root;
    Json::Reader reader;
    bool parsedSuccess = reader.parse(u, root, false);
    if(!parsedSuccess || root.size() <= 0)
    {
        // Report failures and their locations
        // in the document.
        cout<<"Failed to parse JSON"<< endl;
        _result = "";
        return _result;
    }

    Json::Value res(Json::objectValue);
    Json::Value localRes(Json::objectValue);

    for( Json::ValueIterator itr = root.begin() ; itr != root.end() ; itr++ )
    {
        localRes.clear();

        string fct = "";
        fct = itr.key().asString();
        if (fct == "InitGTP")
        {
            if (TMI)
            {
               if (TMI->getInitWS())
               {
                  g3d_draw_allwin_active();
                  TMI->getInitWS()->load();
               }
               delete TMI;
            }

            TMI = new taskManagerInterface();
            global_Project->getActiveScene()->getFactsManager()->cleanWSList();
            TMI->initTaskManager();
//            PlanEnv->setBool(PlanParam::usingSpark,true);
            localRes["status"] = "OK";
            res[fct] = localRes;
            continue;
        }
        else if(!TMI)
        {
            cout<< "GTP not initialized" << endl;
            continue;
        }


        if (fct == "AddGTPAgent" )
        {
            Json::Value t = root[itr.key().asString()];
            string key = t["key"].asString();
            string val = t["value"].asString();
            TMI->setAgent(key,val);
            localRes["status"] = "OK";
        }
        else if (fct == "AddGTPConstraintArea" )
        {
            Json::Value t = root[itr.key().asString()];
            string name = t["nameId"].asString();
            string obj = t["Object"].asString();
            string area = t["area"].asString();
            bool supp = t["isSupport"].asBool();

            Robot* e = global_Project->getActiveScene()->getRobotByName(obj);

            TMI->getConstraints()->addConstraint(name, e, area , supp);
            localRes["status"] = "OK";
        }
        else if (fct == "AddGTPConstraintFact" )
        {
            Json::Value t = root[itr.key().asString()];
//            string key = t.begin().key().asString();
//            string val = t[t.begin().key().asString()].asString();
//            TMI->setAgent(key,val);

            string factType = t["factType"].asString();
            string factSubType = t["factSubType"].asString();
            string mainEntityName = t["mainEntity"].asString();
            string targetEntityName = t["targetEntity"].asString();
            bool boolValue = t["boolValue"].asBool();
            int valueType = t["valueType"].asInt();
            double specificValue = t["specificValue"].asInt();

            fact_id curFI;
            vector<fact_id> vf = global_Project->getActiveScene()->getFactsManager()->getFactList();

            curFI.factType = "none";
            for (unsigned int j = 0; j < vf.size(); j++)
            {
                fact_id tmp = vf.at(j);
                if (tmp.factType == factType && tmp.factSubType == factSubType)
                {
                    curFI = tmp;
                    break;
                }
            }

            if (curFI.factType == "none")
            {
                cout << "fact type not recognized" << endl;
                localRes["status"] = "ERROR";
                res[fct] = localRes;
                break;

            }

            Fact f(curFI);
            Robot* mainEntity = global_Project->getActiveScene()->getRobotByName(mainEntityName);
            if (!mainEntity)
            {
                cout << "no main entity found" << endl;
                localRes["status"] = "ERROR";
                res[fct] = localRes;
                break;
            }

            Robot* targetEntity = global_Project->getActiveScene()->getRobotByName(targetEntityName);

            if (!targetEntity)
            {
                cout << "no target entity found" << endl;
                localRes["status"] = "ERROR";
                res[fct] = localRes;
                break;
            }

            f.setIsComputed(false);
            f.setMainEntity(mainEntity);
            f.setTargetEntity(targetEntity);

            f.setBoolValue(boolValue);

            f.setValueType(valueType);
            f.setSpecValue(specificValue);

            TMI->getConstraints()->addConstraint(f);
            localRes["status"] = "OK";
        }
        else if (fct == "AddGTPData" )
        {
            Json::Value t = root[itr.key().asString()];
            string key = t["key"].asString();
            string val = t["value"].asString();
            TMI->addAdditinalData(key,val);
            localRes["status"] = "OK";
        }
        else if (fct == "AddGTPPoint" )
        {
            Json::Value t = root[itr.key().asString()];
            string key = t["key"].asString();
            p3d_point val;// = t["value"].asString();
            val.x = t["x"].asDouble();
            val.y = t["y"].asDouble();
            val.z = t["z"].asDouble();
            TMI->addPoint(key,val);
            localRes["status"] = "OK";
        }
        else if (fct == "AddGTPObject" )
        {
            Json::Value t = root[itr.key().asString()];
            string key = t["key"].asString();
            string val = t["value"].asString();
            TMI->setObject(key,val);
            localRes["status"] = "OK";
        }
        else if (fct == "Change_obj_col" )
        {
            Json::Value t = root[itr.key().asString()];
            string obj = t["object_name"].asString();
            int color = t["color"].asInt();
            p3d_rob* rob = p3d_get_robot_by_name(obj.c_str());
            cout << "previous color is : " << p3d_set_object_color(rob,color) << ", new color is : " << color << endl;
            localRes["status"] = "OK";
        }
        else if (fct == "CleanGTPPlan" )
        {
            TMI->getTaskManager()->getOnePlan()->clearPlan();
            localRes["status"] = "OK";
        }
        else if (fct == "ClearGTPInputs" )
        {
            TMI->clearAllInputs();
            localRes["status"] = "OK";
        }
        else if (fct == "DisableAllGTPFact" )
        {
            vector<fact_id> vfi =  global_Project->getActiveScene()->getFactsManager()->getFactList();
            vector<string> vs;
            for (unsigned int i = 0; i < vfi.size(); i++)
            {
              fact_id FI = vfi.at(i);
              bool isAlreadyIn = false;
              for (unsigned int j = 0; j < vs.size(); j++)
              {
                if (FI.factType == vs.at(j))
                {
                  isAlreadyIn = true;
                  break;
                }
              }
              if (!isAlreadyIn)
                vs.push_back(FI.factType);
            }


            for (unsigned int i = 0; i < vs.size(); i++)
            {
              global_Project->getActiveScene()->getFactsManager()->disableFacts(vs.at(i));
            }
            localRes["status"] = "OK";
        }
        else if (fct == "DisableGTPFact" )
        {
            Json::Value t = root[itr.key().asString()];
            string factType = t["factType"].asString();
            bool valueT = t["value"].asBool();
            bool test = false;
            if (valueT)
              test = global_Project->getActiveScene()->getFactsManager()->disableFacts(factType);
            else
              test = global_Project->getActiveScene()->getFactsManager()->enableFacts(factType);

            if (!test)
                localRes["status"] = "ERROR";
            else
                localRes["status"] = "OK";

        }
        else if (fct == "FindGTPTaskAlternative" )
        {
            Json::Value t = root[itr.key().asString()];
            int taskId = t["taskId"].asInt();
            bool computeMP = t["computeMP"].asBool();


            TMI->setComputeMP(computeMP);
            TMI->callAlternative(taskId);

            PlanNode* PN  = TMI->getCurrentPN();

            if (!PN->getTaskSolution())
            {
                printf("failed to find a solution");
            }
            else
            {
                localRes["taskId"] = PN->getId().first;
                localRes["taskAlternativeId"] = PN->getId().second;
                localRes["cost"] = PN->getTaskSolution()->getCost();
            }
            localRes["planningInformation"] = TMI->getCurrentPN()->SD->getJsonValue();//planInfo;
        }
        else if (fct == "GetGTPFacts" )
        {
            Json::Value t = root[itr.key().asString()];

            int taskId = t["taskId"].asInt();
            int alternativeId = t["alternativeId"].asInt();

            if (taskId == -1 && alternativeId == -1)
            {
              cout << "Ids -1 -1" << endl;
              if(!TMI->getInitWS())
              {
                  localRes["status"] = "ERROR";
                  res[fct] = localRes;
                  break;
              }
              TMI->getInitWS()->load();
            }
            else
            {
                Task* t = TMI->getTaskManager()->getTask(taskId);
                if (!t)
                {
                    cout << "ERROR: no such task!" << endl;
                    localRes["status"] = "ERROR";
                    res[fct] = localRes;
                    break;
                }
                TaskSolution* TS = t->getResult()->getAlternativeWithId(alternativeId);
                if (!TS)
                {
                    cout << "ERROR: no such alternative!" << endl;
                    localRes["status"] = "ERROR";
                    res[fct] = localRes;
                    break;
                }

                TS->getStopWS()->load();
            }
            int id = global_Project->getActiveScene()->getFactsManager()->computeFacts(false);
            if (id == -1)
            {
                localRes["status"] = "ERROR";
                res[fct] = localRes;
                break;
            }
            WorldStateFacts* WSF = global_Project->getActiveScene()->getFactsManager()->getWSFromId(id);
            if (!WSF)
            {
                localRes["status"] = "ERROR";
                res[fct] = localRes;
                break;
            }
            vector<Fact> vf = WSF->getMeaningfulFacts();

            //if (vf.size() > NB_MAX_FACTS)
            //{
            //  list_gtp_facts->nbFacts = vf.size();
            //  *report = S_mhp_FAIL_FACTS_OVERFLOW_GTP;
            //  return ETHER;
            //}



            localRes["nbFacts"] = (int)vf.size();
            Json::Value facts(Json::arrayValue);

            for(unsigned int i = 0; i < vf.size(); i++)
            {
                Json::Value oneFact(Json::objectValue);

                oneFact["factType"] = vf.at(i).getType().factType;
                oneFact["factSubType"] = vf.at(i).getType().factSubType;
                //changing order for isreachable and isvisible
                if (!vf.at(i).getType().isMainAnAgent)
                {
                    oneFact["mainEntity"] = vf.at(i).getMainEntity()->getName();
                    oneFact["targetEntity"] = vf.at(i).getTargetEntity()->getName();
                }
                else
                {
                    oneFact["targetEntity"] = vf.at(i).getMainEntity()->getName();
                    oneFact["mainEntity"] = vf.at(i).getTargetEntity()->getName();
                }
                oneFact["isComputed"] = vf.at(i).isItComputed();
                oneFact["boolValue"] = vf.at(i).getBoolValue();
                oneFact["valueType"] = vf.at(i).getValueType();
                oneFact["specificValue"] = vf.at(i).getSpecValue();

                facts.append(oneFact);

            }

            localRes["facts"] = facts;
        }
        else if (fct == "PlanGTPTask" )
        {

            Json::Value t = root[itr.key().asString()];
            string type = t["type"].asString();
            int taskId = t["previousTaskId"].asInt();
            int taskAltId = t["previousTaskAlternativeId"].asInt();
            bool computeMP = t["computeMotionPlan"].asBool();
            cout << "computing for " << type << endl;

            WorldState *WS = NULL;
            if (taskId >= 0)
            {
              if (TMI->getTaskManager()->getTask(taskId))
              {
                TaskSolution* TS = TMI->getTaskManager()->getTask(taskId)->getResult()->getAlternativeWithId(taskAltId);
                if(TS)
                {
                   WS = TS->getStopWS();
                }
              }
            }
            if (!WS && taskId == -1 && taskAltId == -1)
            {
              cout << "Ids -1 -1" << endl;
              WS = TMI->getInitWS();
            }

            if (!WS)
            {
              cout << "no world state with this Ids, creating a new one from current" << endl;
              p3d_destroy_all_grasps();
              WS = new WorldState(global_Project->getActiveScene());
              TMI->getTaskManager()->addWorldState(WS);
              WS->saveAll();
              WS->setAttachements(TMI->getAttachements());
            }

            TMI->setType(TMI->getTaskManager()->getTaskIdFromString(type));
            TMI->setTypeStr(type);

            TMI->setWorldState(WS);
            TMI->setComputeMP(computeMP);

            TMI->callGTP();
            TaskSolution* TSol = TMI->getCurrentSol();


//            localRes["returnReport"] = TMI->getCurrentPN()->SD->getStringMapEntry("returnReport");
            if (!TSol)
            {
                printf("failed to find a solution");
            }
            else
            {
                localRes["taskId"] = TSol->getTask()->getId();
                localRes["taskAlternativeId"] = TSol->getId();
                localRes["cost"] = TSol->getCost();
            }
            localRes["planningInformation"] = TMI->getCurrentPN()->SD->getJsonValue();//planInfo;
        }
        else if (fct == "GetGTPDetails" )
        {
            Json::Value t = root[itr.key().asString()];
            int taskId = t["taskId"].asInt();
            int alternativeId = t["alternativeId"].asInt();

            Task* tas = TMI->getTaskManager()->getTask(taskId);
            if (!tas)
            {
                localRes["status"] = "no_action";
            }
            else{
                TaskSolution* TS = tas->getResult()->getAlternativeWithId(alternativeId);
                if (!TS)
                {
                    localRes["status"] = "no_alternative";
                }
                else
                {
                    vector<TaskSubSolution*> vrs = TS->getAllTSS();
                    Json::Value subTrajsList(Json::arrayValue);
                    for (unsigned int i = 0; i < vrs.size(); i++)
                    {
                        Json::Value sti(Json::objectValue);
                        TaskSubSolution* TSS = vrs.at(i);
                        sti["subTrajId"] = TSS->getId();
                        sti["armId"] = TSS->getArmId();
                        sti["subTrajName"] = TSS->getSubSolutionName();
                        sti["agent"] = TSS->getRobot()->getName();
                        subTrajsList.append(sti);
                    }
                    localRes["subTrajs"] = subTrajsList;
                    localRes["status"] = "OK";
                    localRes["nbSubTraj"] = (int)vrs.size();
                }
            }

        }
        else if (fct == "GetGTPTraj" )
        {
            Json::Value t = root[itr.key().asString()];
            int taskId = t["taskId"].asInt();
            int alternativeId = t["alternativeId"].asInt();
            int subTrajId = t["subTrajId"].asInt();

            Task* tas = TMI->getTaskManager()->getTask(taskId);
            if (!tas)
            {
                localRes["status"] = "no_action";
            }
            else{
                TaskSolution* TS = tas->getResult()->getAlternativeWithId(alternativeId);
                if (!TS)
                {
                    localRes["status"] = "no_alternative";
                }
                else
                {
                    TaskSubSolution* TSS = TS->getSubSolutionWithId(subTrajId);
                    if (!TSS)
                    {
                        localRes["status"] = "no_subTraj";
                    }
                    else
                    {
                        Robot* r = TSS->getRobot();

                        Json::Value sti(Json::arrayValue);
                        Json::Value confs(Json::arrayValue);
                        if (!TSS->isNavigation())
                        {
                            p3d_traj* t = TSS->getTraj();
                            localpath* lp = t->courbePt;

                            while(lp)
                            {
                                configPt q = lp->config_at_param(r->getRobotStruct(),lp,0);
                                sti.clear();
                                for (int i = 0; i < r->getRobotStruct()->nb_dof; i++)
                                {
                                    sti.append(q[i]);
                                }
                                confs.append(sti);

                                lp = lp->next_lp;
                            }

                            configPt q = TSS->getEndState()->getRobotConf(r)->getConfigStruct();
                            sti.clear();
                            for (int i = 0; i < r->getRobotStruct()->nb_dof; i++)
                            {
                                sti.append(q[i]);
                            }
                            confs.append(sti);
                            localRes["type"] = "manip";
                        }
                        else
                        {
                            vector<p3d_point> navVect = TSS->getNavVector();
                            for (unsigned int i = 0; i<navVect.size(); i++)
                            {
                                sti.clear();
                                sti.append(navVect.at(i).x);
                                sti.append(navVect.at(i).y);
                                sti.append(navVect.at(i).z);
                                confs.append(sti);
                            }
                            localRes["type"] = "nav";
                        }


                        localRes["confs"] = confs;
                        localRes["status"] = "OK";
                        localRes["nbConfs"] = (int)confs.size();
                        localRes["agent"] = TSS->getRobot()->getName().c_str();
                    }
                }
            }

        }
        else if (fct == "UpdateGTPObjects" )
        {
            Json::Value t = root[itr.key().asString()];
            WorldState* WS = new WorldState(global_Project->getActiveScene());
            WS->saveAll();
            for (Json::Value::iterator it = t.begin(); it != t.end(); ++it)
            {
                const Json::Value& objJson = *it;
                string objName = objJson["name"].asString();
                Robot* r = global_Project->getActiveScene()->getRobotByName(objName);
                confPtr_t q = r->getCurrentPos();
                //cout << objJson.toStyledString() << endl;

                for(int i = 0; i < 6; i++)
                {
                    (*q)[i+6] = objJson["conf"][i].asDouble();
                }

                WS->replaceConf(q);
            }

            TMI->setWorldState(WS);
            WS->load();

            localRes["status"] = "OK";

        }
        else if (fct == "UpdateGTPRobots" )
        {
            Json::Value t = root[itr.key().asString()];
            WorldState* WS = new WorldState(global_Project->getActiveScene());

            WS->saveAll();
            for (Json::Value::iterator it = t.begin(); it != t.end(); ++it)
            {
                const Json::Value& objJson = *it;
                string objName = objJson["name"].asString();
                Robot* r = global_Project->getActiveScene()->getRobotByName(objName);
                confPtr_t q = r->getCurrentPos();
                //cout << objJson.toStyledString() << endl;

                for(unsigned int i = 0; i < objJson["conf"].size(); i++)
                {
                    (*q)[i+6] = objJson["conf"][i].asDouble();
                }

                WS->replaceConf(q);
            }

            TMI->setWorldState(WS);
            WS->load();

            localRes["status"] = "OK";

        }
        else if (fct == "UpdateGTPHumans" )
        {
            Json::Value t = root[itr.key().asString()];
            WorldState* WS = new WorldState(global_Project->getActiveScene());
            WS->saveAll();
            for (Json::Value::iterator it = t.begin(); it != t.end(); ++it)
            {
                const Json::Value& objJson = *it;
                string objName = objJson["name"].asString();
                Robot* r = global_Project->getActiveScene()->getRobotByName(objName);
                confPtr_t q = r->getCurrentPos();
                //cout << objJson.toStyledString() << endl;

                for(unsigned int i = 0; i < objJson["conf"].size(); i++)
                {
                    (*q)[i+6] = objJson["conf"][i].asDouble();
                }
                WS->replaceConf(q);
                WS->load();

                HumanJoints data;
                for (int i = 0; i < 3; i++)
                {
                    data.HAND_RIGHT.pos[i] = objJson["joints"]["arm1"][i].asDouble();
                    data.HAND_LEFT.pos[i] = objJson["joints"]["arm1"][i].asDouble();
                    data.HEAD.pos[i] = objJson["joints"]["head"][i].asDouble();

                    data.FOOT_RIGHT.pos[i] = objJson["joints"]["leg1"][i].asDouble();
                    data.FOOT_LEFT.pos[i] = objJson["joints"]["leg2"][i].asDouble();

                    data.TORSO.pos[i] = objJson["joints"]["torso"][i].asDouble();
                }

                data.HAND_LEFT.confidence = 1;
                data.HAND_LEFT.confidence = 1;
                data.HEAD.confidence = 1;
                data.FOOT_RIGHT.confidence = 1;
                data.FOOT_LEFT.confidence = 1;
                data.TORSO.confidence = 1;



                //configPt q2 = getConfFromJointsPos(r->getRobotStruct(),data);

                confPtr_t q2(new Configuration(r,getConfFromJointsPos(r->getRobotStruct(),data)));
                WS->replaceConf(q2);
            }

            TMI->setWorldState(WS);
            WS->load();

            localRes["status"] = "OK";

        }
        else if (fct == "AddToList" )
        {
            Json::Value t = root[itr.key().asString()];
            string act = t.asString();
            TMI->setTypeStr(act);
            cout << "act = " << act << endl;
            TMI->setType(TMI->getTaskManager()->getTaskIdFromString(act));
            TMI->addCurrentTaskToList();
        }
        else if (fct == "ComputePlan" )
        {
            p3d_destroy_all_grasps();
            WorldState* WS = new WorldState(global_Project->getActiveScene());
            TMI->getTaskManager()->addWorldState(WS);
            WS->saveAll();
            TMI->setWorldState(WS);
            TMI->computePlan(true,0);
        }
        else if (fct == "RemoveGTPConstFromSol" )
        {
            Json::Value t = root[itr.key().asString()];
            int taskId = t["taskId"].asInt();
            int alternativeId = t["alternativeId"].asInt();
            TMI->removeConstraintFromPrevTS(taskId, alternativeId);
            localRes["status"] = "OK";
        }
        else if (fct == "ResetAllGTP" )
        {
            Json::Value t = root[itr.key().asString()];
            cout << "not implemented yet" << endl;
            localRes["status"] = "ERROR";
        }
        else if (fct == "SetGTPInitState" )
        {
            WorldState* WS = new WorldState(global_Project->getActiveScene());
            TMI->getTaskManager()->addWorldState(WS);
            WS->saveAll();
            TMI->setInitWS(WS);
//            TMI->getTaskManager()->addWorldState(WS);
            localRes["status"] = "OK";
        }
        else if (fct == "COMMAND" )
        {
            if (root["COMMAND"].asString() == "CONNECT")
                cout << "connected to hatp" << endl;
            else if (root["COMMAND"].asString() == "EXIT")
                cout << "ended connection to hatp" << endl;
            _result = "command";
            return _result;
        }
        else if (fct == "PING" )
        {
            localRes["status"] = "OK";
        }
        else
        {
            cout << "HATPConnection: fonction not implemented yet" << endl;
            localRes["status"] = "ERROR";
        }

        res[fct] = localRes;

    }
    localRes.clear();
//    cout << res << endl;


    Json::FastWriter wrt;
    _result = wrt.write(res);
    res.clear();
    return _result;
}


void HATPConnection::readFromFile(std::string path)
{

    std::ifstream file_id;
    file_id.open(path.c_str());

    Json::Value root;
    Json::Reader reader;
    bool parsedSuccess = reader.parse(file_id, root, false);
    if(!parsedSuccess || root.size() <= 0)
    {
        cout<<"Failed to parse JSON from file " << path << endl;
    }
    int i = 0;
//    string String = static_cast<ostringstream*>( &(ostringstream() << i) )->str();

    bool endOfFile = false;
    Json::FastWriter fastWriter;
    while (!endOfFile)
    {
        string key = static_cast<ostringstream*>( &(ostringstream() << i++) )->str();
        if (root.isMember(key))
        {
            std::string value = fastWriter.write(root[key]);
            interpreteJson(value);
        }
        else
        {
            endOfFile = true;
        }

    }
}



