#ifndef TASKSUBSOLUTION_H
#define TASKSUBSOLUTION_H

#include "API/Device/robot.hpp"
#include "GTP/Tasks/task.hpp"


class TaskSolution;

class TaskSubSolution
{
public:
    TaskSubSolution();
    TaskSubSolution(TaskSolution* sol, std::string name, int id, p3d_traj* traj, Robot* r, double initRP);

    ~TaskSubSolution();


    bool refineTraj(); //smoothing mainly
    bool reComputeTraj();

    std::string getSubSolutionName() {return _subSolutionName;}
    void setSubSolutionName(std::string name){_subSolutionName = name;}

    Robot* getRobot() {return _robot;}
    int getId() {return _id;}
    void setId(int id) {_id = id;}

    std::vector<TaskSubSolution*> getPrevious() {return _previous; }
    void addToPrevious(TaskSubSolution* prev);
    void setPrevious(std::vector<TaskSubSolution*> prev) {_previous = prev;}

    std::vector<TaskSubSolution*> getNext() { return _next; }
    void setNext(std::vector<TaskSubSolution*> next);
    void addNext(TaskSubSolution* next);

    p3d_traj* getTraj();// {return _trajectory;}
    void setTraj(p3d_traj* traj) {_trajectory = traj;}

    Task* getTask();
    TaskSolution* getSolution() {return _sol;}

    void setIsLast(bool v) {_isLast = v;}
    bool isLast() {return _isLast;}
    void setIsSmoothed(bool v) {_isSmoothed = v;}
    bool isSmoothed() {return _isSmoothed;}

    void setStartState(WorldState* WS) { _start_WS = WS;}
    void setEndState(WorldState* WS) { _end_WS = WS;}

    WorldState* getStartState() {return _start_WS;}
    WorldState* getEndState() {return _end_WS;}

    double getInitRangeParam() {return _initRangeParam;}

    std::string getDescr();

    bool setTrajInRobotStruct();

    double getDuration();

    void setNavigation(bool nav);
    bool isNavigation();

    void setNavTraj(std::vector<p3d_point> p);
    std::vector<p3d_point> getNavVector();

    void saveToDisk(FILE *file);

    int getArmId() {return _armId;}
    void setArmId(int i) {_armId = i;}

    Sem::PlanningPart getPlanningPart() const {return _planningPart;}
    void setPlanningPart(Sem::PlanningPart part){_planningPart = part;}
private:
    std::string _subSolutionName;
    int _id;
    std::vector<TaskSubSolution*> _previous;
    std::vector<TaskSubSolution*> _next;

    double _initRangeParam;

    p3d_traj* _trajectory;

    TaskSolution* _sol;

    bool _isLast;

    WorldState* _start_WS;
    WorldState* _end_WS;

    bool _isSmoothed;
    Robot* _robot;

    std::vector<p3d_point> _nav_path;
    bool _navigation;

    //0 for right, 1 for left, 2 for both
    int _armId;
    Sem::PlanningPart _planningPart;
};



#endif // TASKSUBSOLUTION_H
