#include "taskConstraints.hpp"

using namespace std;

TaskConstraints::TaskConstraints()
{
}

TaskConstraints::TaskConstraints(TaskConstraints* TC)
{
    factConstraints = TC->getFactConstraints();
    ressourceConstraints = TC->getRessourceConstraints();
    regionConstraints = TC->getRegionConstraints();
}

void TaskConstraints::addConstraint(TaskConstraints *TC)
{
    if (!TC)
        return;

    for (unsigned int i = 0; i < TC->getFactConstraints().size(); i++)
    {
        factConstraints.push_back(TC->getFactConstraints().at(i));
    }

    for (unsigned int i = 0; i < TC->getRessourceConstraints().size(); i++)
    {
        ressourceConstraints.push_back(TC->getRessourceConstraints().at(i));
    }

    for (unsigned int i = 0; i < TC->getRegionConstraints().size(); i++)
    {
        regionConstraints.push_back(TC->getRegionConstraints().at(i));
    }
}

void TaskConstraints::addConstraint(Fact f)
{
    factConstraints.push_back(f);
}

void TaskConstraints::addConstraint(ressourceConst r)
{
    ressourceConstraints.push_back(r);
}

void TaskConstraints::addConstraint(regionConst r)
{
    regionConstraints.push_back(r);
}

void TaskConstraints::addConstraint(positionConst r)
{
    positionConstraints.push_back(r);
}

void TaskConstraints::addConstraint(string n, Robot* e)
{
    ressourceConst rc;
    rc.entity = e;
    rc.nameId = n;
    ressourceConstraints.push_back(rc);
}

void TaskConstraints::addConstraint(string n, Robot *obj, string area, bool isSupport)
{
    regionConst rc;
    rc.nameId = n;

    GeometricForm* gf = NULL;
    if (isSupport && area != "")
    {
        for (unsigned int i = 0; i < obj->getObjectRob()->getSupports().size(); i++)
        {
            GeometricForm* tmp = obj->getObjectRob()->getSupports().at(i);
            if (tmp->supportAreas.count(area) != 0)
            {
                gf = tmp->supportAreas[area];
            }
        }
        if (!gf)
        {
            cout << "ERROR this area( "<< area << " ) doesn t exist" << endl;
        }
    }
    else
    {
        GeometricComputations GC = GeometricComputations();
        vector<p3d_point> vp = GC.getLowerPointsInObject(obj,0.03);
        if (vp.size()==4)
        {
            p3d_vector3 m;
            m[0] = 0;
            m[1] = 0;
            m[2] = 0;
            gf = new RectangleForm(vp.at(0),vp.at(1),vp.at(2),vp.at(3),m);
        }
        if (!gf)
        {
            cout << "ERROR: object ( "<< obj->getName() << " ) is too tilted to get an area from it" << endl;
        }

    }

    rc.GeoForm = gf;


    regionConstraints.push_back(rc);
}

void TaskConstraints::addConstraint(string n, GeometricForm* gf)
{
    regionConst rc;
    rc.GeoForm = gf;
    rc.nameId = n;
    regionConstraints.push_back(rc);
}

void TaskConstraints::addConstraint(string n, Robot* e, string objInEntity, p3d_point t, p3d_point r )
{
    positionConst pc;
    pc.nameId = n;
    pc.entity = e;
    pc.objInEntity = objInEntity;
    pc.translations = t;
    pc.rotations = r;
    positionConstraints.push_back(pc);
}

void TaskConstraints::cleanAll()
{
    regionConstraints.clear();
    ressourceConstraints.clear();
    factConstraints.clear();
}

bool TaskConstraints::areFactConstraintRespected(WorldState* WS)
{
    if(factConstraints.size() == 0 )
        return true;


//    vector<Fact> ff = WS->getAllFacts();
    // if what in constraint is in ff, return true
    for (unsigned int i = 0; i < factConstraints.size(); i++)
    {
        Fact fconst = factConstraints.at(i);
        Fact wsFacct = WS->getThisFact(fconst);
        if (fconst.compareFactsValue(wsFacct))
        {
            return true;
        }
    }
    return false;
}

string TaskConstraints::getDescr()
{
    string s = "===============\n";
    s+= "   facts:\n";


    for (unsigned int i = 0; i < factConstraints.size(); i++)
    {
        Fact f = factConstraints.at(i);
        s+=f.getFactDescr();
        s+="\n";
    }

    s+= "\n   ressources:\n";
    for (unsigned int i = 0; i < ressourceConstraints.size(); i++)
    {
        ressourceConst rc = ressourceConstraints.at(i);
        s+= rc.nameId + " " + rc.entity->getName();
        s+="\n";
    }

    s+= "\n   regions:\n";
    for (unsigned int i = 0; i < regionConstraints.size(); i++)
    {
        regionConst rg = regionConstraints.at(i);
        s+= rg.nameId + " " + rg.GeoForm->type;
        s+="\n";
    }
    s += "===============\n";
    return s;
}
