#include "cooperativeTask.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"


CooperativeTask::CooperativeTask() : ManipulationTask()
{

}

bool CooperativeTask::setTargetAgent(std::multimap<std::string,std::string> agents_name)
{
    if (agents_name.count("targetAgent") != 1)
    {
        std::cout << "ERROR: no target agent for " << type << " task." << std::endl;
        SD->setMapEntry("returnReport","no_target_agent");
        return false;
    }
    targetAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("targetAgent")->second);
    if (!targetAgent)
        return false;
    return true;
}

bool CooperativeTask::computeConf(confPtr_t &q_target, p3d_matrix4 objectPosMat, gpGrasp grasp)
{
    confPtr_t qInit = _WS->getRobotConf(mainAgent);
    confPtr_t q;

    p3d_matrix4 /*graspMatWorldFrame,*/ reelTarg;
    p3d_mat4Mult(objectPosMat, grasp.frame, reelTarg);


    p3d_matrix3 matRot;
    p3d_mat4ExtractRotMatrix(objectPosMat, matRot);

    bool testIk = computeInverseKinematic(mainAgent, reelTarg, q, _armId, true );
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbGikCalled");

    _returnReportcount++;
    if (!testIk)
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","no_IK_for_conf");
        }
        return false;
    }


    q_target = confPtr_t(new Configuration(*q));
    mainAgent->setAndUpdate(*qInit);
    return true;
}
