#ifndef MANIPULATIONTASK_HPP
#define MANIPULATIONTASK_HPP

#include "GTP/Tasks/privateTask.hpp"

class ManipulationTask : public PrivateTask
{
public:
    ManipulationTask();
    //virtual bool init();  //should be called after setting main agent
    virtual bool findCandidateSolutions() = 0;
    virtual bool getRandomSol() = 0;
    virtual bool setToSolution(int solutionId, bool computeMP);
    virtual TaskSolution *createSolution(bool computeMP) = 0;
//    virtual vector<TaskSubSolution*> setNextTraj(TaskSubSolution* doneTSS) =0;
    virtual bool computeSolutionTrajectories(int alternativeId);
    /// overides computeSolutionTrajectories(int solutionId)
    virtual bool computeSolutionTrajectories(TaskSolution *ts) = 0;

    virtual bool run(std::multimap<std::string,std::string> agents_name,
             std::multimap<std::string,std::string> objects_name,
             std::multimap<std::string,p3d_point> point3d,
             std::multimap<std::string,std::string> additionalData,
             WorldState* WS, bool computeMP);

    /**
     * @brief get an alternative created by createAlternative and add it to results, or create a new one if no more are presents.
     * @param computeMP
     * @return
     */
    TaskSolution* findAlternative(bool computeMP);
    /**
     * @brief randomly create a solution
     * @param computeMP
     * @return
     */
    TaskSolution* createAlternative(bool computeMP);

    std::vector<confPtr_t> compute_manipulation_data(int armId, Robot *object, gpGrasp grasp);

    bool setMainObject(std::multimap<std::string,std::string> objects_name);

    Robot* getMainObject(){return mainObject;}

    virtual void showMainObjectRandomTransform();

    bool attachObject();
    bool detachObject();
    bool checkHRIAgent(Robot* r);
    bool setMainAgentHand();
    bool isObjectInCollision(Robot *supportObject);
    bool isHandInCollision();

    void removeVirtualObjects();
    void moveBackVirtualObjects();



    /**
     * @brief check the attachements
     * @return true if attachement exist between mainagent and mainobject
     */
    bool checkAttachement();

    /**
     * @brief check if the arm and objects coresponds
     * @return true if either the arm or the object are set, and set the one that is not
     *         false in case of cross inputs
     */
    bool checkAndSetArmAndObj();

    /**
     * @brief use the arm to find the object attached
     * @return the object attached to the arm
     */
    Robot* getMainObjectFromAtt();

    /**
     * @brief use the attachement to find the object
     * @return the arm id of the arm to which the mainObject is attached
     */
    int getArmIdFromAtt();

    /**
     * @brief check if there is an attachement in _WS of the main object in the good hand
     * @return return true if there is.
     */
    bool checkArmAtt();

protected:

    Robot* mainObject;

    confPtr_t graspConfClosed;
    confPtr_t graspConfOpen;
    confPtr_t approachConf;
    confPtr_t escapeConf;

    int currentTransfo;
    std::map<int, Robot*> _hands;
    gpGrasp _curGrasp;

    std::vector<std::pair<Robot*,configPt> > virtualObjs;

    std::vector<std::pair<double,TaskSolution*> > exploredSolutions;

    bool _noArm;


};

#endif // MANIPULATIONTASK_HPP
