#include "placeReachable.hpp"
#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

using namespace std;

PlaceReachable::PlaceReachable() : Place()
{
    type = "placeReachable";
    typeId = placeReachable;
    _armId = -1;

    noSupport = false;

    approachTraj = NULL;
    placeTraj = NULL;
    escapeTraj = NULL;
    releaseTraj = NULL;

    nbMaxLoops = 20;
    maxTests = 500;


    SD->getSDMapEntry("statistics")->setMapEntry("nbMaxLoops", nbMaxLoops);
    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 5);
    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    SD->setMapEntry("maxTests", maxTests);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", maxTests*nbMaxLoops);
    SD->setMapEntry("returnReport", "ERROR");

    _possibleMainAgentType.push_back(HRI_PR2);

}

bool PlaceReachable::findCandidateSolutions()
{
    if (!checkAndSetArmAndObj())
        return false;
    //check if seted hand is the same as the one with the object

    if (!checkAttachement())
    {
        SD->setMapEntry("returnReport","no_corresponding_attachement_in_ws");
        return false;
    }

    if (!checkHRIAgent(mainAgent))
        return false;

    if (!setMainAgentHand())
        return false;

    switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
    {
    case 0:
        //finding support either from set support object, or from env.
        if (!findSupports())
            return false;
        break;
    case 1:
        if (!initMightabilities())
            return false;
        break;
    default:
        SD->setMapEntry("returnReport","typeCase_not_correct");
        return false;
        break;
    }

    if (!findRotations())
        return false;

    //computing human reach
    int nbArms = _targetAgent->getHriAgent()->manip->shoulder_joint_ids_no;
    double agReach = _targetAgent->getHriAgent()->manip->reach_tolerance;
    //TODO: there is a problem if multiple support in an object

//    vector<GeometricForm*> newVs;
    if(!noSupport)
    {
        _supportList.clear();
        _supportList.push_back(_supportObject);
    }
    vector<Robot*> vr;
    for(unsigned int k = 0; k < _supportList.size(); k++)
    {
        for(int j = 0; j < nbArms; j++)
        {
            p3d_point agShoulder;

            agShoulder.x = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[0][3];
            agShoulder.y = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[1][3];
            agShoulder.z = _targetAgent->getJoint(_targetAgent->getHriAgent()->manip->shoulder_joint_ids[j])->getJointStruct()->abs_pos[2][3];

            //this is the real distance, with taking z into acount
            //double disttoP = supp.object->getObjectRob()->distToPoint(agShoulder);
            //this doesn t take z into account
            double disttoP = _supportList.at(k)->getObjectRob()->distToPointZ(agShoulder);
            cout << "objects tested = " << _supportList.at(k)->getName() << " distance = " << disttoP << endl;
            if ( disttoP < agReach)
            {
                vr.push_back(_supportList.at(k));
                cout << "object reachable = " << _supportList.at(k)->getName() << endl;
            }
        }
    }

    _supportList = vr;
//    _supports = newVs;
//    if(!noSupport && _supportList.size() == 1)
//    {
//        _supports = _supportList.at(0)->getObjectRob()->getSupports();
//    }
//    else if (_supportList.size() > 1)
//    {
//        cout << "ERROR: got more than one support." << endl;
//        SD->setMapEntry("returnReport", "too_much_support");
//        return false;
//    }



    if (_supportList.size() == 0)
    {
        cout << "no reachable object for target Agent " << type << " is not feasible" << endl;
        SD->setMapEntry("returnReport", "no_reachable_support");
        return false;
    }

    _WSF = new WorldStateFacts(_WS->getRobotConfs());

    return true;
}

bool PlaceReachable::getRandomSol()
{

    confPtr_t robotCurConf = mainAgent->getCurrentPos();
    confPtr_t objectCurConf = mainObject->getCurrentPos();

    detachObject();

    bool res = false;
    int nbTestsPlace = -1;



    int reachedReturnReportCount = _returnReportcount;
    while (!res && nbTestsPlace < maxTests)
    {
        SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
        nbTestsPlace++;
        _returnReportcount = reachedReturnReportCount;
        mainObject->setAndUpdate(*objectCurConf);
//        cout<< "test nb = " << nbTests << endl;

        if (!findCurrentTestPoint())
            continue;

//        configPt q = findCurrentRotations();
//        if (!q)
//            continue;

        configPt q = findObjectConf();
        if (!q)
            continue;

        p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),q);


        if (!isObjectInSurface())
            continue;

        if(isObjectInCollision(_supportObject))
            continue;

        if(isHandInCollision())
            continue;

         _confSol = confPtr_t(new Configuration(mainObject,q));

         p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), _targetAgent->getRobotStruct());
        if(!checkHumanReach())
        {
//            cout << "noooooooo" << endl;
            continue;
        }
        p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _targetAgent->getRobotStruct());


        if (!testFactConstraints())
            continue;


//        _confSol = confPtr_t(new Configuration(mainObject,q));

        res = true;
    }


    mainAgent->setAndUpdate(*robotCurConf);
    mainObject->setAndUpdate(*objectCurConf);
    attachObject();
    return res;

}


bool PlaceReachable::setTargetAgent(multimap<string,string> agents_name)
{
    if (agents_name.count("targetAgent") != 1)
    {
        cout << "ERROR: no target agent for " << type << " task." << endl;
        return false;
    }
    _targetAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("targetAgent")->second);
    if (!_targetAgent)
        return false;
    return true;
}

bool PlaceReachable::setAgents(multimap<string,string> agents_name)
{
    return setMainAgent(agents_name)&& setTargetAgent(agents_name);
}


bool PlaceReachable::checkHumanReach()
{
    _WSF->setConfiguration(_confSol);
    global_Project->getActiveScene()->getFactsManager()->computeReachability(_WSF,_targetAgent,mainObject,true);

    fact_id FI3;
    FI3.factType = "IsReachable";
    FI3.factSubType = "GIK";
    FI3.isMainAnAgent = true;

    Fact F = _WSF->getFact(FI3, _targetAgent, mainObject);
//    F.printFact();
    return F.getBoolValue();
}


string PlaceReachable::getTextFromValues()
{
    return "Task: " + type
            + "\n    agents: "
            + "\n        mainAgent:    [AGENT ID]"
            + "\n        targetAgent:  [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject:   [OBJECT ID]"
            + "\n        supportObject [OBJECT ID] -optional-"
            + "\n    datas:"
            + "\n        hand: [right OR left]";
}
