#include "stackObj.hpp"
#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

using namespace std;


StackObj::StackObj() : Place()
{
    type = "stackObj";
    typeId = stackObj;
    _armId = -1;

    noSupport = false;

    approachTraj = NULL;
    placeTraj = NULL;
    escapeTraj = NULL;
    releaseTraj = NULL;

    nbMaxLoops = 30;

    SD->getSDMapEntry("statistics")->setMapEntry("nbMaxLoops", nbMaxLoops);
    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 1);
    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", nbMaxLoops);
    SD->setMapEntry("returnReport", "ERROR");

    _possibleMainAgentType.push_back(HRI_PR2);

}

bool StackObj::findCandidateSolutions()
{

    if (!checkAndSetArmAndObj())
        return false;
    //check if seted hand is the same as the one with the object

    if (!checkAttachement())
    {
        SD->setMapEntry("returnReport","no_corresponding_attachement_in_ws");
        return false;
    }

    if (!checkHRIAgent(mainAgent))
        return false;


    if (!setMainAgentHand())
        return false;


    //check support object steepness
    //get midlle of top of support object
    //trasnform point to conf;

//    if (!mainAgent->getHriAgent())
//    {
//        cout << "hri agent doesn't exist, unable to compute task" << endl;
//        return false;
//    }

//    if (!mainAgent->getHriAgent()->manip)
//    {
//        cout << "hri agent doesn't have manip data, unable to compute task" << endl;
//        return false;
//    }



//    for (int i = 0; i < mainAgent->getHriAgent()->manip->external_hand_no; i++)
//    {
//        if (!mainAgent->getHriAgent()->manip->external_hand &&
//                !mainAgent->getHriAgent()->manip->external_hand[i])
//            return false;
//        if (mainAgent->getHriAgent()->manip->external_hand[i]->hand_l == m_arm)
//        {

//            cout << "handName = " << mainAgent->getHriAgent()->manip->external_hand[i]->rob_hand->name << endl;
//            _hand = global_Project->getActiveScene()->getRobotByName(mainAgent->getHriAgent()->manip->external_hand[i]->rob_hand->name);
//            _armId = mainAgent->getHriAgent()->manip->external_hand[i]->hand_id;
//            break;
//        }
//    }
//    if (!_hand)
//    {

//        cout << "ERROR, the hand of the robot was not found" << endl;
//        return false;
//    }

    GeometricComputations GC = GeometricComputations();
//    _targetPointStc = GC.getTopCenterOfObj(_supportObject);


    vector<p3d_point> vpd = GC.getUpperPointsInObject(_supportObject, 0.03);

    if (vpd.size() !=4)
    {
        cout << "object steepness is too high nb of points = " << vpd.size()  << endl;
        return false;
    }

    _targetPointStc.x = (vpd.at(0).x + vpd.at(3).x)/2;
    _targetPointStc.y = (vpd.at(0).y + vpd.at(3).y)/2;
    _targetPointStc.z = (vpd.at(0).z + vpd.at(3).z)/2;


    if (!findRotations())
        return false;

//    if (!mainObject->getObjectRob() && !mainObject->getObjectRob()->isManipulable())
//    {
//        cout << "ERROR: " << mainObject->getName() << " has no rotations" << endl;
//        return false;
//    }
//    _rotations = mainObject->getObjectRob()->getStableRotations();
////    _supportObject->get

//    if (_rotations.size() == 0)
//    {
//        cout << "no rotations defined for the object : " << mainObject->getName() << endl;
//        return false;
//    }

    return true;
}

bool StackObj::getRandomSol()
{
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
    confPtr_t robotCurConf = confPtr_t(new Configuration(*mainAgent->getCurrentPos()));
    confPtr_t objectCurConf = confPtr_t(new Configuration(*mainObject->getCurrentPos()));

    detachObject();

    if (_rotations.size() <= 0)
    {
        cout << "ERROR: no rotation found" << endl;
        SD->setMapEntry("returnReport","no_rotation");
        return false;
    }
    configPt q = p3d_copy_config(mainObject->getRobotStruct(), _rotations.at(p3d_random_integer(0,_rotations.size()-1)));

    //        print_config(mainObject->getRobotStruct(),q);


    double randomRot = p3d_random(-M_PI,M_PI);

    if (_hasTargedPoint)
        randomRot = _targetPoint.z;

    q[6] = _targetPointStc.x + q[6];
    q[7] = _targetPointStc.y + q[7];
    q[8] = _targetPointStc.z + q[8];
    if (q[9] == OPEN_VALUE)
        q[9] = randomRot;
    else if (q[10] == OPEN_VALUE)
        q[10] = randomRot;
    else if (q[11] == OPEN_VALUE)
        q[11] = randomRot;


    p3d_set_and_update_this_robot_conf(mainObject->getRobotStruct(),q);



    if(isObjectInCollision(_supportObject))
    {
        mainAgent->setAndUpdate(*robotCurConf);
        mainObject->setAndUpdate(*objectCurConf);
        return false;
    }

    if(isHandInCollision())
    {
        mainAgent->setAndUpdate(*robotCurConf);
        mainObject->setAndUpdate(*objectCurConf);
        return false;
    }

//    if (!testFactConstraints())
//        continue;

    _confSol = confPtr_t(new Configuration(mainObject,q));



    mainAgent->setAndUpdate(*robotCurConf);
    mainObject->setAndUpdate(*objectCurConf);
//    attachObject();
    return true;










//    p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), _supportObject->getRobotStruct());
//    if(p3d_col_test_robot(mainObject->getRobotStruct(),0) == 1)//there is collision
//    {
//        cout << "colision at mainobj detection" << endl;
//                    p3d_print_col_pair();
//        //            sleep(1);
//        return false;
//    }
//    p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _supportObject->getRobotStruct());


//    //detecting colision between hand and env

//    p3d_matrix4 rot;
//    p3d_mat4Mult(mainObject->getRobotStruct()->joints[1]->abs_pos,_curGrasp.frame, rot);

//    double Tx, Ty, Tz;
//    double Rx, Ry, Rz;
//    p3d_mat4ExtractPosReverseOrder(rot, &Tx, &Ty, &Tz,
//                                   &Rx, &Ry, &Rz);


//    confPtr_t q_hand = _hand->getCurrentPos();


//    double Itx = (*q_hand)[6], Ity = (*q_hand)[7],  Itz = (*q_hand)[8], Irx = (*q_hand)[9], Iry = (*q_hand)[10], Irz = (*q_hand)[11];

//    (*q_hand)[6] = Tx; (*q_hand)[7] = Ty; (*q_hand)[8] = Tz;
//    (*q_hand)[9] = Rx; (*q_hand)[10] = Ry; (*q_hand)[11] = Rz; (*q_hand)[12] = 0.2;
//    _hand->setAndUpdate(*q_hand);


//    //test if hand is in collision with something else
//    p3d_col_deactivate_robot_robot(mainObject->getRobotStruct(), _hand->getRobotStruct());
//    bool colTest = p3d_col_test_robot(_hand->getRobotStruct(),0) == 1;
////        p3d_print_col_pair();
////        print_config(mainObject->getRobotStruct(),q);
////        g3d_draw_allwin_active();
////        sleep(3);
//    p3d_col_activate_robot_robot(mainObject->getRobotStruct(), _hand->getRobotStruct());


//    (*q_hand)[6] = Itx; (*q_hand)[7] = Ity; (*q_hand)[8] = Itz;
//    (*q_hand)[9] = Irx; (*q_hand)[10] = Iry; (*q_hand)[11] = Irz;
//    _hand->setAndUpdate(*q_hand);


//    if(colTest) // there is collision
//    {
//        cout << "collision detected for hand" << endl;
//        return false;
//    }


//    _confSol = confPtr_t(new Configuration(mainObject,q));


//    mainAgent->setAndUpdate(*robotCurConf);
//    mainObject->setAndUpdate(*objectCurConf);
//    attachObject();

//    return true;
}

bool StackObj::setObjects(multimap<string,string> objects_name)
{
    bool test = Place::setObjects(objects_name);
    if (!test)
        return false;

    if(!_supportObject)
        return false;

    return true;
}

string StackObj::getTextFromValues()
{
    return "Task: " + type + " : stacking objects"
            + "\n    agents: "
            + "\n        mainAgent:    [AGENT ID]"
            + "\n    objects: "
            + "\n        mainObject:   [OBJECT ID]"
            + "\n        supportObject [OBJECT ID]"
            + "\n    datas:"
            + "\n        hand: [right OR left]";
}
