#include "navigateTo.hpp"
#include "API/project.hpp"



#include "planner/plannerFunctions.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#include <jsoncpp/json/json.h>

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#include "planner/TrajectoryOptim/plannarTrajectorySmoothing.hpp"


using namespace std;

std::vector<p3d_point> currNavTraj;

vector<p3d_point> _vHP;
double _vMin,_vMax,_dMax;
p3d_point _p0;

static double distTraveledWeight, angleAndRotWeight, distToObjWeight, angleWeight, distHWeight, distWeight, ADistTraveledWeight, AAngleAndRotWeight;

NavigateTo::NavigateTo(): PrivateTask()
{
    type = "navigateto";
    typeId = navigateTo;
    nbMaxLoops = 50;
    SD->getSDMapEntry("statistics")->setMapEntry("nbMaxLoops", nbMaxLoops);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTotPossibleTests", nbMaxLoops);
    _idCurrentSol = -1;

    _possibleMainAgentType.push_back(HRI_PR2);
    _possibleMainAgentType.push_back(HRI_KUKAROBOT);

    _possibleMainAgentType.push_back(HRI_HERAKLES);
    mainAgent = NULL;
    _targetObject = NULL;
}

bool NavigateTo::initAll()
{
    result = new TaskAllSolutions(this);
    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 3);
    _armId = 0;
    _engage = false;
    _disengage = false;

    //in combineCosts
    distTraveledWeight = 1;
    angleAndRotWeight = 1;
    distToObjWeight = 1;
    //in ComputeCosts
    angleWeight = 1;
    distHWeight = 1;
    distWeight = 1;
    //in Compute A*
    ADistTraveledWeight = 1;
    AAngleAndRotWeight = 1;

    return true;
}

bool NavigateTo::initialize()
{    
    _squizTraj = NULL;
    _deployTraj = NULL;

    _disengagedConf.reset();
    _navigationConf.reset();
    _approachConf.reset();
    _manipulationConf.reset();
    _targetConf.reset();


    SD->getSDMapEntry("statistics")->setMapEntry("nbParameters", 1);
    SD->getSDMapEntry("statistics")->setMapEntry("nbGikCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbMPCalled", 0);
    SD->getSDMapEntry("statistics")->setMapEntry("nbTests", 0);
    SD->getSDMapEntry("times")->setMapEntry("timeMp", 0.0);
    SD->setMapEntry("returnReport", "ERROR");

    _returnReportcount = 0;
    _returnReportcountMax = 0;

    //remove col with object in hand
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        p3d_col_deactivate_robot_robot(att.agent->getRobotStruct(),att.object->getRobotStruct());
    }


    _vHP.clear();
    for (unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
    {
        if (global_Project->getActiveScene()->getRobot(i)->isAgent())
        {
            if (global_Project->getActiveScene()->getRobot(i)->getHriAgent()->is_human == 1)
            {
                p3d_point HP;
                confPtr_t q = global_Project->getActiveScene()->getRobot(i)->getCurrentPos();
                HP.x = (*q)[6];
                HP.y = (*q)[7];
                HP.z = (*q)[11];
//                cout << "HP ( " << HP.x << " , " << HP.y << " , " << HP.z << " )" << endl;
                _vHP.push_back(HP);
            }
        }
    }
    return  initMotionPlaner(mainAgent);
}

bool NavigateTo::finalize()
{
    //put back col with object in hand
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        p3d_col_activate_robot_robot(att.agent->getRobotStruct(),att.object->getRobotStruct());
    }

//    ENV.setBool(Env::drawTraj,false);
    return true;
}


bool NavigateTo::run(multimap<string,string> agents_name,
               multimap<string,string> objects_name,
               multimap<string,p3d_point> point3d,
               multimap<string,string> additionalData,
               WorldState *WS, bool computeMP)
{

    WS->load();
    _WS = WS;
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!_targetObject)
    {
        if (!setObjects(objects_name))
        {
            if (!setPoints(point3d))
            {
                cout << "ERROR: setObjects and setPoint fails for task " << type << " with id " << task_id << endl;
                return false;
            }
        }
    }
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setdata fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    bool res = false;
    int count = 0;
    while (!res)
    {
        cout << "count loop = " << count << endl;
        _returnReportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
            continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        if(!testConstraints(0))
            continue;

        res = true;
    }

    if (res)
        SD->setMapEntry("returnReport","OK");

    computeCost(0);
    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    return res;
}

double NavigateTo::computeCost(int altId)
{

    TaskSolution* TS = result->getAlternativeWithId(altId);
    if (!TS)
        return 1000;
    vector<TaskSubSolution*> vtss = TS->getTSSByRobot(mainAgent);

    vector<p3d_point> vp;
    for (unsigned int i = 0; i < vtss.size(); i++)
    {
        TaskSubSolution* TSS = vtss.at(i);
        if (TSS->getSubSolutionName().find("navigate") != string::npos)
        {
            vp = TSS->getNavVector();
        }

    }
    if (vp.size() == 0)
        return 1000;
    double cost = 0;




    GeometricComputations GC = GeometricComputations();
    TS->setCost(GC.computeCost(_vHP,vp, angleWeight,distHWeight,1));

    return cost;
}

double NavigateTo::combineCosts(double distTraveled, double angleAndRot, double distToObj)
{
    return (distTraveled*distTraveledWeight + angleAndRot*angleAndRotWeight + distToObj*distToObjWeight)/ (distTraveledWeight + angleAndRotWeight + distToObjWeight);
//    return angleAndRot;//(distTraveled + 3*angleAndRot + distToObj)/5;
}

bool NavigateTo::compareCells(GTP2DCell* i, GTP2DCell* j)
{

//    return (i->getDistCurrentRobot() < j->getDistCurrentRobot());
    switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
    {
    case 0:
    {
        i->setDrawn(true);
        j->setDrawn(true);
        double iDist, iCost, iCenter, jDist, jCost, jCenter;
        GeometricComputations GC = GeometricComputations();
        //distance from init
        iDist = i->getDistCurrentRobot()/_dMax;
        jDist = j->getDistCurrentRobot()/_dMax;

        double d = (_vMax + _vMin)/2;
        //angle and distance to humans
        vector<p3d_point> vp;
        p3d_point po;
        po.x = i->getCenter()[0];
        po.y = i->getCenter()[1];
        po.z = 0;
        vp.push_back(po);
        iCost = GC.computeCost(_vHP,vp,angleWeight,distHWeight,distWeight);

        iCenter = (fabs(GC.distBetweenPoints2D(_p0, po) - d))/((_vMax-_vMin)/2);

        vp.clear();
        po.x = j->getCenter()[0];
        po.y = j->getCenter()[1];
        po.z = 0;
        vp.push_back(po);
        jCost = GC.computeCost(_vHP,vp,angleWeight,distHWeight,distWeight);
        jCenter = (GC.distBetweenPoints2D(_p0, po) - d)/((_vMax-_vMin)/2);

        i->Ecolor = NavigateTo::combineCosts(iDist,iCost,iCenter);
        j->Ecolor = NavigateTo::combineCosts(jDist,jCost,jCenter);

        return i->Ecolor<j->Ecolor;
    }
    case 1:
        return (i->getDistCurrentRobot() < j->getDistCurrentRobot());
    default:
        return (i->getDistCurrentRobot() < j->getDistCurrentRobot());
    }

}

bool NavigateTo::findCandidateSolutions()
{

    //building the 2D grid
    std::vector<double> m_EnvSize;
    // initCostSpace
    m_EnvSize.resize(4);
    m_EnvSize[0] = XYZ_ENV->box.x1; m_EnvSize[1] = XYZ_ENV->box.x2;
    m_EnvSize[2] = XYZ_ENV->box.y1; m_EnvSize[3] = XYZ_ENV->box.y2;
    _grid = new GTP2DGrid(0.1,m_EnvSize);

    if (!_grid->initCylinders())
    {
        SD->setMapEntry("returnReport","no_cylinders");
        return false;
    }

    _disengagedConf = (confPtr_t) new Configuration(*_WS->getRobotConf(mainAgent));
    if (_disengage)
    {
        (*_disengagedConf)[6] = (*_disengagedConf)[6]- _distTodisengage*cos((*_disengagedConf)[11]);
        (*_disengagedConf)[7] = (*_disengagedConf)[7]- _distTodisengage*sin((*_disengagedConf)[11]);
        mainAgent->setAndUpdate(*_disengagedConf);
    }
    _grid->addAgentAndComputeNav(mainAgent);

    if (!isTargetAPoint)
    {
        cout << "target is an object or an agent" << endl;

        if(!_targetObject)
        {
            cout << "No target object/agent!" << endl;
            SD->setMapEntry("returnReport","no_target_agent_object");
            return false;
        }

        if (_targetObject->isAgent())
        {
            cout << "NavigateTo a target agent" << endl;
            _reachableCells = _grid->getCellsArroundAgent(mainAgent,_targetObject);
            if (_reachableCells.size() == 0)
            {
                cout << "the agent object " << _targetObject->getName() << " is not reachable by the robot" << endl;
                SD->setMapEntry("returnReport","not_reachable");
                return false;
            }
        }
        else
        {
            if(!_targetObject->getObjectRob())
            {
                cout<< _targetObject->getName() << " is not an object " << endl;
                SD->setMapEntry("returnReport","object_not_object");
                return false;
            }


            if (_targetObject->getObjectRob()->isManipulable())
            {
                cout << "NavigateTo a manipulable object" << endl;
                _reachableCells = _grid->getCellsArroundObject(mainAgent,_targetObject);
                if (_reachableCells.size() == 0)
                {
                    cout << "the manipulable object " << _targetObject->getName() << " is not reachable by the robot" << endl;
                    SD->setMapEntry("returnReport","not_reachable");
                    return false;
                }
            }
            else if (_targetObject->getObjectRob()->isSupport())
            {
                cout << "NavigateTo a support object" << endl;
                _reachableCells = _grid->getCellsArroundSupport(mainAgent,_targetObject);
                if (_reachableCells.size() == 0)
                {
                    cout << "the support " << _targetObject->getName() << " is not reachable by the robot" << endl;
                    return false;
                }
            }
        }



        _grid->setCurrentRobot(mainAgent);
        _vMin = 1000;
        _vMax = 0;
        _dMax = 0;

        confPtr_t q = _targetObject->getCurrentPos();
        _p0.x = (*q)[6];
        _p0.y = (*q)[7];
        _p0.z = 0;

        GeometricComputations GC = GeometricComputations();
        for (unsigned int i = 0; i<_reachableCells.size(); i++)
        {
            p3d_point p1;
            p1.x = _reachableCells.at(i)->getCenter()[0];
            p1.y = _reachableCells.at(i)->getCenter()[1];
            p1.z = 0;

            double distw = GC.distBetweenPoints2D(_p0,p1);
            if (_vMin > distw)
                _vMin = distw;
            else if (_vMax < distw)
                _vMax = distw;

            if (_dMax < _reachableCells.at(i)->getDistCurrentRobot())
                _dMax = _reachableCells.at(i)->getDistCurrentRobot();
        }




        std::sort(_reachableCells.begin(), _reachableCells.end(), NavigateTo::compareCells);

        _idCurrentSol = -1;
    }

    return true;
}

bool NavigateTo::getRandomSol()
{
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbTests");
    if (isTargetAPoint)
    {
        return true;
    }
    else
    {
        if (_reachableCells.size() == 0)
        {
            cout << "the agent cannot reach the object" << endl;
            return false;
        }

//        int id = p3d_random_integer(0,_reachableCells.size()-1);

        _idCurrentSol += 1;

        if (_reachableCells.size() <=_idCurrentSol )
        {
            return false;
        }


        _targetPoint.x = _reachableCells.at(_idCurrentSol)->getCenter()[0];
        _targetPoint.y = _reachableCells.at(_idCurrentSol)->getCenter()[1];
        _targetPoint.z = atan2((*_targetObject->getCurrentPos())[7] - _targetPoint.y, (*_targetObject->getCurrentPos())[6] - _targetPoint.x );

//        cout << "not yet implemented for objects" << endl;
        return true;
    }

}

bool NavigateTo::findConfigurations()
{
    bool ok = true;
    _WS->load();

    _disengagedConf = (confPtr_t) new Configuration(*_WS->getRobotConf(mainAgent));
    if (_disengage)
    {
        _disengagedConf->setAsNotTested();
        (*_disengagedConf)[6] = (*_disengagedConf)[6]- _distTodisengage*cos((*_disengagedConf)[11]);
        (*_disengagedConf)[7] = (*_disengagedConf)[7]- _distTodisengage*sin((*_disengagedConf)[11]);
        mainAgent->setAndUpdate(*_disengagedConf);
        _returnReportcount++;
        if (p3d_col_test_robot(mainAgent->getRobotStruct(),1))
        {
            ok = false;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","disengagedConf_in_col");
            }
        }
    }

    if (ok && _disengage)
        _grid->addAgentAndComputeNav(mainAgent);
    if (ok)
    {
        _navigationConf = getRandomArmPosFromFileBasedOnConf(DISENGAGECONF,_disengagedConf);
        _navigationConf->setAsNotTested();
        mainAgent->setAndUpdate(*_navigationConf);
        if (p3d_col_test_robot(mainAgent->getRobotStruct(),1))
        {
            ok = false;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","navigationConf_in_col");
            }
        }
    }


    if (ok && mainAgent->getHriAgent() && mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
    {
        double distMid = 0.3;
        _targetPoint.x = _targetPoint.x - distMid*cos(_targetPoint.z);
        _targetPoint.y = _targetPoint.y - distMid*sin(_targetPoint.z);
    }

    if (ok)
    {
        if (_engage)
        {
            _approachConf = (confPtr_t)new Configuration(*_navigationConf);
            _approachConf->setAsNotTested();
            (*_approachConf)[6] = _targetPoint.x- _distToDeploy*cos(_targetPoint.z);
            (*_approachConf)[7] = _targetPoint.y- _distToDeploy*sin(_targetPoint.z);
            (*_approachConf)[11] = _targetPoint.z;
            mainAgent->setAndUpdate(*_approachConf);
            _returnReportcount++;
            if (p3d_col_test_robot(mainAgent->getRobotStruct(),1))
            {
                ok = false;
                if (_returnReportcountMax < _returnReportcount)
                {
                    _returnReportcountMax = _returnReportcount;
                    SD->setMapEntry("returnReport","approachConf_in_col");
                }
            }
            if (ok)
            {
                _manipulationConf = getRandomArmPosFromFileBasedOnConf(ENGAGECONF,_approachConf);
                _manipulationConf->setAsNotTested();
                mainAgent->setAndUpdate(*_manipulationConf);
                _returnReportcount++;
                if (p3d_col_test_robot(mainAgent->getRobotStruct(),1))
                {
                    ok = false;
                    if (_returnReportcountMax < _returnReportcount)
                    {
                        _returnReportcountMax = _returnReportcount;
                        SD->setMapEntry("returnReport","manipulationConf_in_col");
                    }
                }
            }
            if (ok)
            {
                _targetConf = (confPtr_t)new Configuration(*_manipulationConf);
            }
        }
        else
        {
            _targetConf = (confPtr_t)new Configuration(*_navigationConf);
        }
        (*_targetConf)[6] = _targetPoint.x;
        (*_targetConf)[7] = _targetPoint.y;
        (*_targetConf)[11] = _targetPoint.z;
        _targetConf->setAsNotTested();
        mainAgent->setAndUpdate(*_targetConf);
        _returnReportcount++;
        if (p3d_col_test_robot(mainAgent->getRobotStruct(),1))
        {
            ok = false;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","targetConf_in_col");
            }
        }
    }

    return ok;
}

bool NavigateTo::findTrajectories()
{
    _WS->load();
    vector<SurfaceAttachement> sss = createAndLoadSurfaceAttachements(mainAgent);
    vector<Attachement> att =_WS->getAttachements();
    for (unsigned int i = 0; i < att.size(); i++)
    {
        vector<SurfaceAttachement> tmp = createAndLoadSurfaceAttachements(att.at(i).object);
        sss.insert(sss.end(), tmp.begin(),tmp.end());
    }
    vector<SurfaceAttachement> tmp = createAndLoadSurfaceAttachements(mainAgent);
    sss.insert(sss.end(), tmp.begin(),tmp.end());
    bool ok = true;

    // if disengage enabled, compute the traj
    if (_disengage)
    {
        int status = -1;
        confPtr_t q_init = _WS->getRobotConf(mainAgent);
        compute_traj_using_localPath(q_init,_disengagedConf,_targetObject,status);
        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND)
        {
            ok = false;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","disengageTraj_not_found");
            }
        }
        else
        {
            p3d_point initP,disengageP;
            initP.x = (*q_init)[6]; initP.y = (*q_init)[7]; initP.z = (*q_init)[11];
            disengageP.x = (*_disengagedConf)[6]; disengageP.y = (*_disengagedConf)[7]; disengageP.z = (*_disengagedConf)[11];
            _disengageTraj.clear();
            _disengageTraj.push_back(initP);
            _disengageTraj.push_back(disengageP);
        }
    }

    // set arm to navigation
    if (ok)
    {
        setArmToUse(Right);
        int status = -1;
        if (mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
            _squizTraj = compute_traj(_disengagedConf, _navigationConf, _targetObject, status);
        else
            _squizTraj = compute_traj_using_localPath(_disengagedConf, _navigationConf, _targetObject, status);

        _returnReportcount++;
        if (status != PLAN_STATUS_START_EQUALS_GOAL && status != PLAN_STATUS_FOUND)
        {
            cout << "no squiz traj" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","no_squiz_traj");
            }
            ok = false;
        }
        else if (status == PLAN_STATUS_START_EQUALS_GOAL)
        {
            _squizTraj = NULL;
        }
    }

    confPtr_t endingConf;
    if (ok)
    {
        mainAgent->setAndUpdate(*_disengagedConf);
        if (_engage)
            endingConf = _approachConf;
        else
            endingConf = _targetConf;
    }



    //looking for the target cell
    GTP2DCell* cell;
    Eigen::Vector2d endPos;
    if (ok)
    {

        endPos[0] = (*endingConf)[6];
        endPos[1] = (*endingConf)[7];
        cell = dynamic_cast<GTP2DCell*>(_grid->getCell(endPos));
        _returnReportcount++;
        if (!cell)
        {
            cout << "ERROR, no cell found for position " << endPos[0] << ", " << endPos[1] << endl;
            ok = false;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","cell_not_found");
            }
        }
    }

    vector<GTP2DCell*> path;
    if (ok)
    {
//        angleWeight = 1;
//        distHWeight = 1;
//        distWeight = 1;
//        //in Compute A*
//        ADistTraveledWeight = 1;
//        AAngleAndRotWeight = 1;

        SD->getSDMapEntry("statistics")->incrementMapEntry("nbMPCalled");

        _returnReportcount++;
        switch(PlanEnv->getInt(PlanParam::GTPTypeCase))
        {
        case 0:
            path = _grid->getHumanAwarePath(mainAgent,cell,ADistTraveledWeight,AAngleAndRotWeight,angleWeight,distHWeight,distWeight);
            break;
        case 1:
            path = _grid->getPath(mainAgent,cell);
            break;
        default:
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","typeCase_not_correct");
            }
            ok = false;
            break;
        }
    }

    if (ok)
    {
        PlannarTrajectorySmoothing PTS(mainAgent);
        std::vector<Eigen::Vector2d,Eigen::aligned_allocator<Eigen::Vector2d> > vp;

        confPtr_t q_cur = _WS->getRobotConf(mainAgent);

        Eigen::Vector2d vinit;
        vinit[0] = (*q_cur)[6];
        vinit[1] = (*q_cur)[7];

        vp.push_back(vinit);

        for(unsigned int i = path.size(); i > 0; i--)
        {
            Eigen::Vector2d vtmp;
            vtmp[0] = path.at(i-1)->getCenter()[0];
            vtmp[1] = path.at(i-1)->getCenter()[1];
            vp.push_back(vtmp);
        }

        vp.push_back(endPos);

        std::vector<Eigen::Vector3d,Eigen::aligned_allocator<Eigen::Vector3d> > vp3 = PTS.add3DimwithoutTrajChange(vp, mainAgent, 0.05);

        _traj.clear();
        for (unsigned int j = 0; j < vp3.size(); j++)
        {
            p3d_point ptmp;
            ptmp.x= vp3.at(j)[0];
            ptmp.y= vp3.at(j)[1];
            ptmp.z= vp3.at(j)[2];
            _traj.push_back(ptmp);
        }

        p3d_point vend;
        vend.x = (*endingConf)[6];
        vend.y = (*endingConf)[7];
        vend.z = (*endingConf)[11];
        _traj.push_back(vend);
        currNavTraj = _traj;

        //computing dist and switching to p3d_points
        int trajSiz = 0;
        GeometricComputations GC = GeometricComputations();
        for (unsigned int i = 1; i < _traj.size(); i++)
        {
            trajSiz += GC.distBetweenPoints2D(_traj.at(i-1),_traj.at(i));
        }
        SD->getSDMapEntry("graph")->setMapEntry("nbGraphNode",(int)_grid->getNbExploredCells());
        SD->getSDMapEntry("graph")->setMapEntry("nbComputedNodes",(int)_traj.size());
        SD->getSDMapEntry("graph")->setMapEntry("trajSize",trajSiz);
    }



    if (ok && _engage)
    {
        int status = -1;
        if (mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
            _deployTraj = compute_traj(_approachConf, _manipulationConf, _targetObject, status);
        else
            _deployTraj = compute_traj_using_localPath(_approachConf, _manipulationConf, _targetObject, status);

        _returnReportcount++;
        if (status != PLAN_STATUS_START_EQUALS_GOAL && status != PLAN_STATUS_FOUND)
        {
            cout << "no deploy traj" << endl;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","no_deploy_traj");
            }
            ok = false;
        }
        else if (status == PLAN_STATUS_START_EQUALS_GOAL)
        {
            _deployTraj = NULL;
        }
    }


    if (ok && _engage)
    {
        int status = -1;
        compute_traj_using_localPath(_manipulationConf,_targetConf,_targetObject,status);
        _returnReportcount++;
        if (status != PLAN_STATUS_FOUND)
        {
            ok = false;
            if (_returnReportcountMax < _returnReportcount)
            {
                _returnReportcountMax = _returnReportcount;
                SD->setMapEntry("returnReport","disengageTraj_not_found");
            }
        }
        else
        {
            p3d_point initP,disengageP;
            initP.x = (*_manipulationConf)[6]; initP.y = (*_manipulationConf)[7]; initP.z = (*_manipulationConf)[11];
            disengageP.x = (*_targetConf)[6]; disengageP.y = (*_targetConf)[7]; disengageP.z = (*_targetConf)[11];
            _engageTraj.clear();
            _engageTraj.push_back(initP);
            _engageTraj.push_back(disengageP);
        }
    }
    return ok;
}

bool NavigateTo::computeSolutionTrajectories(int alternativeId)
{
    bool res =true;
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if(!TS)
    {
        cout << "no alternative for task " << to_string(task_id) << " with alternative id = " << alternativeId << endl;
        res = false;
    }
    if(res){
        confPtr_t q = TS->getStopWS()->getRobotConf(mainAgent);
        _targetPoint.x = (*q)[6];
        _targetPoint.y = (*q)[7];
        _targetPoint.z = (*q)[11];

        res = findTrajectories();
    }
    return res;
}



bool NavigateTo::setToSolution(int solutionId, bool computeMP)
{
    TaskSolution* TS = new TaskSolution(result,solutionId);

    TS->setTrajComputed(computeMP);

    TS->setTaskConstraints(_currConstraints);

    int index = 0;

    TaskSubSolution* TSS0,*TSS1,*TSS2,*TSS3,*TSS4;

    if (_disengage)
    {
        TSS0 = new TaskSubSolution(TS, "disengage_navigate", index++, NULL, mainAgent, 0);
        cout << "index" << index << endl;
        TSS0->setNavigation(true);
        TSS0->setNavTraj(_disengageTraj);
    }

    TSS1 = new TaskSubSolution(TS, "squiz", index++, _squizTraj, mainAgent, 0);
    TSS1->setArmId(2);

    if (_disengage)
    {
        vector<TaskSubSolution*> v1;
        v1.push_back(TSS1);
        TSS0->setNext(v1);
    }

    TSS2 = new TaskSubSolution(TS, "navigate", index++, NULL, mainAgent, 0);
    TSS2->setNavigation(true);
    TSS2->setNavTraj(_traj);

    vector<TaskSubSolution*> v2;
    v2.push_back(TSS2);
    TSS1->setNext(v2);

    if (_engage)
    {
         TSS3 = new TaskSubSolution(TS, "deploy", index++, _deployTraj, mainAgent, 0);
         TSS3->setArmId(2);

         vector<TaskSubSolution*> v3;
         v3.push_back(TSS3);
         TSS2->setNext(v3);

         TSS4 = new TaskSubSolution(TS, "engage_navigate", index++, NULL, mainAgent, 0);
         TSS4->setNavigation(true);
         TSS4->setNavTraj(_engageTraj);

         vector<TaskSubSolution*> v4;
         v4.push_back(TSS4);
         TSS3->setNext(v4);
    }

    TS->setStartWS(_WS);

    _WS->load();
    vector<Attachement> atts;
    vector<SurfaceAttachement> sss;
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        if (att.agent == mainAgent)
        {
            PrivateTask::attachObject(mainAgent, att.object, true ,att.armId);
            atts.push_back(att);
            vector<SurfaceAttachement> tmp = createAndLoadSurfaceAttachements(att.object);
            sss.insert(sss.end(), tmp.begin(), tmp.end());
        }
    }
    vector<SurfaceAttachement> tmp = createAndLoadSurfaceAttachements(mainAgent);
    sss.insert(sss.end(), tmp.begin(), tmp.end());
    _WS->setSurfaceAttachements(sss);

    _WS->load();
    WorldState *startWS = _WS;

    //---------------------------------------------
    if (_disengage)
    {
        TSS0->setStartState(_WS);
        WorldState* WS1 = new WorldState(_WS);
        WS1->replaceConf(_disengagedConf);

        mainAgent->setAndUpdate(*_disengagedConf);
        for (unsigned int i = 0; i < atts.size(); i++)
            WS1->replaceConf(atts.at(i).object->getCurrentPos());
        for (unsigned int i = 0; i < sss.size(); i++)
            WS1->replaceConf(sss.at(i).attache->getCurrentPos());

        WS1->setSurfaceAttachements(sss);

        TSS0->setEndState(WS1);

        WS1->load();
        startWS = WS1;
    }

    TSS1->setStartState(startWS);
    //---------------------------------------------

    WorldState* WS2 = new WorldState(startWS);
    WS2->replaceConf(_navigationConf);

    mainAgent->setAndUpdate(*_navigationConf);
    for (unsigned int i = 0; i < atts.size(); i++)
        WS2->replaceConf(atts.at(i).object->getCurrentPos());
    for (unsigned int i = 0; i < sss.size(); i++)
        WS2->replaceConf(sss.at(i).attache->getCurrentPos());

    WS2->setSurfaceAttachements(sss);

    TSS1->setEndState(WS2);
    TSS2->setStartState(WS2);

    confPtr_t ws3EndingConf;
    if (_engage)
        ws3EndingConf = _approachConf;
    else
        ws3EndingConf = _targetConf;
    //---------------------------------------------
    WS2->load();
    WorldState* WS3 = new WorldState(WS2);
    WS3->replaceConf(ws3EndingConf);

    mainAgent->setAndUpdate(*ws3EndingConf);
    for (unsigned int i = 0; i < atts.size(); i++)
        WS3->replaceConf(atts.at(i).object->getCurrentPos());
    for (unsigned int i = 0; i < sss.size(); i++)
        WS3->replaceConf(sss.at(i).attache->getCurrentPos());


    WS3->setSurfaceAttachements(sss);
    TSS2->setEndState(WS3);

    WorldState* WSEnd = WS3;

    if (_engage)
    {
        TSS3->setStartState(WS3);

        //---------------------------------------------
        WS3->load();
        WorldState* WS4 = new WorldState(WS3);
        WS4->replaceConf(_manipulationConf);

        mainAgent->setAndUpdate(*_manipulationConf);
        for (unsigned int i = 0; i < atts.size(); i++)
            WS4->replaceConf(atts.at(i).object->getCurrentPos());
        for (unsigned int i = 0; i < sss.size(); i++)
            WS4->replaceConf(sss.at(i).attache->getCurrentPos());

        WS4->setSurfaceAttachements(sss);

        TSS3->setEndState(WS4);
        TSS4->setStartState(WS4);


        //---------------------------------------------
        WS4->load();
        WorldState* WS5 = new WorldState(WS4);
        WS5->replaceConf(_targetConf);

        mainAgent->setAndUpdate(*_targetConf);
        for (unsigned int i = 0; i < atts.size(); i++)
            WS5->replaceConf(atts.at(i).object->getCurrentPos());
        for (unsigned int i = 0; i < sss.size(); i++)
            WS5->replaceConf(sss.at(i).attache->getCurrentPos());

        WS5->setSurfaceAttachements(sss);

        TSS4->setEndState(WS5);
        WSEnd = WS5;
    }

    WSEnd->setTCForNextTask(_currConstraints);

    TS->setStopWS(WSEnd);
    WSEnd->setTSThatEndWithThisWS(TS);

    return true;
}


bool NavigateTo::setAgents(multimap<string,string> agents_name)
{
    if (agents_name.count("targetAgent") == 1)
    {
        _targetObject = global_Project->getActiveScene()->getRobotByName(agents_name.find("targetAgent")->second);
        if (!_targetObject)
        {
            cout << "error in target agent name :" << agents_name.find("targetAgent")->second << endl;
            SD->setMapEntry("returnReport","error_targetagent_name");
            return false;
        }
        isTargetAPoint = false;
    }
    return setMainAgent(agents_name);
}

bool NavigateTo::setObjects(multimap<string,string> objects_name)
{
    if (objects_name.count("target") != 1)
    {
        cout << "one and only one target at a time" << endl;
        return false;
    }
    _targetObject = global_Project->getActiveScene()->getRobotByName(objects_name.find("target")->second);
    if (!_targetObject)
        return false;
    isTargetAPoint = false;
    return true;
}

bool NavigateTo::setPoints(multimap<string,p3d_point> points)
{
    if (points.count("target") != 1)
    {
        cout << "one and only one target at a time" << endl;
        return false;
    }
    _targetPoint = points.find("target")->second;
    isTargetAPoint = true;
    return true;
}

bool NavigateTo::setdata(multimap<string,string> additionalData)
{
    bool ok = true;
    int countCost = additionalData.count("cost");
    string costStr = "";
    if (countCost > 1)
    {
        SD->setMapEntry("returnReport","too_many_costs_in_map");
        ok = false;
    }
    else if (countCost == 0)
    {
        cout << "using default costs: 1 10 0 2 1 0 0 1" << endl;
        costStr = "1 10 0 2 1 0 0 1";
    }
    else if (countCost == 1)
        costStr = additionalData.find("cost")->second;

    std::vector<std::string> tokens;
    if (ok)
    {
        std::istringstream buf(costStr);
        std::istream_iterator<std::string> beg(buf), end;

        tokens = std::vector<std::string>(beg, end);

        if (tokens.size() != 8)
        {
            SD->setMapEntry("returnReport","wrong_nb_of_costs");
            ok =false;
        }
    }

    if (ok)
    {
        //in combineCosts
        distTraveledWeight = atof(tokens.at(0).c_str());
        angleAndRotWeight = atof(tokens.at(1).c_str());
        distToObjWeight = atof(tokens.at(2).c_str());
        //in ComputeCosts
        angleWeight = atof(tokens.at(3).c_str());
        distHWeight = atof(tokens.at(4).c_str());
        distWeight = atof(tokens.at(5).c_str());
        //in Compute A*
        ADistTraveledWeight = atof(tokens.at(6).c_str());
        AAngleAndRotWeight = atof(tokens.at(7).c_str());

        //    for(unsigned int i = 0; i< tokens.size(); i++)
        //        std::cout << '"' << tokens.at(i) << '"' << '\n';

        //    cout <<distTraveledWeight << endl;
        //    cout <<angleAndRotWeight << endl;
        //    cout <<distToObjWeight << endl;
        //    cout <<angleWeight << endl;
        //    cout <<distHWeight << endl;
        //    cout <<distWeight << endl;
        //    cout <<ADistTraveledWeight << endl;
        //    cout <<AAngleAndRotWeight << endl;

    }

    if (ok)
    {
        if (additionalData.count("distance"))
        {
            string dist = additionalData.find("distance")->second;
            _distToDeploy = atof(dist.c_str());

        }
        else
        {
            cout << "setting distance to deployement to 0.3m as default" << endl;
            _distToDeploy = 0.3;
        }
        //TODO; have different values
        _distTodisengage = _distToDeploy;
    }

    if (ok)
    {
        if (additionalData.count("engage") == 1)
        {
            _engage = true;
        }
        if (additionalData.count("disengage") == 1)
        {
            _disengage = true;
        }

    }

    return ok;
}

void NavigateTo::smoothSolution(int alternativeId)
{

    if (mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
    {
        for (int i = 0; i < result->getAlternativeWithId(alternativeId)->getNbsubSolPerRobot(getMainAgent()); i++)
        {
            TaskSubSolution* tss = result->getAlternativeWithId(alternativeId)->getTSSByRobot(getMainAgent()).at(i);
            if (tss->getSubSolutionName() == "extract")
            {
                tss->getRobot()->getRobotStruct()->tcur = tss->getTraj();
                tss->getStartState()->load();
                tss->refineTraj();
                break;
            }
        }
        result->getAlternativeWithId(alternativeId)->getStartWS()->load();
    }

    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    if (!TS)
        return;
    vector<TaskSubSolution*> vtss = TS->getTSSByRobot(mainAgent);
    TaskSubSolution* TSS = NULL;
    for (unsigned int i = 0; i < vtss.size(); i++)
    {
        TSS = vtss.at(i);
        if (TSS->getSubSolutionName().find("navigate") != string::npos)
        {
            break;
        }
    }
    if (!TSS)
        return;

    vector<p3d_point> vp3d = TSS->getNavVector();

    GeometricComputations GC = GeometricComputations();

    Robot *cyl;
//    r->getHriAgent()->type
    if (mainAgent->getHriAgent()->type == HRI_HERAKLES)
        cyl = global_Project->getActiveScene()->getRobotByName("HUMCYLINDER");
    else if (mainAgent->getHriAgent()->type == HRI_PR2 || mainAgent->getHriAgent()->type == HRI_KUKAROBOT )
        cyl = global_Project->getActiveScene()->getRobotByName("PR_2CYLINDER");

    p3d_col_deactivate_robot_robot(cyl->getRobotStruct(),mainAgent->getRobotStruct());
    _traj = GC.smoothTraj(vp3d,_vHP,cyl,500,-1,distWeight,distHWeight,angleWeight);

    currNavTraj = _traj;

    TSS->setNavTraj(_traj);

}

int NavigateTo::checkHandAttachement()
{
    int res = 0;
    bool rarm = false;
    bool larm = false;
    for (unsigned int i = 0; i < _WS->getAttachements().size(); i++)
    {
        Attachement att = _WS->getAttachements().at(i);
        if (att.agent->getName() == mainAgent->getName())
        {
            if (att.armId == 0)
                rarm = true;
            else if (att.armId == 1)
                larm = true;
        }
    }
    if (rarm && larm)
        res = 3;
    else if (rarm)
        res = 1;
    else if (larm)
        res = 2;
    return res;
}

string NavigateTo::getDescr()
{
    string s= "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type;
    if (mainAgent)
        s+= "\nMain agent \t\t= " + mainAgent->getName();
    if (_targetObject)
        s+= "\nMain object \t= " + _targetObject->getName();

    return s;
}

string NavigateTo::getTextFromValues()
{
    return "Task: " + type + " : navigate to target (either point or object)"
            + "\n    agents: "
            + "\n        mainAgent: [AGENT ID]"
            + "\n    objects: "
            + "\n        target: [OBJECT ID]"
            + "\n    points:"
            + "\n        target: [X Y THETA]";
}


confPtr_t NavigateTo::getRandomArmPosFromFileBasedOnConf(navigateToFileType fileType, confPtr_t q_tmp)
{
    bool ok=true;
    string path = string(getenv("HOME_MOVE3D")) + "/Configurations/navigateTo/";

    if (fileType == ENGAGECONF)
        path += "engage/";
    else if (fileType == DISENGAGECONF)
        path += "disengage/";
    else
    {
        ok = false;
        SD->setMapEntry("returnReport","file_unknown");
    }
    path+= mainAgent->getName();

    std::ifstream ifs;
    Json::Value root;


    if (ok)
        ifs.open(path.c_str());

    if(ok && !ifs){
        ok=false;
        cout << "could not open "<< path << endl;
        SD->setMapEntry("returnReport","file_does_not_exist");
    }
    if(ok){
        try{
            ifs>>root;
        }catch (std::runtime_error e){
            cout << "error while parsing file "<<path<<": "<<e.what() << endl;
            ok=false;
            SD->setMapEntry("returnReport","parsing_error1");
        }catch(std::exception e){
            cout << "error while parsing file "<<path<<": "<<e.what() << endl;
            ok=false;
            SD->setMapEntry("returnReport","parsing_error2");
        }
    }
    if(ok){
        if(!root){
            cout << "could not parse json file "<<path << endl;
            ok=false;
            SD->setMapEntry("returnReport","bad_json_format");
        }
        if (root.size() <= 0){
            cout << "file "<< path << "is empty" << endl;
            ok=false;
            SD->setMapEntry("returnReport","empty_file");
        }
    }

    confPtr_t q = (confPtr_t) new Configuration(*q_tmp);
    if (ok)
    {
        int t = p3d_random_integer(0, root.size()-1);

        std::vector<std::string> vdr = root.getMemberNames();
        Json::Value v = root[vdr.at(t)];
//        cout << v.toStyledString() << endl;
        vector<string> membersName = v.getMemberNames();
        map<string, vector<double> > confParts;
        for (unsigned int i = 0; i < membersName.size(); i++)
        {
            v[membersName.at(i)];
            vector<double> vectD;
            for(Json::Value::iterator it = v[membersName.at(i)].begin(); it !=v[membersName.at(i)].end(); ++it)
            {
                vectD.push_back((*it).asDouble());
            }
            confParts[membersName.at(i)] = vectD;
        }


        int rArm = -1;
        int lArm = -1;
        int head = -1;
        unsigned int nbArmDof = -1;
        if (mainAgent->getHriAgent()->type == HRI_HERAKLES)
        {
            rArm = 18;
            lArm = 27;
            head = 15;
            nbArmDof = 7;
        }
        else if (mainAgent->getHriAgent()->type == HRI_PR2 )
        {
            rArm = 16;
            lArm = 25;
            head = 13;
            nbArmDof = 7;
        }
        else if (mainAgent->getHriAgent()->type == HRI_KUKAROBOT)
        {
            rArm = 12;
            nbArmDof = 7;
        }
        for (map<string, vector<double> >::iterator it=confParts.begin(); it!=confParts.end(); ++it)
        {
            if (it->second.size() != nbArmDof)
            {
                SD->setMapEntry("returnReport", "bad_nb_dofs_input");
                ok = false;
                q.reset();
                break;
            }
            int c = -1;
            if (it->first == "RARM")
                c = rArm;
            else if (it->first == "LARM")
                c = lArm;
            else if (it->first == "HEAD")
                c = head;
            else
                break;
            if (c == -1)
            {
                cout << "wrong assignation of arm/head to the wrong robot" << endl;
                q.reset();
                break;
            }

            for (unsigned int i = 0; i < it->second.size(); i++)
                (*q)[i+c] = it->second.at(i)*M_PI/180;


        }

    }
    return q;
}




