#include "math.h"
#include "fixOnWall.hpp"
#include "API/project.hpp"
#include "GTP/GTPTools/worldState.hpp"
#include "GTP/GTPTools/grid.hpp"

#include "planner/plannerFunctions.hpp"

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif
using namespace std;


//####################### Point ########################

fixOnWall::fixOnWall() : ManipulationTask()
{
    type = "fixOnWall";
//    _armId = -1;
    nbMaxLoops = 50;

    //_possibleMainAgentType.push_back("PR2_ROBOT");
    //_possibleMainAgentType.push_back("MAX_ROBOT");
    //_possibleMainAgentType.push_back("BOB_ROBOT");
    _possibleMainAgentType.push_back(HRI_PR2);

}

bool fixOnWall::initAll()
{
    result = new TaskAllSolutions(this);
    currMain = NULL;
    currTarget = NULL;
    return true;
}

bool fixOnWall::run(multimap<string,string> agents_name,
                           multimap<string,string> objects_name,
                           multimap<string,p3d_point> point3d,
                           multimap<string,string> additionalData,
                           WorldState *WS)
{

    WS->load();
    _WS = WS;
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if(!setObjects(objects_name))
    {
        cout << "ERROR: no target specified for task " << type << " with id " << task_id << endl;
        return false;

    }
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (!setdata(additionalData))
    {
        cout << "ERROR: no hand specified " << type << " with id " << task_id << endl;
        return false;
    }

    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    bool res = false;
    int is = 0;
//    double currCost = 0;
    while (is++ < 5)
    {
        for (int i = 0; i < 100; i++)
        {
            cout << "count loop = " << i << " in test " << is << endl;
            if(!getRandomSol())
               continue;
            if(!findConfigurations())
                continue;
//            double cTmp = getConfigCost();
//            if(cTmp >currCost)
//                replaceSol();
        }
        if(findTrajectories())
            break;
    }




//    bool res = false;
//    int count = 0;
//    while (!res)
//    {
//        cout << "count loop = " << count << endl;
//        if(count++ > 50)
//            break;
//        if(!getRandomSol())
//           continue;
//        if(!findConfigurations())
//            continue;
//        if(!findTrajectories())
//            continue;
//        res = true;
//    }
    return res;
}

bool fixOnWall::findCandidateSolutions()
{
    std::vector<double> m_EnvSize;
    // initCostSpace
    m_EnvSize.resize(4);
    m_EnvSize[0] = XYZ_ENV->box.x1; m_EnvSize[1] = XYZ_ENV->box.x2;
    m_EnvSize[2] = XYZ_ENV->box.y1; m_EnvSize[3] = XYZ_ENV->box.y2;
    _grid = new GTP2DGrid(0.05,m_EnvSize);
    _grid->addAgentAndComputeNav(mainAgent);
    _grid->addAgentAndComputeNav(targetAgent);

    p3d_oriented_BB *OBB = p3d_get_oriented_BB_obj(mainObject->getJoint(1)->getJointStruct()->o);

    _target1 = OBB->BottomFrontLeft;
    _target2 = OBB->BottomFrontRight;
    p3d_destroy_oriented_BB(OBB);



    mainReach = _grid->getCellsWithinReach(_target1.x,_target1.y, mainAgent->getHriAgent()->manip->reach_tolerance, mainAgent);
    targetReach = _grid->getCellsWithinReach(_target2.x,_target2.y, targetAgent->getHriAgent()->manip->reach_tolerance, targetAgent);
    return true;
}

bool fixOnWall::getRandomSol()
{

    //tirer une solution aleatoire dans les reach, qui ne se marche pas dessus!!
    double dist =0;
    while (dist < 0.6)
    {
        if (currMain && currTarget)
        {
            pair<GTP2DCell*,GTP2DCell*> p;
            p.first = currMain;
            p.second = currTarget;
            failedPairs.push_back(p);
        }
        currMain = mainReach.at(p3d_random_integer(0,mainReach.size()-1));
        currTarget = targetReach.at(p3d_random_integer(0,targetReach.size()-1));

        dist = currMain->dist(currTarget);

        for (unsigned int i = 0; i < failedPairs.size(); i++)
        {
            if (failedPairs.at(i).first->getIndex() != currMain->getIndex())
                continue;
            if (failedPairs.at(i).second->getIndex() != currTarget->getIndex())
                continue;
            dist =0;
            break;
        }
    }

    return true;
}

bool fixOnWall::findConfigurations()
{
    //utiliser freeholding conf
//    currMain->print();
//    currTarget->print();

    bool res = true;
    confPtr_t q_main_curr = confPtr_t(new Configuration(*mainAgent->getCurrentPos()));
    confPtr_t q_target_curr = confPtr_t(new Configuration(*targetAgent->getCurrentPos()));

    confPtr_t q_main = mainAgent->getCurrentPos();
    confPtr_t q_target = targetAgent->getCurrentPos();

    (*q_main)[6] = currMain->getCenter()[0];
    (*q_main)[7] = currMain->getCenter()[1];
    (*q_main)[11] = M_PI + atan2(currMain->getCenter()[1]- _target1.y, currMain->getCenter()[0]- _target1.x);
    mainAgent->setAndUpdate(*q_main);


    (*q_target)[6] = currTarget->getCenter()[0];
    (*q_target)[7] = currTarget->getCenter()[1];
    (*q_target)[11] = M_PI + atan2(currTarget->getCenter()[1]- _target2.y, currTarget->getCenter()[0]- _target2.x);
    targetAgent->setAndUpdate(*q_target);

    if (p3d_col_test_robot(mainAgent->getRobotStruct(),0) != 0)
        res = false;
    if (p3d_col_test_robot(targetAgent->getRobotStruct(),0) != 0)
        res = false;

    if (!res)
    {
        mainAgent->setAndUpdate(*q_main_curr);
        targetAgent->setAndUpdate(*q_target_curr);
        return false;
    }


    /*ManipulationPlanner* mp = _MP->getMotionPlanner(mainAgent->getName());*/


    int armId = getArmId();
    ArmManipulationData& armData = (*mainAgent->getRobotStruct()->armManipulationData)[armId];
    ManipulationConfigs manipConf(mainAgent->getRobotStruct());
    gpGrasp grasp;
    double confCost = -1;
    vector<double> target(6);
    target.at(0) = _target1.x;
    target.at(1) = _target1.y;
    target.at(2) = _target1.z;
    target.at(3) = P3D_HUGE;
    target.at(4) = P3D_HUGE;
    target.at(5) = P3D_HUGE;

    p3d_col_deactivate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());
    configPt q = NULL;
    q = manipConf.getFreeHoldingConf(NULL, armId, grasp, armData.getCcCntrt()->Tatt, confCost, target, NULL);
    if (!q)
        res = false;
    else
    {
        mainEndConf = confPtr_t(new Configuration(mainAgent,q));
        mainAgent->setAndUpdate(*mainEndConf);

        if (p3d_col_test_robot(mainAgent->getRobotStruct(),0) != 0)
            res = false;

    }
    p3d_col_activate_robot_robot(mainAgent->getRobotStruct(), mainObject->getRobotStruct());


    if (!res)
    {
        mainAgent->setAndUpdate(*q_main_curr);
        targetAgent->setAndUpdate(*q_target_curr);
        return false;
    }



    HRI_GIK_TASK_TYPE task;
    if (armId == 0){
      task = GIK_RATREACH;
    }else{
      task = GIK_LATREACH;
    }

    p3d_vector3 Tcoord;

    Tcoord[0] = _target2.x;
    Tcoord[1] = _target2.y;
    Tcoord[2] = _target2.z;
    double distance_tolerance = 0.2;
    configPt q2 = p3d_get_robot_config(targetAgent->getRobotStruct());

    /*int IKSucceded = */hri_agent_single_task_manip_move(targetAgent->getHriAgent(), task, &Tcoord, distance_tolerance, &q2);

////    _MP->getMotionPlanner(targetAgent->getName());
//    ArmManipulationData& armData2 = (*targetAgent->getRobotStruct()->armManipulationData)[armId];
//    ManipulationConfigs manipConf2(targetAgent->getRobotStruct());
//    gpGrasp grasp2;
//    double confCost2 = -1;
//    target.at(0) = _target2.x;
//    target.at(1) = _target2.y;
//    target.at(2) = _target2.z;
//    target.at(3) = P3D_HUGE;
//    target.at(4) = P3D_HUGE;
//    target.at(5) = P3D_HUGE;

//    configPt q2 = NULL;
//    q2 = manipConf2.getFreeHoldingConf(NULL, armId, grasp2, armData2.getCcCntrt()->Tatt, confCost2, target, NULL);
    if(!q2)
        res = false;
    else
    {
        targetEndConf = confPtr_t(new Configuration(targetAgent,q2));
        targetAgent->setAndUpdate(*targetEndConf);
        p3d_col_deactivate_robot_robot(targetAgent->getRobotStruct(), mainObject->getRobotStruct());
        if (pqp_robot_all_no_self_collision_test(targetAgent->getRobotStruct()) != 0)
        {
            p3d_print_col_pair();
            res = false;
        }
        p3d_col_activate_robot_robot(targetAgent->getRobotStruct(), mainObject->getRobotStruct());
    }




    mainAgent->setAndUpdate(*q_main_curr);
    targetAgent->setAndUpdate(*q_target_curr);
    return res;
}


bool fixOnWall::findTrajectories()
{

//    result.push_back(TSS1);
    return true;
}

bool fixOnWall::computeSolutionTrajectories(TaskSolution *TS)
{
    TS = 0;
    return true;
}

TaskSolution *fixOnWall::createSolution(bool computeMP)
{
    p3d_traj* t = NULL;
    TaskSolution* TS = new TaskSolution();
    TS->setTrajComputed(computeMP);
    TS->setTaskConstraints(_currConstraints);
    TaskSubSolution* TSS1 = new TaskSubSolution(TS, "approach", 0, t,mainAgent,0);
    TSS1->setStartState(_WS);
    WorldState* s = new WorldState(_WS);
    s->replaceConf(mainEndConf);
    s->replaceConf(targetEndConf);
    s->setTSThatEndWithThisWS(TS);
    s->setTCForNextTask(_currConstraints);
    TSS1->setEndState(s);
    TS->setStopWS(s);

    return TS;
}




bool fixOnWall::setAgents(multimap<string,string> agents_name)
{
    if (agents_name.count("mainAgent") != 1)
    {
        cout << "ERROR: no main agent for " << type << " task." << endl;
        return false;
    }
    mainAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("mainAgent")->second);
    if (!mainAgent)
        return false;

    if (agents_name.count("targetAgent") != 1)
    {
        cout << "ERROR: no target agent for " << type << " task." << endl;
        return false;
    }
    targetAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("targetAgent")->second);
    if (!targetAgent)
        return false;
    return true;
}

bool fixOnWall::setObjects(multimap<string,string> objects_name)
{
    if (objects_name.count("mainObject") != 1)
    {
        cout << "ERROR: no main object for " << type << " task." << endl;
        return false;
    }
    mainObject = global_Project->getActiveScene()->getRobotByName(objects_name.find("mainObject")->second);
    return true;
}

bool fixOnWall::setPoints(multimap<string,p3d_point> points)
{
    return true;
}

bool fixOnWall::setdata(multimap<string,string> additionalData)
{
    if (additionalData.count("hand") != 1)
    {
        cout << "no arm choosen, go for right" << endl;
        setArmToUse(Right);
        return true;
    }

    string handStr = additionalData.find("hand")->second;
    if (handStr == "right")
        setArmToUse(Right);
    else if (handStr == "left")
        setArmToUse(Left);
    else
    {
        cout << "the arm selection is not rigth: choosen : " << handStr << "available : right or left." << endl;
        cout << "setting default arm to right" << endl;
        setArmToUse(Right);
    }

    return true;
}

//vector<TaskSubSolution*> fixOnWall::setNextTraj(TaskSubSolution* doneTSS)
//{
//    vector<TaskSubSolution*> vt;
//    if (!doneTSS->getSolution()->isInPreviousDoneTSS(doneTSS))
//    {
//        return vt;
//    }

//    vector<TaskSubSolution*> vTSS = doneTSS->getNext();

//    for(unsigned int k = 0; k < vTSS.size(); k++)
//    {
//        TaskSubSolution* toAddTSS = vTSS.at(k);
//        bool available = true;
//        for (unsigned int l = 0; l < toAddTSS->getPrevious().size(); l++)
//        {
//            TaskSubSolution* previousToAddTSS = toAddTSS->getPrevious().at(l);
//            if (!doneTSS->getSolution()->isInPreviousDoneTSS(previousToAddTSS))
//            {
//                available = false;
//                break;
//            }
//        }
//        if (available)
//        {
//            if (toAddTSS->getSubSolutionName() == "release")
//                detachObject();
//            else
//                toAddTSS->getRobot()->getRobotStruct()->tcur = toAddTSS->getTraj();

//            toAddTSS->getStartContext()->load();

//            doneTSS->getSolution()->addToPreviousDoneTSS(toAddTSS);
//            vt.push_back(toAddTSS);
//            //tcur
//            //addtoPrevious
//        }

//    }
//    return vt;
////    if (result.at(subSolId)->getSubSolutionName() == "grasp")
////        attachObject();
////    else if (result.at(subSolId)->getSubSolutionName() == "release")
////        detachObject();
////    else
////        getMainAgent()->getRobotStruct()->tcur = result.at(subSolId)->getTraj();

////    result.at(subSolId)->getStartContext()->load();
//}

string fixOnWall::getDescr()
{
    return "Task: \nId \t\t= " + to_string(task_id) + "\nType \t\t= " + type+ "\nMain agent \t\t= " + mainAgent->getName() + "\ntarget \t= is to be specified";
}

string fixOnWall::getTextFromValues()
{
    return "Task: " + type + " this task is not implemented yet!";
}




double fixOnWall::getConfigCost()
{
    return 1;
}

