#include "privateTask.hpp"

#include "API/HPP_Wrap/Romeo_Wrap.hpp"
#include "P3d-pkg.h"
#include "API/project.hpp"

#include "planner/plannerFunctions.hpp"
#include "plannerFunctions.hpp"
#include "planner/cost_space.hpp"

#ifdef LIGHT_PLANNER
#include "LightPlanner-pkg.h"
#endif

#include "database/semantics/Sem_SemanticsManager.hpp"

#ifdef HPP
RomeoWrap *_romeoWrap;
#endif

#include <iostream>
#include <map>

#include "GTP/GTPTools/solutionTools/taskAllSolutions.hpp"
#include "GTP/GTPTools/solutionTools/taskSolution.hpp"

using namespace std;

PrivateTask::PrivateTask() : Task()
{
    mainAgent = NULL;
}

bool PrivateTask::initAll()
{
    return true;
}

bool PrivateTask::initialize()
{
    return true;
}

bool PrivateTask::finalize()
{
    return true;
}

bool PrivateTask::setMainAgent(multimap<string,string> agents_name)
{
    if (agents_name.count("mainAgent") != 1)
    {
        cout << "ERROR: no main agent for " << type << " task." << endl;
        SD->setMapEntry("returnReport","no_main_agent");
        return false;
    }
    mainAgent = global_Project->getActiveScene()->getRobotByName(agents_name.find("mainAgent")->second);
    if (!mainAgent)
    {
        SD->setMapEntry("returnReport","no_main_agent_wrong_name");
        return false;
    }
    if (mainAgent->getName() == "ROMEO")
    {
#ifdef HPP
        if (!_romeoWrap)
            _romeoWrap = new RomeoWrap(mainAgent);
        if (!_romeoWrap)
#endif
            return false;
    }
    return true;
}

bool PrivateTask::isPossibleMainAgent(Robot* r)
{
    bool ok = false;
    if(r->getHriAgent()){
        std::string sem_report;
        ok = Sem::SemanticsManager::mainSemanticsManager()->actionsManager()->isActionPossible(Sem::ActionType(getType()),r->getHriAgent()->type,sem_report);
        if(!ok){
            SD->setMapEntry("returnReport","agent_cant_do_that");
            //returnReport = "agent_not_in_list";
        }
    }else{
        SD->setMapEntry("returnReport","agent_has_no_hri_struct");
    }
    return ok;
}

void PrivateTask::openGripper(confPtr_t &q, bool open, int armId)
{
    if (q->getRobot()->getHriAgent()) {
        switch (q->getRobot()->getHriAgent()->type) {
            case HRI_PR2:
            {
                double v = 0.02;
                if (open)
                    v = 0.5;

                if (armId == 0)
                {
                    (*q)[23] = v;
                    (*q)[24] = v;
                }
                else
                {
                    (*q)[32] = v;
                    (*q)[33] = v;
                }
                break;
            }
            case HRI_KUKAROBOT:
            {
                double v = 0.02;
                if (open)
                    v = 0.5;

                (*q)[19] = v;
                (*q)[20] = v;
                break;
            }
            case HRI_ROMEO:
            {
                double v = 0.;
                double u = 0.;
                if (open) {
                    v = 1.064651;
                    u = 0.610865;
                }

                if (armId == 0) {
                    (*q)[61] = v;
                    (*q)[62] = v;
                    (*q)[63] = v;
                    (*q)[64] = v;
                    (*q)[65] = v;
                    (*q)[66] = v;
                    (*q)[67] = v;
                    (*q)[68] = v;
                    (*q)[69] = v;
                    (*q)[70] = v;
                    (*q)[71] = u;
                    (*q)[72] = v;
                } else {
                    (*q)[30] = v;
                    (*q)[31] = v;
                    (*q)[32] = v;
                    (*q)[33] = v;
                    (*q)[34] = v;
                    (*q)[35] = v;
                    (*q)[36] = v;
                    (*q)[37] = v;
                    (*q)[38] = v;
                    (*q)[39] = v;
                    (*q)[40] = u;
                    (*q)[41] = v;
                }
                break;
            }
            default:
                break;
        }
    }
}

bool PrivateTask::testConstraints(int alternativeId)
{
    TaskSolution* TS = result->getAlternativeWithId(alternativeId);
    _returnReportcount++;
    if (!TS)
    {
        cout << "test of constraint fail, no solution with id " << alternativeId << endl;
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","alternative_creation_fail");
        }
        return false;
    }
    _returnReportcount++;
    if (testConstraints(TS))
        return true;
    else
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","constraints_not_respected");
        }
    }

    result->removeAlternativeWithId(alternativeId);
    return false;
}

bool PrivateTask::testConstraints(TaskSolution *TS)
{
    return TS && TS->getTaskConstraints()->areFactConstraintRespected(TS->getStopWS());
}

bool PrivateTask::run(multimap<string,string> agents_name,
                           multimap<string,string> objects_name,
                           multimap<string,p3d_point> point3d,
                           multimap<string,string> additionalData,
                           WorldState *WS, bool computeMP)
{

    WS->load();
    _WS = WS;
    if (!initAll())
    {
        cout << "ERROR: init fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setAgents(agents_name))
    {
        cout << "ERROR: setAgents fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!isPossibleMainAgent(mainAgent))
    {
        cout << "ERROR: mainAgent is not in the list of possible agent for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setObjects(objects_name))
    {
        cout << "ERROR: setObjects fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setPoints(point3d))
    {
        cout << "ERROR: setPoints fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!setdata(additionalData))
    {
        cout << "ERROR: setData fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }


    if (!findCandidateSolutions())
    {
        cout << "ERROR: findCandidateSolutions fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    bool res = false;
    int count = 0;
    while (!res)
    {
        //cout << "count loop = " << count << endl;
        _returnReportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
            continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        if(!setToSolution(0,computeMP))
            continue;
        if(!testConstraints(0))
            continue;

        res = true;
    }

    if (res)
        SD->setMapEntry("returnReport","OK");

    computeCost(0);
    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }
    return res;
}


TaskSolution* PrivateTask::findAlternative(bool computeMP)
{
    int count = 0;
    if (!getResult()->getAlternativeWithId(0))
    {
        cout<< "This task has no first solution, no alternative can be computed in that case" << endl;
        return NULL;
    }
    getResult()->getAlternativeWithId(0)->getStartWS()->load(); //
    int id = -1;
    bool res = false;


    if (!initialize())
    {
        cout << "ERROR: initialize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    while (!res)
    {
        //cout << "count loop = " << count << endl;
        _returnReportcount = 0;
        if(count++ > nbMaxLoops)
            break;
        if(!getRandomSol())
           continue;
        if(!findConfigurations())
           continue;
        if (computeMP)
            if(!findTrajectories())
                continue;
        id = result->getNextId();

        if(!setToSolution(id,computeMP))
           continue;
        if(!testConstraints(id))
            continue;
        res = true;
    }

    if (res)
        SD->setMapEntry("returnReport","OK");

    computeCost(id);

    if (!finalize())
    {
        cout << "ERROR: finalize fails for task " << type << " with id " << task_id << endl;
        return false;
    }

    if (res && id != -1)
        return getResult()->getAlternativeWithId(id);
    return NULL;
}


armLabel PrivateTask::getArmLabelFromId(int armId)
{
    if (armId == 1)
        return Left;
    else
        return Right;
}

bool PrivateTask::setArmToUse(int armId){
    if (!mainAgent)
        return false;

    ArmManipulationData m=mainAgent->getRobotStruct()->armManipulationData->at(armId);

    m_arm = mainAgent->getHriAgent()->manip->external_hand[armId]->hand_l;

    handType = m.getHandProperties().type;
    _armId = armId;

    return true;
}

bool PrivateTask::setArmToUse(armLabel arm_t)
{
    m_arm = arm_t;
    if (!mainAgent)
        return false;
    //define, from the arm string and the type of main robot the hand type:
    // need to be changed!!!!!!!!!!!!!! should be stocked in agent.h

    bool t = false;
    if (mainAgent->getHriAgent()->type == HRI_PR2 ||
           mainAgent->getHriAgent()->type == HRI_KUKAROBOT )
    {

        t =true;
        handType = GP_PR2_GRIPPER;
        if (ManipPlMap.count(mainAgent->getName()) > 0)
        {
            if (arm_t == Left)
                fixAllJointsWithoutArm(ManipPlMap[mainAgent->getName()]->robot(),1);
            else
                fixAllJointsWithoutArm(ManipPlMap[mainAgent->getName()]->robot(),0);
        }
    }
    else if (mainAgent->getName().find("HUMAN") !=std::string::npos)
    {
        t =true;
        if (arm_t == Left)
            handType = GP_SAHAND_LEFT;
        else
            handType = GP_SAHAND_RIGHT; //by default right hand
    }
    else if (mainAgent->getName().find("ROMEO") !=std::string::npos)
    {
        t = true;
        if (arm_t == Left)
            handType = GP_ROMEO_LEFT_HAND;
        else
            handType = GP_ROMEO_RIGHT_HAND; //by default right hand
    }

    if (t)
    {
        for (int i = 0; i < mainAgent->getHriAgent()->manip->external_hand_no; i++)
        {
            if (!mainAgent->getHriAgent()->manip->external_hand)
                return false;
            if (!mainAgent->getHriAgent()->manip->external_hand[i])
                return false;
            if (mainAgent->getHriAgent()->manip->external_hand[i]->hand_l == m_arm)
            {
                _armId = mainAgent->getHriAgent()->manip->external_hand[i]->hand_id;
                break;
            }
        }
    }
//    cout << "armid = " << _armId << endl;
    return t;
}

int PrivateTask::getArmId()
{
    return _armId;
//    if (m_arm == Left)
//        return 1;
//    else if (m_arm == Right)
//        return 0;
//    else
//    {
//        cout << "arm not defined as it should be, by default choosing id 0, string defined = " << m_arm << endl;
//        return 0;
//    }
}

bool PrivateTask::initMotionPlaner(Robot* rob)
{
    if (rob->getHriAgent() && rob->getHriAgent()->type == HRI_PR2)
    {
        if (_armId > 1 || _armId < 0)
        {
//            cout << "wrong _armId param, " << _armId << " going with right arm" <<endl;
            _armId = 0;
        }

        p3d_cntrt* ct;
        // over all constraints
        for(int i=0; i<rob->getRobotStruct()->cntrt_manager->ncntrts; i++)
        {
            string name = rob->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

            if( name == "p3d_fix_jnts_relpos" ){
                continue;
            }
            // get constraint from the cntrts manager
            ct = rob->getRobotStruct()->cntrt_manager->cntrts[i];
            p3d_desactivateCntrt( rob->getRobotStruct(), ct );
        }

        p3d_multiLocalPath_disable_all_groupToPlan( rob->getRobotStruct() , false );
        p3d_multiLocalPath_set_groupToPlan( rob->getRobotStruct(), 2, 1, false);


        //Fix every joint except the active arm(s)
        for (int i = 0; i < rob->getRobotStruct()->njoints + 1; i++) {
            p3d_jnt * joint = rob->getRobotStruct()->joints[i];
            if (joint->type != P3D_BASE && joint->type != P3D_FIXED) {
                fixJoint(rob->getRobotStruct(), joint, joint->jnt_mat);
            }
        }


        ArmManipulationData& mData = (*rob->getRobotStruct()->armManipulationData)[_armId];
        ct = mData.getCcCntrt();
        for(int i = 0; i < ct->npasjnts; i++){
            p3d_jnt * joint = ct->pasjnts[i];
            if(joint->type != P3D_FIXED){
                unFixJoint(rob->getRobotStruct(), joint);
            }
        }
        if((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_lwr_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_pr2_arm_ik"))){
            p3d_jnt * joint = rob->getRobotStruct()->joints[ct->argu_i[0]];
            unFixJoint(rob->getRobotStruct(), joint);
        }




        desactivateTwoJointsFixCntrt(rob->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);

        activateCcCntrts(rob->getRobotStruct(), _armId, false);
        ManipulationUtils::unfixManipulationJoints(rob->getRobotStruct(), _armId);
        shootTheObjectArroundTheBase(rob->getRobotStruct(), rob->getRobotStruct()->baseJnt, mData.getManipulationJnt(), 2.0);
        mData.getManipulationJnt()->dist  = 0.10;
    }


    return true;

}

p3d_traj* PrivateTask::compute_traj(confPtr_t q_init, confPtr_t q_goal, Robot* object, int& status)
{

    clock_t begin = clock();
//    p3d_plan();


    SD->getSDMapEntry("statistics")->incrementMapEntry("nbMPCalled");
    if (q_init->getRobot()->getName() != q_goal->getRobot()->getName())
        return NULL;

    if (!q_init->getRobot()->getHriAgent())
    {
        SD->setMapEntry("returnReport","not_an_agent");
        status = 5;
        return NULL;
    }

    int part = PLAN_FULL;
    HRI_AGENT_TYPE agent_type = q_init->getRobot()->getHriAgent()->type;
    switch (agent_type)
    {
        case HRI_ROMEO:
        {
#ifdef HPP
            double *timer;
            p3d_traj *outputTraj;

            bool result = false;
            if (PlanEnv->getBool(PlanParam::useGraph))
            {

            }
            else
            {
                result = _romeoWrap->findTrajectory(q_init, q_goal, outputTraj, timer);
            }
            std::cout << "Path planning : " << (result ? "SUCCESS" : "FAIL") << ", in " << timer << "s" << std::endl;
            if (!result)
            {
                SD->setMapEntry("returnReport","no_path_found");
                return NULL;
            }
            timeMp += (double)(clock() - begin)/CLOCKS_PER_SEC;
            return outputTraj;
#else
            int ress = -1;
            p3d_traj *outputTraj = NULL;
            confPtr_t initConf = mainAgent->getCurrentPos();

            bool prevFeetStatus = ENV.getBool(Env::FixFeet);
            if (!prevFeetStatus)
                ENV.setBool(Env::FixFeet, true);

            confPtr_t q_cur = q_init->getRobot()->getCurrentPos();
            ress = p3d_plan(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct(), q_goal->getConfigStruct(), part, ENV.getBool(Env::useCartesian));
            mainAgent->setAndUpdate(*q_cur);

            if (ress == PLAN_STATUS_FOUND)
                outputTraj = p3d_create_traj_by_copy(q_init->getRobot()->getRobotStruct()->tcur);

            status = ress;

            if (!outputTraj) {
                _returnReportcount++;
                if (_returnReportcountMax<_returnReportcount) {
                    _returnReportcountMax=_returnReportcount;
                    SD->setMapEntry("returnReport","no_traj");
                }
                return NULL;
            }

            ENV.setBool(Env::FixFeet, prevFeetStatus);
            double timeMp_ = SD->getSDMapEntry("times")->getDoubleMapEntry("timeMp");
            SD->getSDMapEntry("times")->setMapEntry("timeMp", timeMp_ + (double)(clock() - begin)/CLOCKS_PER_SEC);
            return outputTraj;
#endif
        }
        case HRI_PR2:
        {
            for (int i = 0; i < q_init->getRobot()->getRobotStruct()->nb_dof; i++)
            {
    //            cout << " for " << i << " diff = " << fabs(q_init->at(i) - q_goal->at(i)) << endl;
                if (fabs(q_init->at(i) - q_goal->at(i)) > EPS3)
                {
                    if (5<i && i<12)
                    {
                        part = PLAN_BASE;
                        i = 12;
                    }

                    else if (14<i && i<23)
                    {
                        if (part == PLAN_BASE)
                        {
                            part=PLAN_FULL;
                            break;
                        }
                        part=PLAN_RIGHT_ARM;
                        i=23;
                    }

                    else if (24<i && i<32)
                    {
                        if (part == PLAN_BASE || part == PLAN_RIGHT_ARM)
                        {
                            part=PLAN_FULL;
                            break;
                        }
                        part=PLAN_LEFT_ARM;
                        i=32;
                    }
                }
            }

            mainAgent->setAndUpdate(*q_init);

            double Tx, Ty, Tz;
            double Rx, Ry, Rz;
            p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "rIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

            q_init->getConfigStruct()[37] = Tx;
            q_init->getConfigStruct()[38] = Ty;
            q_init->getConfigStruct()[39] = Tz;
            q_init->getConfigStruct()[40] = Rx;
            q_init->getConfigStruct()[41] = Ry;
            q_init->getConfigStruct()[42] = Rz;

            p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "lIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

            q_init->getConfigStruct()[43] = Tx;
            q_init->getConfigStruct()[44] = Ty;
            q_init->getConfigStruct()[45] = Tz;
            q_init->getConfigStruct()[46] = Rx;
            q_init->getConfigStruct()[47] = Ry;
            q_init->getConfigStruct()[48] = Rz;



            int ress = -1;
            p3d_traj* t = NULL;
            bool testOn = true;
            if (PlanEnv->getBool(PlanParam::useGraph))
            {
                testOn = false;
                PlanNode* PN = _planTree->getClosestNode(_WS,currentDataType);
                if (!PN)
                {
                    testOn = true;
                }
                else
                {

                }
            }
            if (testOn)
            {
                confPtr_t q_cur = q_init->getRobot()->getCurrentPos();
                std::pair<Sem::PlanningPart,bool> part_to_use = getPlanPartToUse();

                if(part_to_use.second){

                    Json::Value json;
                    //Sem::PlanningPart planPart = Sem::SemanticsManager::mainSemanticsManager()->actionsManager()->getJointsForAction(Sem::ActionType(getType()),agent_type,false)[0];
                    Sem::PlanningPart planPart = part_to_use.first;
                    Sem::SemanticsManager::mainSemanticsManager()->planParts()->appendJointsSetToJson(json,&planPart);
                    cout << "plan part:\n" << json <<endl;
                    ress = p3d_plan(q_init->getRobot()->getRobotStruct(),q_init->getConfigStruct(), q_goal->getConfigStruct(),
                                    planPart.getAllJointIds(),
                                    ENV.getBool(Env::useCartesian));
                }else{
                    ress = p3d_plan(q_init->getRobot()->getRobotStruct(),q_init->getConfigStruct(), q_goal->getConfigStruct(), part, ENV.getBool(Env::useCartesian));
                }
                mainAgent->setAndUpdate(*q_cur);

                if (ress == PLAN_STATUS_FOUND)
                    t = p3d_create_traj_by_copy(q_init->getRobot()->getRobotStruct()->tcur);
            }

            status = ress;
            SD->getSDMapEntry("times")->setMapEntry("timeMp",(double)SD->getSDMapEntry("times")->getDoubleMapEntry("timeMp") + (double)(clock() - begin)/CLOCKS_PER_SEC);
            return t;
        }
    case HRI_KUKAROBOT:
    {

#ifdef P3D_CONSTRAINTS
        p3d_cntrt* ct;
        for(int i=0; i<q_init->getRobot()->getRobotStruct()->cntrt_manager->ncntrts; i++)
        {
            string name = q_init->getRobot()->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

            if( name == "p3d_fix_jnts_relpos" ){
                continue;
            }
            // get constraint from the cntrts manager
            ct = q_init->getRobot()->getRobotStruct()->cntrt_manager->cntrts[i];
            p3d_desactivateCntrt( q_init->getRobot()->getRobotStruct(), ct );
        }

#endif

        for (unsigned int i = 0; i< q_init->getRobot()->getJoint(0)->getNumberOfDof();i++)
        {
            q_init->getRobot()->getJoint(0)->getJointStruct()->dof_data[i].is_active_for_planner = 0;
        }
        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "IkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q_init->getConfigStruct()[21] = Tx;
        q_init->getConfigStruct()[22] = Ty;
        q_init->getConfigStruct()[23] = Tz;
        q_init->getConfigStruct()[24] = Rx;
        q_init->getConfigStruct()[25] = Ry;
        q_init->getConfigStruct()[26] = Rz;


        int ress = -1;
        p3d_traj* t = NULL;

        confPtr_t q_cur = q_init->getRobot()->getCurrentPos();

//        q_init->print(true);



        //ress = p3d_plan(q_init->getRobot()->getRobotStruct(),q_init->getConfigStruct(), q_goal->getConfigStruct(), part, ENV.getBool(Env::useCartesian));
        Json::Value json;
        Sem::PlanningPart planPart = Sem::SemanticsManager::mainSemanticsManager()->actionsManager()->getJointsForAction(Sem::ActionType(getType()),agent_type,false)[0];
        Sem::SemanticsManager::mainSemanticsManager()->planParts()->appendJointsSetToJson(json,&planPart);
        cout << "plan part:\n" << json <<endl;
        ress = p3d_plan(q_init->getRobot()->getRobotStruct(),q_init->getConfigStruct(), q_goal->getConfigStruct(),
                        planPart.getAllJointIds(),
                ENV.getBool(Env::useCartesian));
        mainAgent->setAndUpdate(*q_cur);

        if (ress == PLAN_STATUS_FOUND)
            t = p3d_create_traj_by_copy(q_init->getRobot()->getRobotStruct()->tcur);

        status = ress;
        SD->getSDMapEntry("times")->setMapEntry("timeMp",(double)SD->getSDMapEntry("times")->getDoubleMapEntry("timeMp") + (double)(clock() - begin)/CLOCKS_PER_SEC);
        return t;
    }
        default:
            return NULL;
    }

}

p3d_traj* PrivateTask::compute_traj_using_localPath(confPtr_t q_init, confPtr_t q_goal, Robot* object, int& status, bool testingStraightPath)
{
    q_init->getRobot()->setAndUpdate(*q_init);

    if (q_init->getRobot()->getHriAgent() && q_init->getRobot()->getHriAgent()->type == HRI_PR2 )
    {


        for(int i=0; i<q_init->getRobot()->getRobotStruct()->cntrt_manager->ncntrts; i++)
        {
            string name = q_init->getRobot()->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

            if( name == "p3d_fix_jnts_relpos" ){
                continue;
            }

//            cout << "deactivate : " << name << endl;

            // get constraint from the cntrts manager
            p3d_cntrt* ct = q_init->getRobot()->getRobotStruct()->cntrt_manager->cntrts[i];
            p3d_desactivateCntrt( q_init->getRobot()->getRobotStruct(), ct );
        }


//        cout << "arm to use = " << _armId << endl;
        ArmManipulationData& mData = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[_armId];
        p3d_cntrt* ct = mData.getCcCntrt();
        for(int i = 0; i < ct->npasjnts; i++){
            p3d_jnt * joint = ct->pasjnts[i];
            if(joint->type != P3D_FIXED){
                unFixJoint(q_init->getRobot()->getRobotStruct(), joint);
            }
        }



        if((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
            (!strcmp(ct->namecntrt, "p3d_lwr_arm_ik")) ||
            (!strcmp(ct->namecntrt, "p3d_pr2_arm_ik"))){
            p3d_jnt * joint = q_init->getRobot()->getRobotStruct()->joints[ct->argu_i[0]];
            unFixJoint(q_init->getRobot()->getRobotStruct(), joint);
        }
//        mData.setCartesian( true );
        mData.setCartesian( false );


        ArmManipulationData& mDataOther = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[(_armId+1)%2];
        mDataOther.setCartesian( false );

        for (int i=0; i < int((*q_init->getRobot()->getRobotStruct()->armManipulationData).size()); i++)
        {
            ArmManipulationData& armData  = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[i];

            desactivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),armData.getManipulationJnt(), armData.getCcCntrt()->pasjnts[ armData.getCcCntrt()->npasjnts-1 ]);


            if (armData.getCartesian())
            {
                cout << "Arm Data " << i <<  " is set cartesian" << endl;
                activateCcCntrts(q_init->getRobot()->getRobotStruct(), i, false);
                ManipulationUtils::unfixManipulationJoints(q_init->getRobot()->getRobotStruct(), i);
                shootTheObjectArroundTheBase(q_init->getRobot()->getRobotStruct(), q_init->getRobot()->getRobotStruct()->baseJnt, armData.getManipulationJnt(), 2.0);
                armData.getManipulationJnt()->dist  = 0.10;

            } else {
                deactivateCcCntrts(q_init->getRobot()->getRobotStruct(), i);
                p3d_set_and_update_this_robot_conf(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct());
                setAndActivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),armData.getManipulationJnt(), armData.getCcCntrt()->pasjnts[ armData.getCcCntrt()->npasjnts-1 ]);
                ManipulationUtils::fixManipulationJoints(q_init->getRobot()->getRobotStruct(), i, q_init->getConfigStruct(), NULL);
                shootTheObjectInTheWorld(q_init->getRobot()->getRobotStruct(), armData.getManipulationJnt());
            }
        }


        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "rIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q_init->getConfigStruct()[37] = Tx;
        q_init->getConfigStruct()[38] = Ty;
        q_init->getConfigStruct()[39] = Tz;
        q_init->getConfigStruct()[40] = Rx;
        q_init->getConfigStruct()[41] = Ry;
        q_init->getConfigStruct()[42] = Rz;

        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "lIkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q_init->getConfigStruct()[43] = Tx;
        q_init->getConfigStruct()[44] = Ty;
        q_init->getConfigStruct()[45] = Tz;
        q_init->getConfigStruct()[46] = Rx;
        q_init->getConfigStruct()[47] = Ry;
        q_init->getConfigStruct()[48] = Rz;


    }
    else if (q_init->getRobot()->getHriAgent() && q_init->getRobot()->getHriAgent()->type == HRI_KUKAROBOT )
    {
        p3d_cntrt* ct = mainAgent->getRobotStruct()->cntrt_manager->cntrts[0];
        p3d_desactivateCntrt(mainAgent->getRobotStruct(), ct);


        double Tx, Ty, Tz;
        double Rx, Ry, Rz;
        p3d_mat4ExtractPosReverseOrder(p3d_get_robot_jnt_by_name(q_init->getRobot()->getRobotStruct(), "IkJoint")->abs_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);

        q_init->getConfigStruct()[21] = Tx;
        q_init->getConfigStruct()[22] = Ty;
        q_init->getConfigStruct()[23] = Tz;
        q_init->getConfigStruct()[24] = Rx;
        q_init->getConfigStruct()[25] = Ry;
        q_init->getConfigStruct()[26] = Rz;
    }


    confPtr_t q_curr = q_init->getRobot()->getCurrentPos();
    if (q_init->getRobot()->getName() != q_goal->getRobot()->getName())
        return NULL;

    if (q_init->equal(*q_goal))
    {
        status = PLAN_STATUS_START_EQUALS_GOAL;
        return NULL;
    }

    p3d_localpath *lp = p3d_local_planner(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct(), q_goal->getConfigStruct());
    if (!lp || lp->range_param == P3D_HUGE)
    {
        status = PLAN_STATUS_NOT_FOUND;
        return NULL;
    }

    p3d_traj *traj = p3d_create_empty_trajectory(q_init->getRobot()->getRobotStruct());
    traj->courbePt = lp;
    traj->nlp = 1;
    traj->range_param = lp->range_param;
    double d = 0;
    if (object)
        activateFingertipCol(false,q_init->getRobot(),object,NULL);
//        (q_init->getRobot()->getRobotStruct(), object->getRobotStruct());
    bool ok = true;
    while (ok && d < traj->range_param)
    {
        d += traj->range_param / 100;
        configPt q =p3d_config_at_param_along_traj(traj,d);
        p3d_set_and_update_this_robot_conf(q_init->getRobot()->getRobotStruct(),q);
        p3d_update_this_robot_pos_without_cntrt(q_init->getRobot()->getRobotStruct());

        vector<p3d_rob*> colRob = pqp_robot_robot_list_collision_test(q_init->getRobot()->getRobotStruct());

        if (colRob.size()>0) //col
        {
            for (unsigned int i = 0; i < colRob.size(); i++)
            {
                SD->getSDMapEntry("colision")->incrementMapEntry(colRob.at(i)->name);
            }
            p3d_destroy_traj(q_init->getRobot()->getRobotStruct(),traj);
            traj = NULL;
            status = PLAN_STATUS_NOT_FOUND;
            ok = false;
        }
    }
    if (object)
        activateFingertipCol(true,q_init->getRobot(),object,NULL);
//        p3d_col_activate_robot_robot(q_init->getRobot()->getRobotStruct(), object->getRobotStruct());

    q_init->getRobot()->setAndUpdate(*q_curr);

    if (q_init->getRobot()->getHriAgent() && q_init->getRobot()->getHriAgent()->type == HRI_PR2)
    {
        ArmManipulationData& mData = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[0];
        mData.setCartesian( false );
        desactivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);
        deactivateCcCntrts(q_init->getRobot()->getRobotStruct(), 0);
        p3d_set_and_update_this_robot_conf(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct());
        setAndActivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);
        ManipulationUtils::fixManipulationJoints(q_init->getRobot()->getRobotStruct(), 0, q_init->getConfigStruct(), NULL);
        shootTheObjectInTheWorld(q_init->getRobot()->getRobotStruct(), mData.getManipulationJnt());


        ArmManipulationData&  mData1 = (*q_init->getRobot()->getRobotStruct()->armManipulationData)[1];
        mData1.setCartesian( false );
        desactivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),mData1.getManipulationJnt(), mData1.getCcCntrt()->pasjnts[ mData1.getCcCntrt()->npasjnts-1 ]);
        deactivateCcCntrts(q_init->getRobot()->getRobotStruct(), 1);
        p3d_set_and_update_this_robot_conf(q_init->getRobot()->getRobotStruct(), q_init->getConfigStruct());
        setAndActivateTwoJointsFixCntrt(q_init->getRobot()->getRobotStruct(),mData1.getManipulationJnt(), mData1.getCcCntrt()->pasjnts[ mData1.getCcCntrt()->npasjnts-1 ]);
        ManipulationUtils::fixManipulationJoints(q_init->getRobot()->getRobotStruct(), 1, q_init->getConfigStruct(), NULL);
        shootTheObjectInTheWorld(q_init->getRobot()->getRobotStruct(), mData1.getManipulationJnt());


    }


    if(ok && traj && testingStraightPath){
        if(!this->isPathStraight(mainAgent,traj)){
            cout << "localpath is not straight"<<endl;
            p3d_destroy_traj(q_init->getRobot()->getRobotStruct(),traj);
            traj=NULL;
            status = PLAN_STATUS_NOT_FOUND;
        }
    }
    if (ok && traj)
        status = PLAN_STATUS_FOUND;

    return traj;
}

double PrivateTask::computeCost(int altId)
{
    //get the taskSolution
    //get subtrajs
    //get beg and end confs
    //compute dist
    //add dists
    TaskSolution* TS = result->getAlternativeWithId(altId);
    if (!TS)
        return 1000;
    vector<TaskSubSolution*> vtss = TS->getTSSByRobot(mainAgent);
    double cost = 0;
    GeometricComputations GC = GeometricComputations();
    for (unsigned int i = 0; i < vtss.size(); i++)
    {
        TaskSubSolution* TSS = vtss.at(i);
        confPtr_t q1 = TSS->getStartState()->getRobotConf(mainAgent);
        confPtr_t q2 = TSS->getEndState()->getRobotConf(mainAgent);

        cost += GC.distConfs(q1,q2);
//        cost += p3d_dist_config(mainAgent->getRobotStruct(), q1->getConfigStruct(), q2->getConfigStruct(), true);
//        q1->

    }
    TS->setCost(cost);

    return cost;
}

bool PrivateTask::attachObject(Robot* rob, Robot* obj, bool attach)
{
    return attachObject(rob,obj,attach,getArmId());
}

bool PrivateTask::attachObject(Robot* rob, Robot* obj, bool attach, int  armId)
{
    if (!rob || ! obj)
        return false;
    int jointId = -1;
    if (attach)
    {
        if (!rob->isAgent())
            return false;
        if (!rob->getHriAgent())
            return false;
        if (!rob->getHriAgent()->manip)
            return false;
        if (!rob->getHriAgent()->manip->external_hand[armId])
            return false;
        jointId = rob->getHriAgent()->manip->external_hand[armId]->joint_id_in_rob;

        if (!rob->isObjectAttached(obj,jointId))
        {
            rob->attachObj(obj,jointId);
        }
    }
    else
        rob->dettachObj(obj);

    return rob->whereObjectAttached(obj);
    //    int r = hri_agent_is_grasping_obj_rob_pt(robot, attach , c , armId);

    //    return !r;
}


bool PrivateTask::computeManipulationConfs(confPtr_t& approachC, confPtr_t& graspC, confPtr_t& graspOC, confPtr_t& escapeC, p3d_matrix4 objectPosMat,
                                           gpGrasp grasp, p3d_vector3 approachDir, double approachDist, p3d_vector3 escapeDir, double escapeDist)
{
    confPtr_t qInit = (confPtr_t)new Configuration(*mainAgent->getCurrentPos());

    confPtr_t q;

    p3d_matrix4 graspMatWorldFrame, approachMat, escapeMat, /*rotMat,*/ reelTarg;
    p3d_mat4Mult(objectPosMat, grasp.frame, graspMatWorldFrame);

    p3d_mat4Copy(graspMatWorldFrame, reelTarg);

    p3d_vector3 v_new;

    p3d_matrix3 matRot;
    p3d_mat4ExtractRotMatrix(objectPosMat, matRot);
    p3d_xform3(matRot, approachDir,  v_new);

    p3d_mat4Copy(reelTarg, approachMat);
    if(mainAgent->getHriAgent()->type == HRI_ROMEO) { // Romeo hand frame is oriented differently
        approachMat[0][3] = reelTarg[0][3] + v_new[0]*approachDist;
        approachMat[1][3] = reelTarg[1][3] + v_new[1]*approachDist;
        approachMat[2][3] = reelTarg[2][3] + v_new[2]*approachDist;
    } else {
        approachMat[0][3] = reelTarg[0][3] - v_new[0]*approachDist;
        approachMat[1][3] = reelTarg[1][3] - v_new[1]*approachDist;
        approachMat[2][3] = reelTarg[2][3] - v_new[2]*approachDist;
    }

//    cout << "x " << v_new[0] << " y " << v_new[1] << " z " << v_new[2] << endl;
//    cout << "approach dist = " << approachDist << endl ;

    p3d_mat4Copy(reelTarg, escapeMat);
    escapeMat[0][3] = reelTarg[0][3] + escapeDir[0]*escapeDist;
    escapeMat[1][3] = reelTarg[1][3] + escapeDir[1]*escapeDist;
    escapeMat[2][3] = reelTarg[2][3] + escapeDir[2]*escapeDist;

    bool testIk = computeInverseKinematic(mainAgent, reelTarg, q, _armId, true );
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbGikCalled");

    _returnReportcount++;
    if (!testIk)
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","no_IK_for_grasp");
        }
        return false;
    }

    //graspC = confPtr_t(new Configuration(*q));
    openGripper(q,true,_armId);

    graspC = graspOC = confPtr_t(new Configuration(*q));


    mainAgent->setAndUpdate(*graspC);
    testIk = computeInverseKinematic(mainAgent, approachMat  ,q , _armId, true );
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbGikCalled");

    _returnReportcount++;
    if (!testIk)
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","no_IK_for_hand_free");
        }
        return false;
    }

    openGripper(q,true,_armId);
    approachC = confPtr_t(new Configuration(*q));

    //    mainAgent->setAndUpdate(*approachConf);

    mainAgent->setAndUpdate(*graspC);
    testIk = computeInverseKinematic(mainAgent, escapeMat  ,q , _armId, true );
    SD->getSDMapEntry("statistics")->incrementMapEntry("nbGikCalled");

    _returnReportcount++;
    if (!testIk)
    {
        if (_returnReportcountMax < _returnReportcount)
        {
            _returnReportcountMax = _returnReportcount;
            SD->setMapEntry("returnReport","no_IK_for_obj_in_hand");
        }
        return false;
    }

    openGripper(q,true,_armId);
    escapeC = confPtr_t(new Configuration(*q));

    mainAgent->setAndUpdate(*approachC);

//    _WS->replaceConf(qInit);

    mainAgent->setAndUpdate(*qInit);
    return true;
}


bool PrivateTask::computeInverseKinematic(Robot* ag, p3d_matrix4 target_pos, confPtr_t &solution, int armId, bool fromCurPos)
{
    if (!ag->getHriAgent())
    {
        SD->setMapEntry("returnReport","not_an_agent");
        return false;
    }

    p3d_matrix4 rotMat, tmp;

//    p3d_mat4Print(target_pos,"target_pos");

    switch(ag->getHriAgent()->type)
    {
        bool found, i_can;
        int nbIKloop;
        case HRI_ROMEO:
        {
            rotMat[0][0] = 0;
            rotMat[0][1] = -1;
            rotMat[0][2] = 0;
            rotMat[0][3] = 0;
            rotMat[1][0] = 0;
            rotMat[1][1] = 0;
            rotMat[1][2] = 1;
            rotMat[1][3] = 0;
            rotMat[2][0] = -1;
            rotMat[2][1] = 0;
            rotMat[2][2] = 0;
            rotMat[2][3] = 0;
            rotMat[3][0] = 0;
            rotMat[3][1] = 0;
            rotMat[3][2] = 0;
            rotMat[3][3] = 1;
            p3d_mat4Mult(target_pos, rotMat, tmp);
            p3d_mat4Copy(tmp, target_pos);
#ifdef HPP
            bool randomize = true;
            if (!getResult()->getAlternativeWithId(0))
                randomize = false;
            found = i_can = false;
            nbIKloop = 0;

            confPtr_t outputConf;
            confPtr_t inputConf = confPtr_t(new Configuration(*ag->getCurrentPos()));
            i_can = _romeoWrap->generateConfig(inputConf, outputConf, target_pos, armId, randomize);
            if (i_can)
            {
                // std::cout << "RV -- Stability of generated config : "
                //           << (_romeoWrap->isPositionStable(outputConf) ? "Stable" : "Not stable")
                //           << std::endl;
                if (p3d_col_test_robot(ag->getRobotStruct(),0) == 0)
                    found = true;
                else
                    i_can = false;
            }

            if (!i_can)
            {
                while (nbIKloop < MAXIKLOOP)
                {
                    nbIKloop++;
                    i_can = _romeoWrap->generateConfig(inputConf, outputConf, target_pos, armId, true);
                    if (i_can)
                    {
                        // std::cout << "RV -- Stability of generated config : "
                        //           << (_romeoWrap->isPositionStable(outputConf) ? "Stable" : "Not stable")
                        //           << std::endl;
                        if (p3d_col_test_robot(ag->getRobotStruct(),0) == 0)
                        {
                            found = true;
                            break;
                        }
                    }
                }
            }

            if (found)
            {
    //            q = ag->getCurrentPos()->getConfigStruct();
                solution = confPtr_t(new Configuration(*outputConf));
                return true;
            }
#else
            i_can = found = false;
            configPt q = p3d_copy_config(ag->getRobotStruct(), _WS->getRobotConf(ag)->getConfigStruct());

            // Place virtual object on target
            double Tx, Ty, Tz, Rx, Ry, Rz;
            p3d_mat4ExtractPosReverseOrder(target_pos, &Tx, &Ty, &Tz, &Rx, &Ry, &Rz);
            int ffjntIndex = (*ag->getRobotStruct()->armManipulationData)[armId].getManipulationJnt()->index_dof;
            q[ffjntIndex + 0] = Tx;
            q[ffjntIndex + 1] = Ty;
            q[ffjntIndex + 2] = Tz;
            q[ffjntIndex + 3] = Rx;
            q[ffjntIndex + 4] = Ry;
            q[ffjntIndex + 5] = Rz;

            p3d_set_and_update_this_robot_conf(ag->getRobotStruct(), q);

            // Set orientation and fix feet
            ENV.setBool(Env::eefOrientation, true);
            ENV.setBool(Env::FixFeet, true);

            // Activate constraint
            p3d_cntrt* ct;
            for (int i = 0; i < ag->getRobotStruct()->cntrt_manager->ncntrts; i++) {
                string name = ag->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;
                ct = ag->getRobotStruct()->cntrt_manager->cntrts[i];
                if (name == "p3d_romeo_ik")
                    p3d_activateCntrt(ag->getRobotStruct(), ct);
            }

            // Call constraints resolution (IK)
            i_can = p3d_set_and_update_this_robot_conf(ag->getRobotStruct(), q);
            g3d_draw_allwin_active();
            sleep(1);

            // Deactivate constraints : back to initial state
            for (int i = 0; i < ag->getRobotStruct()->cntrt_manager->ncntrts; i++) {
                string name = ag->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;
                ct = ag->getRobotStruct()->cntrt_manager->cntrts[i];
                if (name == "p3d_romeo_ik")
                    p3d_desactivateCntrt(ag->getRobotStruct(), ct);
            }
            ENV.setBool(Env::eefOrientation, false);
            ENV.setBool(Env::FixFeet, false);

            if (p3d_col_test_robot(ag->getRobotStruct(),0) != 0) {
                p3d_print_col_pair();
            }

            if (i_can && (p3d_col_test_robot(ag->getRobotStruct(),0) == 0)
                      && !p3d_isOutOfBounds(ag->getRobotStruct(), ag->getCurrentPos()->getConfigStruct()))
                found = true;

            if (i_can && found) {
                solution = confPtr_t(new Configuration(*ag->getCurrentPos()));
                return true;
            }
            q =NULL;
#endif
            break;
        }
        case HRI_PR2:
        case HRI_KUKAROBOT:
        {
            if (ag->getHriAgent()->type == HRI_KUKAROBOT)
            {
                rotMat[0][0] = 0;
                rotMat[0][1] = 0;
                rotMat[0][2] = -1;
                rotMat[0][3] = 0;

                rotMat[1][0] = 0;
                rotMat[1][1] = 1;
                rotMat[1][2] = 0;
                rotMat[1][3] = 0;

                rotMat[2][0] = 1;
                rotMat[2][1] = 0;
                rotMat[2][2] = 0;
                rotMat[2][3] = 0;

                rotMat[3][0] = 0;
                rotMat[3][1] = 0;
                rotMat[3][2] = 0;
                rotMat[3][3] = 1;

                p3d_mat4Mult(target_pos, rotMat, tmp);
                p3d_mat4Copy(tmp, target_pos);

                rotMat[0][0] = 0;
                rotMat[0][1] = 1;
                rotMat[0][2] = 0;
                rotMat[0][3] = 0;

                rotMat[1][0] = -1;
                rotMat[1][1] = 0;
                rotMat[1][2] = 0;
                rotMat[1][3] = 0;

                rotMat[2][0] = 0;
                rotMat[2][1] = 0;
                rotMat[2][2] = 1;
                rotMat[2][3] = 0;

                rotMat[3][0] = 0;
                rotMat[3][1] = 0;
                rotMat[3][2] = 0;
                rotMat[3][3] = 1;

                p3d_mat4Mult(target_pos, rotMat, tmp);
            }
            else
            {
                rotMat[0][0] = 0;
                rotMat[0][1] = 1;
                rotMat[0][2] = 0;
                rotMat[0][3] = 0;

                rotMat[1][0] = 0;
                rotMat[1][1] = 0;
                rotMat[1][2] = 1;
                rotMat[1][3] = 0;

                rotMat[2][0] = 1;
                rotMat[2][1] = 0;
                rotMat[2][2] = 0;
                rotMat[2][3] = 0;

                rotMat[3][0] = 0;
                rotMat[3][1] = 0;
                rotMat[3][2] = 0;
                rotMat[3][3] = 1;

                p3d_mat4Mult(target_pos, rotMat, tmp);


            }
            p3d_mat4Copy(tmp, target_pos);


            found = i_can = false;
            double Tx, Ty, Tz;
            double Rx, Ry, Rz;
            nbIKloop = 0;
            p3d_mat4ExtractPosReverseOrder(target_pos, &Tx, &Ty, &Tz,
                                           &Rx, &Ry, &Rz);

            configPt q = p3d_copy_config(ag->getRobotStruct(), ag->getCurrentPos()->getConfigStruct());
            int ffjntIndex = (*ag->getRobotStruct()->armManipulationData)[armId].getManipulationJnt()->index_dof;

            q[ffjntIndex + 0] = Tx;
            q[ffjntIndex + 1] = Ty;
            q[ffjntIndex + 2] = Tz;
            q[ffjntIndex + 3] = Rx;
            q[ffjntIndex + 4] = Ry;
            q[ffjntIndex + 5] = Rz;

            int *ikSol = NULL;
            p3d_set_and_update_this_robot_conf( ag->getRobotStruct(), q );

            p3d_cntrt* ct;

            if(ag->getHriAgent()->type == HRI_PR2)
            {
                for (int i=0; i<ag->getRobotStruct()->cntrt_manager->ncntrts; i++)
                {
                    string name = ag->getRobotStruct()->cntrt_manager->cntrts[i]->namecntrt;

                    if ( name == "p3d_fix_jnts_relpos" )
                        continue;
                    // get constraint from the cntrts manager
                    ct = ag->getRobotStruct()->cntrt_manager->cntrts[i];
                    p3d_desactivateCntrt( ag->getRobotStruct(), ct );
                }

                p3d_multiLocalPath_disable_all_groupToPlan( ag->getRobotStruct() , false );
                p3d_multiLocalPath_set_groupToPlan( ag->getRobotStruct(), 2, 1, false);

                //Fix every joint except the active arm(s)
                for (int i = 0; i < ag->getRobotStruct()->njoints + 1; i++)
                {
                    p3d_jnt * joint = ag->getRobotStruct()->joints[i];
                    if (joint->type != P3D_BASE && joint->type != P3D_FIXED)
                        fixJoint(ag->getRobotStruct(), joint, joint->jnt_mat);
                }
            }


            ArmManipulationData& mData = (*ag->getRobotStruct()->armManipulationData)[armId];
            ct = mData.getCcCntrt();

            if(ag->getHriAgent()->type == HRI_PR2)
            {
                for (int i = 0; i < ct->npasjnts; i++)
                {
                    p3d_jnt * joint = ct->pasjnts[i];
                    if (joint->type != P3D_FIXED)
                        unFixJoint(ag->getRobotStruct(), joint);
                }
            }

            if ((!strcmp(ct->namecntrt, "p3d_kuka_arm_ik")) ||
                (!strcmp(ct->namecntrt, "p3d_lwr_arm_ik"))  ||
                (!strcmp(ct->namecntrt, "p3d_pr2_arm_ik")))
            {
                p3d_jnt * joint = ag->getRobotStruct()->joints[ct->argu_i[0]];
                unFixJoint(ag->getRobotStruct(), joint);
            }
//            mData.setCartesian(true);

            desactivateTwoJointsFixCntrt(ag->getRobotStruct(),mData.getManipulationJnt(), mData.getCcCntrt()->pasjnts[ mData.getCcCntrt()->npasjnts-1 ]);

            activateCcCntrts(ag->getRobotStruct(), armId, false);
            ManipulationUtils::unfixManipulationJoints(ag->getRobotStruct(), armId);
            shootTheObjectArroundTheBase(ag->getRobotStruct(), ag->getRobotStruct()->baseJnt, mData.getManipulationJnt(), 2.0);
            mData.getManipulationJnt()->dist  = 0.10;

            p3d_set_and_update_this_robot_conf(ag->getRobotStruct(), q);

            int begId = -1;
            if(ag->getHriAgent()->type == HRI_PR2)
            {
                if (armId == 0)
                    begId = 5;
                else
                    begId = 15;
            }
            else
            {
                begId = 1;
            }

            if (fromCurPos)
            {
                i_can = p3d_update_this_robot_pos_multisol(ag->getRobotStruct(), NULL, 0, ikSol);

                if (p3d_col_test_robot(ag->getRobotStruct(),0) == 0
                        && !p3d_isOutOfBounds(ag->getRobotStruct(), ag->getCurrentPos()->getConfigStruct()))
                    found = true;
                else
                {

                    if (p3d_col_test_robot(ag->getRobotStruct(),0) != 0)
                    {

                        vector<p3d_rob*> colRob = pqp_robot_robot_list_collision_test(ag->getRobotStruct());
                        for (unsigned int i = 0; i < colRob.size(); i++)
                        {
                            SD->getSDMapEntry("colision")->incrementMapEntry(colRob.at(i)->name);
                        }
                    }
                    i_can = false;
                }
            }

            // in case of faillure, trying different initial pose before resolving the IK
            if (!i_can)
            {
                while (nbIKloop < MAXIKLOOP)
                {
                    nbIKloop++;


                    for(int i=begId; i<begId+7; i++)
                    {
                        double vmin,vmax;
                        ag->getJoint(i)->getDofBounds(0,vmin,vmax);
                        q[ag->getJoint(i)->getIndexOfFirstDof()] = p3d_random(vmin,vmax);
                    }

                    q[ffjntIndex + 0] = Tx;
                    q[ffjntIndex + 1] = Ty;
                    q[ffjntIndex + 2] = Tz;
                    q[ffjntIndex + 3] = Rx;
                    q[ffjntIndex + 4] = Ry;
                    q[ffjntIndex + 5] = Rz;

                    p3d_set_and_update_this_robot_conf( ag->getRobotStruct(), q );
                    i_can = p3d_update_this_robot_pos_multisol(ag->getRobotStruct(), NULL, 0, ikSol);
//                    g3d_draw_allwin_active();
//                    sleep(1);

                    vector<p3d_rob*> colRob = pqp_robot_robot_list_collision_test(ag->getRobotStruct());
                    if (i_can && colRob.size() == 0
                            && !p3d_isOutOfBounds(ag->getRobotStruct(), ag->getCurrentPos()->getConfigStruct()))
                    {
                        found = true;
                        break;
                    }
                    else if (colRob.size() != 0)
                    {

                        for (unsigned int i = 0; i < colRob.size(); i++)
                        {
                            SD->getSDMapEntry("colision")->incrementMapEntry(colRob.at(i)->name);
                        }
                    }

                }
            }

            if (i_can && found)
            {
                solution = confPtr_t(new Configuration(*ag->getCurrentPos()));
                return true;
            }
            q =NULL;
            break;
        }
        default:
            cout << "inverse kinematic not implemented for robot : " << mainAgent->getName() << " yet" << endl;
            break;
    }

    return false;
}


void PrivateTask::activateFingertipCol(bool val, Robot* robi , Robot* object, Robot* supportObj)
{
    switch (mainAgent->getHriAgent()->type) {
        case HRI_PR2:
        case HRI_KUKAROBOT:
        {
            if (val)
            {
                if (supportObj)
                    gpActivate_object_fingertips_collisions( robi->getRobotStruct(), supportObj->getRobotStruct()->joints[1]->o, _armId);
                gpActivate_object_fingertips_collisions( robi->getRobotStruct(), object->getRobotStruct()->joints[1]->o, _armId);
                gpActivate_object_fingertips_collisions( robi->getRobotStruct(), object->getJoint(0)->getJointStruct()->o, _armId);
            }
            else
            {
                if (supportObj)
                    gpDeactivate_object_fingertips_collisions( robi->getRobotStruct(), supportObj->getRobotStruct()->joints[1]->o, _armId);
                gpDeactivate_object_fingertips_collisions( robi->getRobotStruct(), object->getRobotStruct()->joints[1]->o, _armId);
                gpDeactivate_object_fingertips_collisions( robi->getRobotStruct(), object->getJoint(0)->getJointStruct()->o, _armId);
            }
            break;
        }
        case HRI_ROMEO:
        {
            std::string handBaseJntName = (_armId ? "L" : "R");
            handBaseJntName += "WristPitch";
            p3d_jnt *handBaseJnt = p3d_get_robot_jnt_by_name(robi->getRobotStruct(), handBaseJntName.c_str());
            if (val) {
                for (int i = 1; i < handBaseJnt->n_next_jnt; i++) { // Start @ 1 to avoid gripper_joint
                    p3d_jnt *curJnt = handBaseJnt->next_jnt[i];
                    p3d_col_activate_pair_of_objects(object->getRobotStruct()->o[0], curJnt->o);
                    while (curJnt->n_next_jnt > 0) {
                        curJnt = curJnt->next_jnt[0];
                        p3d_col_activate_pair_of_objects(object->getRobotStruct()->o[0], curJnt->o);
                    }
                    if (supportObj) {
                        p3d_col_activate_pair_of_objects(supportObj->getRobotStruct()->o[0], curJnt->o);
                        while (curJnt->n_next_jnt > 0) {
                            curJnt = curJnt->next_jnt[0];
                            p3d_col_activate_pair_of_objects(supportObj->getRobotStruct()->o[0], curJnt->o);
                        }
                    }
                }
            } else {
                for (int i = 1; i < handBaseJnt->n_next_jnt; i++) { // Start @ 1 to avoid gripper_joint
                    p3d_jnt *curJnt = handBaseJnt->next_jnt[i];
                    p3d_col_deactivate_pair_of_objects(object->getRobotStruct()->o[0], curJnt->o);
                    while (curJnt->n_next_jnt > 0) {
                        curJnt = curJnt->next_jnt[0];
                        p3d_col_deactivate_pair_of_objects(object->getRobotStruct()->o[0], curJnt->o);
                    }
                    if (supportObj) {
                        p3d_col_deactivate_pair_of_objects(supportObj->getRobotStruct()->o[0], curJnt->o);
                        while (curJnt->n_next_jnt > 0) {
                            curJnt = curJnt->next_jnt[0];
                            p3d_col_deactivate_pair_of_objects(supportObj->getRobotStruct()->o[0], curJnt->o);
                        }
                    }
                }
            }
            break;
        }
        default:
            break;
    }
}

std::pair<Sem::PlanningPart, bool> PrivateTask::getPlanPartToUse()
{
    return std::make_pair(Sem::PlanningPart(),false);
}

bool PrivateTask::isPathStraight(Robot *r,p3d_traj *t)
{
    API::Trajectory p(r,t);
    API::Trajectory *path = &p;
    bool ok(true);
    Configuration q_save= *mainAgent->getCurrentPos();
    int j_id = mainAgent->getHriAgent()->manip->external_hand[getArmId()]->joint_id_in_rob - 1;
    double length_straight(0),length_path(0);
    double dmax(0);
    double param(0),param_max(0);
    Eigen::Vector3d v1,v2;
    p3d_col_get_dmax(&dmax);

    mainAgent->setAndUpdate(*path->getBegin());
    v1=mainAgent->getJointPos(j_id);
    mainAgent->setAndUpdate(*path->getEnd());
    v2=mainAgent->getJointPos(j_id);
    length_straight = (v1-v2).norm();


    param_max = path->getRangeMax();
    assert(param_max!=P3D_HUGE);
    param=std::min(dmax,param_max);
    while(param<=param_max){
//        cout << param << " max = " <<  param_max << endl;
        mainAgent->setAndUpdate(*path->configAtParam(param));
        v2=mainAgent->getJointPos(j_id);
        length_path += (v1-v2).norm();

        if(param<param_max){
            //add dmax or stop at the end of the traj
            param=std::min(param+dmax,param_max);
        }else{
            //end the loop
            param+=dmax;
        }
        v1=v2;
    }

    if(length_path > length_straight *1.3){
        cout << "end effector local path is "<<length_path/length_straight<<" times longer than straight line"<<endl;
        ok=false;
    }

    mainAgent->setAndUpdate(q_save);

    return ok;
}


