#include "HRICS_CF_TimeToCollision.h"
#include "utils/Chronometer.hpp"
#include "HRICS_costfunction.h"

using namespace std;
namespace HRICS {
namespace CF {

INIT_MOVE3D_STATIC_LOGGER(TimeToCollision,"planners.hrics.cf.timetocollision");

TimeToCollision::TimeToCollision(CostFunctionCommon *commonCF, double min, double max, double low, double up, double maxToMeanRatio, FunctionShape shape):
    VirtualCostFunction(commonCF,min,max,low,up,maxToMeanRatio,shape)
{
}

TimeToCollision::~TimeToCollision()
{

}

double TimeToCollision::getCost(API::Trajectory const & traj){
    double dmax=global_Project->getActiveScene()->getDMax();
    double timetocol(lowerBound());
    Chronometer chrono;
    for (double p=0;p+dmax<=traj.getRangeMax();p+=dmax){
        //at each loops, stop projecting when we can't find a worst time to collision
        timetocol = min(timetocol,this->getTimeToCollision(traj.configAtParam(p),traj.configAtParam(p+dmax),dmax,timetocol));
        if(chrono.getMilliSec() > 10e6){
            M3D_DEBUG("aborting traj cost computation after 10 seconds delay exhausted");
            break;
        }
        M3D_DEBUG("timetocol:"<<timetocol);
    }
    return getCost(timetocol);

}

double TimeToCollision::getCost(){
    return getCost(_common->getRobot()->getCurrentTraj());
}

double TimeToCollision::getCost(std::vector<confPtr_t> &confs,double dt){
    return getTimeToCollision(confs[confs.size()-2],confs[confs.size()-1],dt,upperBound());
}

double TimeToCollision::getTimeToCollision(confPtr_t cprev,confPtr_t ccurrent,double dt,double maxt)
{
    Robot *human,*rob;
    rob=_common->getRobot();
    double t(0);
    bool collision(false);
    while (!collision && t<maxt){
        confPtr_t cnext = this->projectConfiguration(cprev,ccurrent,dt,t);
        cnext->getRobot()->setAndUpdate(*cnext);
        for(unsigned int h=0;h<_common->getHumans().size();h++){
            human=_common->getHumans()[h];
            collision |= pqp_robot_robot_collision_test(rob->getRobotStruct(),human->getRobotStruct());
        }
        t+=dt;
    }
    return t;
}

confPtr_t  TimeToCollision::projectConfiguration(confPtr_t cprev, confPtr_t  ccurrent, double dt, double t){
    //cprev was dt seconds before ccurrent. Returned configuration is the projected one at time t seconds after ccurrent.


    Robot* rob = cprev->getRobot();
    confPtr_t cnext(new Configuration(rob)),cdiff(new Configuration(rob));
    cdiff = *cprev - *ccurrent;

    cnext = *ccurrent + (*cdiff * (t/dt));
    return cnext;
}

int TimeToCollision::test(CostFunctionCommon* common){
    TimeToCollision ttc(common,0,1,0,1,0.5,VirtualCostFunction::LINEAR);
    Robot *rob = global_Project->getActiveScene()->getActiveRobot();
    API::Trajectory traj = rob->getCurrentTraj();

    double dmax=global_Project->getActiveScene()->getDMax();

    M3D_DEBUG("dmax="<<dmax);

    M3D_DEBUG("traj cost: "<<ttc.getCost(traj));

    double rangemax=traj.getRangeMax();
    confPtr_t qnext = ttc.projectConfiguration(traj.getBegin(),traj.configAtParam(rangemax/100),rangemax/100,rangemax/10);
    rob->setAndUpdate(*qnext);
    return 0;
}

} // namespace CF
} // namespace HRICS

