#include "HRICS_distance2d.h"
#include "utils/Geometry.h"
#include "HRICS_costfunction.h"

namespace HRICS {
namespace CF {

INIT_MOVE3D_STATIC_LOGGER(Distance2D,"planners.hrics.distance2d");

Distance2D::Distance2D(CostFunctionCommon *commonCF, double min, double max, double low, double up, double maxToMeanRatio, FunctionShape shape):
    VirtualCostFunction(commonCF,min,max,low,up,maxToMeanRatio,shape)
{
    if(commonCF == NULL){
        throw(std::invalid_argument("NULL CostFunctionCommon"));
    }
}

Distance2D::~Distance2D()
{

}

double Distance2D::getCost(){
    double cost_mean(0), cost_max(0);
    double sum(0);
    double weights_sum(0);

    try{
        Robot *r=this->commonCF()->getRobot();
        HRI_AGENT_TYPE t = r->getHriAgent()->type;
        std::map<int,double> jointWeights = commonCF()->jointWeightsMovingCurrentCost(t);
        for (std::map<int,double>::const_iterator it=jointWeights.begin();it!=jointWeights.end();++it){
            if(it->second!=0){
                Joint* jnt=r->getJoint(it->first);
                Eigen::Vector3d jpos = jnt->getVectorPos();
                Eigen::Vector2d jvect; //2D
                jvect[0]=(jpos.x()); jvect[1] = jpos.y();
                for( unsigned int h=0;h<_common->getHumanPos().size();h++){
                    double c= it->second*getCost((jvect - _common->getHumanPos()[h]).norm());
                    weights_sum += it->second;
                    sum += c;
                    cost_max = std::max(cost_max,c);
                }
            }
        }

        if(weights_sum !=0){
            cost_mean = sum / weights_sum;
        }else{
            cost_mean=_minCost;
            M3D_WARN("no joints specified to compute cost");
        }

        assert(_maxToMeanRatio <=1.);
        return _maxToMeanRatio * cost_max + (1-_maxToMeanRatio) * cost_mean;
    }catch (std::runtime_error e){
        M3D_ERROR(e.what());
    }
    return _maxCost;
}

double Distance2D::getCost(double d)
{
    return std::min(_maxCost,std::max(_minCost,_maxCost + (_minCost - _maxCost)/_ignore_threshold * d));
}

void Distance2D::setMaxToMeanRatio(double a)
{
    _maxToMeanRatio = std::min(1.,std::max(0.,a));
    if (a!=_maxToMeanRatio){
        M3D_WARN("maxToMeanRatio must be in [0,1]");
        throw std::invalid_argument("maxToMeanRatio must be in [0,1]");
    }
}


} // namespace CF
} // namespace HRICS

