#ifndef HRICS_DISTANCE2D_H
#define HRICS_DISTANCE2D_H

#include "API/planningAPI.hpp"
#include "Logging/Logger.h"
#include "HRICS_CF_VirtualCostFunction.h"

namespace HRICS {
namespace CF{

class Distance2D :public VirtualCostFunction
{
    MOVE3D_STATIC_LOGGER;
public:
    using VirtualCostFunction::getCost;

    Distance2D(CostFunctionCommon *commonCF,double min, double max, double low,double up,double maxToMeanRatio, FunctionShape shape);
    virtual ~Distance2D();

    static std::string getName(){return "Distance2D";}

    double getCost();
    double getCost(double d);

    /**
     * @brief setMaxToMeanRatio
     * @param a
     * @throw invalid_argument when a not in [0,1]
     */
    void setMaxToMeanRatio(double a);
    double getMaxToMeanRatio(){return _maxToMeanRatio;}

    void setMaxCost(double max){_maxCost = max;}
    void setMinCost(double min){_minCost = min;}
    double getMaxCost(){return _maxCost;}
    double getMinCost(){return _minCost;}

private:

    double _ignore_threshold;
    double _minCost; /// the cost to apply where the robot is at more than _ignore_threshold from humans
    double _maxCost; /// the cost when the robot is on the human
    double _maxToMeanRatio; ///< the ratio "a" when using getCost, computes the cost as c = a*cmax + (1-a)*cmean. needs to be in [0,1]

};

} // namespace CF
} // namespace HRICS

#endif // HRICS_DISTANCE2D_H
