#ifndef MOVE3D_LOG_OFF

#include "plotsocketappender.h"

#include <log4cxx/helpers/loglog.h>
#include <log4cxx/helpers/optionconverter.h>
#include <log4cxx/helpers/stringhelper.h>
#include <log4cxx/spi/loggingevent.h>
#include <log4cxx/helpers/synchronized.h>
#include <log4cxx/helpers/objectoutputstream.h>
//#include <apr_time.h>
//#include <apr_atomic.h>
//#include <apr_thread_proc.h>
#include <log4cxx/helpers/socketoutputstream.h>


using namespace log4cxx;
using namespace log4cxx::helpers;

// Register this class with log4cxx
IMPLEMENT_LOG4CXX_OBJECT(PlotSocketAppender)
INIT_MOVE3D_STATIC_LOGGER(PlotSocketAppender,"planners.logger.plotsocketappender");

// The default port number of remote logging server (4560)
int PlotSocketAppender::DEFAULT_PORT                 = 4560;

// The default reconnection delay (3000 milliseconds or 3 seconds).
int PlotSocketAppender::DEFAULT_RECONNECTION_DELAY   = 3000;

PlotSocketAppender::PlotSocketAppender()
: SocketAppenderSkeleton(DEFAULT_PORT, DEFAULT_RECONNECTION_DELAY) {
}

PlotSocketAppender::PlotSocketAppender(InetAddressPtr& address1, int port1)
: SocketAppenderSkeleton(address1, port1, DEFAULT_RECONNECTION_DELAY) {
    Pool p;
    activateOptions(p);
}

PlotSocketAppender::PlotSocketAppender(const LogString& host, int port1)
: SocketAppenderSkeleton(host, port1, DEFAULT_RECONNECTION_DELAY) {
    Pool p;
    activateOptions(p);
}

PlotSocketAppender::~PlotSocketAppender()
{
    finalize();
}

int PlotSocketAppender::getDefaultDelay() const {
    return DEFAULT_RECONNECTION_DELAY;
}

int PlotSocketAppender::getDefaultPort() const {
    return DEFAULT_PORT;
}

void PlotSocketAppender::setSocket(log4cxx::helpers::SocketPtr& socket, Pool& p) {
    synchronized sync(mutex);
    oos = new ObjectOutputStream(new SocketOutputStream(socket), p);
}

void PlotSocketAppender::cleanUp(Pool& p) {
    if (oos != 0) {
        try {
            oos->close(p);
            oos = 0;
        } catch(std::exception& e) {
        }
    }
}

void PlotSocketAppender::append(const log4cxx::spi::LoggingEventPtr& event, Pool& p)
{
       if ( this->layout == NULL ) {
             LOG4CXX_ENCODE_CHAR(nameStr, name);
             std::string msg("No Layout set for the appender named [ ");
             msg.append(nameStr);
             msg.append(" ].");

             LOG4CXX_DECODE_CHAR(msgL, msg);
             errorHandler->error(msgL);
           return;
       }

       log4cxx::LogString fMsg;

       this->layout->format(fMsg, event, p);

       if (oos != 0) {
           try {
              //event->write(*oos, p);
              oos->writeUTFString(fMsg.c_str(),p);
              oos->flush(p);
           } catch(std::exception& e) {
              oos = 0;
              LogLog::warn(LOG4CXX_STR("Detected problem with connection: "), e);
              if (getReconnectionDelay() > 0) {
                  fireConnector();
              }
           }
       }


       LOG4CXX_ENCODE_CHAR(fMsgStr, fMsg);

       // This example appender prints to the screen just like
       // a console appender. Do something with the data here
       //std::cout<<"PlotSocketAppender: " <<fMsgStr.c_str();

       //LogVector.push_back(fMsgStr);
}


//void PlotSocketAppender::close()
//{
//    if (this->closed)
//    {
//        return;
//    }
//
//    this->closed = true;
//}

#endif //MOVE3D_LOG_OFF
