#include "ent_entityManager.hpp"
#include "API/project.hpp"
#include "VimanTags.h"

namespace Ent {

INIT_MOVE3D_STATIC_LOGGER(EntityManager,"planners.entities.manager");

EntityManager *EntityManager::__mainEntityManager = 0;

EntityManager::EntityManager()
{
    std::string path(getenv("HOME_MOVE3D"));
    path+="/EntityTypes";
    setGlobalJsonPath(path);
    createModules();
}

EntityManager::~EntityManager()
{
}

EntityManager *EntityManager::mainEntityManager()
{
    if(__mainEntityManager ==0){
        M3D_INFO("creating new main entity Manager");
        __mainEntityManager=new EntityManager();
    }
    return __mainEntityManager;
}

void EntityManager::createModules()
{
    M3D_DEBUG("creating modules");
    setManipulables(new Manipulable(globalJsonPath()));
    setSurfaces(new Surface(globalJsonPath()));
    setAgents(new Agent(globalJsonPath()));
    setVerticals(new Vertical(globalJsonPath()));
    setVirtuals(new Virtual(globalJsonPath()));
}

bool EntityManager::updateEntities()
{
    bool ok = fetchData();
    M3D_INFO("list of retrieved viman tags");
    for (unsigned int i = 0; i < global_Project->getActiveScene()->getNumberOfRobots(); i++)
    {
        Robot* r = global_Project->getActiveScene()->getRobot(i);
        if (r->getObjectRob()->isObject())
        {
            unsigned int tags=VimanTags::readTag(r,r->getObjectRob());
            M3D_INFO("robot "<< r->getName() << " has " << tags << " viman tags");
        }
    }
    return ok;
}

bool EntityManager::fetchData()
{
    M3D_DEBUG("reading files");
    bool ok = true;
    ok = ok && manipulables()->fetchData();
    ok = ok && surfaces()->fetchData();
    ok = ok && agents()->fetchData();
    ok = ok && verticals()->fetchData();
    ok = ok && virtuals()->fetchData();

    return ok;
}

void EntityManager::setManipulables(Manipulable *manip)
{
    _manipulable = manip;
}


Manipulable *EntityManager::manipulables() const
{
    return _manipulable;
}

void EntityManager::setSurfaces(Surface *surface)
{
    _surface = surface;
}


Surface *EntityManager::surfaces() const
{
    return _surface;
}

void EntityManager::setAgents(Agent *agent)
{
    _agent = agent;
}

Agent *EntityManager::agents() const
{
    return _agent;
}

void EntityManager::setVerticals(Vertical *vertical)
{
    _vertical = vertical;
}

Vertical *EntityManager::verticals() const
{
    return _vertical;
}


void EntityManager::setVirtuals(Virtual *virtua)
{
    _virtual = virtua;
}

Virtual *EntityManager::virtuals() const
{
    return _virtual;
}




std::string EntityManager::globalJsonPath() const
{
    return _globalJsonPath;
}

void EntityManager::setGlobalJsonPath(const std::string &globalJsonPath)
{
    _globalJsonPath = globalJsonPath;
}


}
