#ifndef ENT_ENTITYMANAGER_H
#define ENT_ENTITYMANAGER_H

#include "Logging/Logger.h"
#include "ent_manipulable.hpp"
#include "ent_surface.hpp"
#include "ent_agent.hpp"
#include "ent_vertical.hpp"
#include "ent_virtual.hpp"

namespace Ent {


/**
  * @brief EntityManager: Manage all the entity type reader classes
  */
class EntityManager
{
    MOVE3D_STATIC_LOGGER;
public:
    EntityManager();
    ~EntityManager();

    static EntityManager* mainEntityManager();

    Manipulable *manipulables() const;
    void setManipulables(Manipulable *manip);

    Surface *surfaces() const;
    void setSurfaces(Surface *surface);

    Agent *agents() const;
    void setAgents(Agent *agent);

    Vertical *verticals() const;
    void setVerticals(Vertical *vertical);

    Virtual *virtuals() const;
    void setVirtuals(Virtual *virtua);


    std::string globalJsonPath() const;
    void setGlobalJsonPath(const std::string &globalJsonPath);

    bool updateEntities();
    bool fetchData();

protected:
    void createModules();


private:
    static EntityManager *__mainEntityManager;

    std::string _globalJsonPath;

    Manipulable* _manipulable;
    Surface* _surface;
    Agent* _agent;
    Vertical* _vertical;
    Virtual* _virtual;

};



} //namespace Ent

#endif // ENT_ENTITYMANAGER_H
