#include "ent_manipulable.hpp"
#include "project.hpp"
#include <fstream>

namespace Ent {


INIT_MOVE3D_STATIC_LOGGER(Manipulable, "planners.entities.manipulable");

Manipulable::Manipulable(std::string global_json_path):
    _global_json_path(global_json_path)
{
}

Manipulable::~Manipulable()
{
}




std::string Manipulable::getPath()
{
    std::string file=_global_json_path;
    file.append("/Manipulable.json");
    return file;
}


bool Manipulable::fetchData()
{
    bool ok = importJsonFile(getPath());
    return ok;
}


bool Manipulable::importJsonFile(const std::string &path)
{
    std::ifstream ifs;
    Json::Value root;
    bool ok=true;

    ifs.open(path.c_str());

    if(!ifs){
        ok=false;
        M3D_ERROR("could not open "<<path);
    }
    if(ok){
        try{
            ifs>>root;
        }catch (std::runtime_error e){
            M3D_ERROR("error while parsing file "<<path<<": "<<e.what());
            ok=false;
        }catch(std::exception e){
            M3D_ERROR("error while parsing file "<<path<<": "<<e.what());
            ok=false;
        }
    }
    if(ok){
        if(!root){
            M3D_ERROR("could not parse json file "<<path);
            ok=false;
        }
    }
    if(ok){
        if(!createManipulableObjects(root)){
            ok=false;
            M3D_ERROR("could not properly parse objects in file "<<path);
        }
    }
    return ok;
}


bool Manipulable::createManipulableObjects(Json::Value const & value)
{
    bool ok = true;
    M3D_INFO("List of manipulable entities:");
    for(Json::Value::iterator it = value.begin(); it !=value.end(); ++it)
    {
        //        Json::Value key = it.key();
        Json::Value cont = (*it);



        if(cont.isMember("complete_name"))
        {
            for(Json::Value::iterator it2 = cont["complete_name"].begin(); it2 !=cont["complete_name"].end(); ++it2)
            {
                if (global_Project->getActiveScene()->getRobotByName((*it2).asString()))
                {
                    Robot* r = global_Project->getActiveScene()->getRobotByName((*it2).asString());
                    if (r)
                    {
                        M3D_INFO(r->getName());
                        ok = ok && setToObjRob(r, cont["rotations"]);
                    }
                }
            }
        }
        if (cont.isMember("partial_name"))
        {
            for(Json::Value::iterator it2 = cont["partial_name"].begin(); it2 !=cont["partial_name"].end(); ++it2)
            {
                std::vector<Robot*> vr = global_Project->getActiveScene()->getAllRobotWithNameContaining((*it2).asString());
                for (unsigned int i = 0; i < vr.size();i++)
                {
                    M3D_INFO(vr.at(i)->getName());
                    ok = ok && setToObjRob(vr.at(i), cont["rotations"]);
                }
            }
        }

        if (!cont.isMember("complete_name") && !cont.isMember("partial_name") && !cont.isMember("_comment"))
        {
            M3D_ERROR("complete_name and partial_name does not exit for an entry");
            ok = false;
        }
    }

    return ok;
}

bool Manipulable::setToObjRob(Robot* rob, Json::Value const & value)
{
    configPt q_cur = p3d_get_robot_config(rob->getRobotStruct());
    bool ok = true;
    for(Json::Value::const_iterator it = value.begin(); it !=value.end(); ++it)
    {
        configPt q = p3d_copy_config(rob->getRobotStruct(),q_cur);
        int i =6;
        for(Json::Value::const_iterator it2 = (*it).begin(); it2 !=(*it).end(); ++it2)
        {
            if ((*it2).isNull())
                q[i] = OPEN_VALUE;
            else
            {
                if (i < 9)
                    q[i] = (*it2).asDouble();
                else
                    q[i] = (*it2).asDouble()*M_PI/180;
            }
            i++;
        }
        if (i > rob->getRobotStruct()->nb_dof)
        {
            ok = false;
            M3D_ERROR("nb dofs for entity "<<rob->getName()<< " should be: "<<rob->getRobotStruct()->nb_dof<<" received: "<<i);
        }
        rob->getObjectRob()->addStableRotation(q);

    }

    return ok;
}



}
