#include "ent_vertical.hpp"
#include "project.hpp"
#include <fstream>

namespace Ent {


INIT_MOVE3D_STATIC_LOGGER(Vertical, "planners.entities.vertical");

Vertical::Vertical(std::string global_json_path):
    _global_json_path(global_json_path)
{
}

Vertical::~Vertical()
{
}

std::string Vertical::getPath()
{
    std::string file=_global_json_path;
    file.append("/Vertical.json");
    return file;
}


bool Vertical::fetchData()
{
    bool ok = importJsonFile(getPath());
    return ok;
}


bool Vertical::importJsonFile(const std::string &path)
{
    std::ifstream ifs;
    Json::Value root;
    bool ok=true;

    ifs.open(path.c_str());

    if(!ifs){
        ok=false;
        M3D_ERROR("could not open "<<path);
    }
    if(ok){
        try{
            ifs>>root;
        }catch (std::runtime_error e){
            M3D_ERROR("error while parsing file "<<path<<": "<<e.what());
            ok=false;
        }catch(std::exception e){
            M3D_ERROR("error while parsing file "<<path<<": "<<e.what());
            ok=false;
        }
    }
    if(ok){
        if(!root){
            M3D_ERROR("could not parse json file "<<path);
            ok=false;
        }
    }
    if(ok){
        if(!createVerticalConst(root)){
            ok=false;
            M3D_ERROR("could not properly parse objects in file "<<path);
        }
    }
    return ok;
}


bool Vertical::createVerticalConst(Json::Value const & value)
{
    bool ok = true;
    M3D_INFO("List of vertical entities:");
    for(Json::Value::iterator it = value.begin(); it !=value.end(); ++it)
    {
        //        Json::Value key = it.key();
        Json::Value cont = (*it);

        if(cont.isMember("complete_name"))
        {
            for(Json::Value::iterator it2 = cont["complete_name"].begin(); it2 !=cont["complete_name"].end(); ++it2)
            {
                if (global_Project->getActiveScene()->getRobotByName((*it2).asString()))
                {
                    Robot* r = global_Project->getActiveScene()->getRobotByName((*it2).asString());
                    if (r)
                    {
                        M3D_INFO(r->getName());
                        ok = ok && setToObjRob(r, cont);
                    }
                }
            }
        }
        if (cont.isMember("partial_name"))
        {
            for(Json::Value::iterator it2 = cont["partial_name"].begin(); it2 !=cont["partial_name"].end(); ++it2)
            {
                std::vector<Robot*> vr = global_Project->getActiveScene()->getAllRobotWithNameContaining((*it2).asString());
                for (unsigned int i = 0; i < vr.size();i++)
                {
                    M3D_INFO(vr.at(i)->getName());
                    ok = ok && setToObjRob(vr.at(i), cont);
                }
            }
        }

        if (!cont.isMember("complete_name") && !cont.isMember("partial_name") && !cont.isMember("_comment"))
        {
            M3D_ERROR("complete_name and partial_name does not exit for an entry");
            ok = false;
        }
    }
    return ok;
}

bool Vertical::setToObjRob(Robot* rob, Json::Value const & value)
{
    rob->getObjectRob()->setHasVerticalityConstraint(true);
    rob->getObjectRob()->setMaxVerticalTolerance(value["angle_max"].asDouble()*M_PI/180);
    Eigen::Vector3d v;
    bool ok = true;
    if (!value.isMember("vector"))
    {
        ok = false;
        M3D_ERROR("vector not available for entity "<<rob->getName());
    }
    if (ok & value["vector"].isArray() && value["vector"].size() == 3)
    {
        v[0] = value["vector"][0].asDouble();
        v[1] = value["vector"][1].asDouble();
        v[2] = value["vector"][2].asDouble();
    }
    else if (!value["vector"].isArray())
    {
        ok = false;
        M3D_ERROR("vector not an array for entity "<<rob->getName());
    }
    else if (value["vector"].size() != 3)
    {
        ok = false;
        M3D_ERROR("vector for entity "<<rob->getName()<<" greater than 3");
    }
    rob->getObjectRob()->setKeepVerticalAxis(v);
    return ok;
}



}
