#include <stdlib.h>
#include "Sem_AgentType.hpp"
#include <stdlib.h>



namespace Sem {
using namespace std;

AgentType::AgentType():
    _isDef(false)
{
}

AgentType::AgentType(HRI_AGENT_TYPE t):
    _val(t),
    _isDef(true)
{
    assert( _agent_types_str.count(t) != 0 );
    if (_agent_types_str.count(t) == 0 ){
        std::ostringstream oss;
        oss << "no string defined for agent type " << t;
        throw (std::runtime_error(oss.str()));
    }
}

AgentType::AgentType(string s)
{
    _isDef = getAgentTypeFromStr(s,_val);
    //assert(_isDef);
    //if(!_isDef){
    //    std::ostringstream oss;
    //    oss << "no matching agent type value for string " << s;
    //    throw (std::runtime_error(oss.str()));
    //}
}

AgentType::~AgentType()
{
}

AgentType AgentType::fromRobot(Robot *rob)
{
    if(rob && rob->getHriAgent()){
        return AgentType(rob->getHriAgent()->type);
    }else{
        return AgentType();
    }
}

bool AgentType::isDefined() const
{
    return _isDef;
}

std::map<HRI_AGENT_TYPE,std::string> createAgentTypeStr(){
    std::map<HRI_AGENT_TYPE,std::string> agent_types_str;
    agent_types_str[HRI_PR2]="PR2";
    agent_types_str[HRI_BALL_ROBOT]="BALL_ROBOT";
    agent_types_str[HRI_JIDOKUKA]="JIDOKUKA";
    agent_types_str[HRI_KUKAROBOT]="KUKAROBOT";
    agent_types_str[HRI_HERAKLES]="HERAKLES";
    agent_types_str[HRI_ACHILE]="ACHILE";
    agent_types_str[HRI_ROMEO]="ROMEO";
    return agent_types_str;
}

bool AgentType::getStrAgentType(HRI_AGENT_TYPE t,std::string &s){
    bool ok=(_agent_types_str.find(t) != _agent_types_str.end());
    if (ok){
        s=_agent_types_str[t];
    }
    return ok;
}

string AgentType::agentTypeToString(HRI_AGENT_TYPE type)
{
    std::string s;
    bool ok=(_agent_types_str.find(type) != _agent_types_str.end());
    if (ok){
        s=_agent_types_str[type];
    }
    return s;
}

string AgentType::agentTypeToHumanString(HRI_AGENT_TYPE type)
{
    std::string s=agentTypeToString(type);
    if (s.size()==0){
        ostringstream os;
        os<<static_cast<int>(type);
        s=os.str();
    }
    return s;
}

HRI_AGENT_TYPE AgentType::toHriAgentType() const
{
    assert(isDefined());
    if (!isDefined()){
        throw(std::runtime_error("asking HRI_AGENT_TYPE of an undefined AgentType"));
    }
    return _val;
}

string AgentType::toString() const
{
    if (!isDefined()){
        return "undef!";
    }else{
        return _agent_types_str[_val];
    }
}

bool operator==(const AgentType &t1, const AgentType &t2)
{
    // equals if both defined and values are equal, or if both not defined (no matter the values).
    return (t1._isDef == t2._isDef ) && ( (!t1._isDef) || (t1._val == t2._val) );
}

bool operator!=(const AgentType &t1, const AgentType &t2)
{
    return !(t1==t2);
}

bool operator<(const AgentType &t1, const AgentType &t2)
{
    if(t1._isDef && t2._isDef){
        return t1._val < t2._val;
    }else if (t1._isDef != t2._isDef){
        return t1._isDef < t2._isDef;
    }else{
        //both not def
        return 0<0;
    }
}
bool operator>(const AgentType &t1, const AgentType &t2)
{
    if(t1._isDef && t2._isDef){
        return t1._val > t2._val;
    }else if (t1._isDef != t2._isDef){
        return t1._isDef > t2._isDef;
    }else{
        //both not def
        return 0>0;
    }
}

bool AgentType::getAgentTypeFromStr(std::string const &s, HRI_AGENT_TYPE &type){
    map<HRI_AGENT_TYPE,string>::iterator tstr=_agent_types_str.begin();
    while (tstr->second != s && tstr != _agent_types_str.end()){ ++tstr;}
    bool ok=(tstr!=_agent_types_str.end());
    if(ok){
        type=tstr->first;
    }
    return ok;
}

std::map<HRI_AGENT_TYPE,std::string> AgentType::_agent_types_str = createAgentTypeStr();

} // namespace Sem

