#include "MultiHandOverAlgos.h"
#include "HRI_costspace/HRICS_MultiHandOver.h"
#include "Chronometer.hpp"

#include "Collision-pkg.h"
#include "planner/planner.hpp"
#include "planEnvironment.hpp"

#include "planner/TrajectoryOptim/plannarTrajectorySmoothing.hpp"
#include "API/project.hpp"
#include "plannerFunctions.hpp"
#include "Planner-pkg.h"
#include "LightPlanner-pkg.h"

namespace HRICS {
namespace MHO{
using namespace Eigen;
using namespace std;
using namespace tr1;

// //////////////// TOOL SET ///////////////
ToolSet::ToolSet(FrontierTester *frontier_tester, AbstractSolutionTools *sol_tools, HRICS::MultiHandOver *mho,FrontierCells::FrontierCheckStatus max_check_status):
    solutionTools(sol_tools),
    frontierTester(frontier_tester),
    _mho(mho),
    _max_check_status(max_check_status),
    _full_ho_cnt(0), _full_ho_time_mean(0), _full_ho_time_M2(0)
{}
void ToolSet::setRobots(unsigned int a1, unsigned int a2){
    assert(max(a1,a2)<_mho->getAgents().size());
    //handoverObjectChecker->robots(_mho->getAgents()[a1],_mho->getAgents()[a2]);
    //handoverObjectChecker->reach(_mho->getAgentStruct(a1)->reach(),_mho->getAgentStruct(a2)->reach());
}
void ToolSet::updateObject(){
    //handoverObjectChecker->object(_mho->getObject());
}
bool ToolSet::checkHandOverFull(MultiAgentCell *c1, unsigned int a1, MultiAgentCell *c2, unsigned int a2,double &cost,double *time){
    bool ok(0),found(0);
    this->_full_ho_cnt++;
    double delta=0;
    Chronometer chrono;
    vector<FrontierPtr> frontiers=c1->createAllRealFrontiersFor(a1,a2);
    //vector<FrontierPtr> &frontiers=c1->getAgentAt(a1)->frontiers(c2->getAgentAt(a2));
    for(vector<FrontierPtr>::iterator it=frontiers.begin();it!=frontiers.end();++it){
        assert(c1==(*it)->cellOf(a1) || c1->getAgentAt(a1)->agentInGrid->getMetaAgent() == c1->getAgentAt(a2)->agentInGrid->getMetaAgent());
        if((*it)->cellOf(a2)==c2){
            //found a frontier corresponding to the one asked
            found=true;
            ok=this->checkFrontierFull(*it,c1,a1,c2,a2);
            if(ok){
                cost=solutionTools->handOverCost(*it,time);
                break;
            }
        }
    }
    if(!found){
        FrontierPtr frontier;
        frontier=c1->addFrontierComplete(c2,!this->_mho->getPlanGrid()->areCellsInNeighbourhood(c1,c2,a1,c1->computeDist(c2)*sqrt(2)),
                                         a1,a2,c1->computeDist(c2),false);
        ok=this->checkFrontierFull(frontier,c1,a1,c2,a2);
        cost=solutionTools->handOverCost(frontier,time);
    }
    if(!ok){
        cost=numeric_limits<double>::infinity();
    }
    //STAT FOR TIME
    double t=chrono.getMilliSec();
    delta=t-_full_ho_time_mean;
    _full_ho_time_mean+=delta/_full_ho_cnt;
    _full_ho_time_M2+= delta*(t-_full_ho_time_mean);

    return ok;
}

double ToolSet::getStatsCheckHoFull(double &mean,double &variance,unsigned int &n){
    mean=_full_ho_time_mean;
    n=_full_ho_cnt;
    variance = _full_ho_time_M2/(n-1);
    return _full_ho_time_mean;
}

bool ToolSet::checkFrontierFull(FrontierPtr frontier,MultiAgentCell *c1,unsigned int a1,MultiAgentCell *c2, unsigned int a2){
    bool ok(0);
    bool log_this=false;//(_mho->getAgentStruct(a1)->is_human || _mho->getAgentStruct(a2)->is_human);
    if(a2!=frontier->agentB_ind){
        swap(a1,a2);
        swap(c1,c2);
    }
    assert(frontier->cellB()==c2 && frontier->cellA()==c1);
    if(frontier->agentA_ind==a2){
        //reverse frontier
        frontier->select_reverse=true;
    }else{
        frontier->select_reverse=false;
    }
    ok=frontierTester->test(frontier,_max_check_status);
    if(ok && log_this){
        cout<<"MHO checkFrontierFull : OK"<<endl;
    }
    return ok;
}

double ToolSet::getReach(unsigned int a){
    return _mho->getAgentStruct(a)->reach();
}
double ToolSet::getReach(unsigned int a1, unsigned int a2){
    return this->getReach(a1)+getReach(a2);
}

INIT_MOVE3D_STATIC_LOGGER(SolutionTools,"planners.hrics.mho.soltools");
SolutionTools::SolutionTools(MultiHandOver *mho) :
    AbstractSolutionTools(),
    _mho(mho)
{ }

double SolutionTools::computeCost(MultiHandOverSolution *solution){
    return this->_mho->computeSolutionCost(solution);
}
double SolutionTools::computeNavigationTrajectories(MultiHandOverSolution *solution,int action_index){
    return this->_mho->computeNavigationTrajectories(solution,action_index);
}
//bool SolutionTools::handOverFindConfs(MultiHandOverSolution *solution,int frontier_index){
//    return this->_mho->handOverFindConfs(solution,frontier_index);
//}
//bool SolutionTools::handOverFindConfs(FrontierPtr frontier){
//    return this->_mho->handOverFindConfs(frontier);
//}
//bool SolutionTools::handOverFindConfs(MultiAgentCell *c1, unsigned int a1,
//                                      MultiAgentCell *c2, unsigned int a2){
//    FrontierPtr frontier;
//    //HandOverWindowCheckerRRT ho_checker()
//    vector<FrontierPtr> frontiers=c1->createAllRealFrontiersFor(a1,a2);
//    //vector<FrontierPtr> &frontiers=c1->getAgentAt(a1)->frontiers(c2->getAgentAt(a2));
//    for(vector<FrontierPtr>::iterator it=frontiers.begin();it!=frontiers.end();++it){
//        if((*it)->cellOf(a2)==c2 && (*it)->status_ok[FrontierCells::FCS_OBJECT]){
//            //TODO check for object when its not done before
//            frontier=*it;
//            break;
//        }
//    }
//    return frontier && this->handOverFindConfs(frontier);
//}
double SolutionTools::distanceCostNoTime(double d, unsigned int a){
    double c(numeric_limits<double>::infinity());
    double tf,uf,df,hf;//time, use, distance, hri cost factors
    _mho->getCostsFactors(df,tf,uf,hf);
    if(d!=numeric_limits<double>::infinity()){ // avoid inf * 0 operation, resulting in NaN
        c=uf * _mho->getAgentUseCost(a) * d / _mho->getAgentSpeed(a) + df * d;
    }
    assert(c>=0);//not negative, not NaN, may be inf
    return c;
}

double SolutionTools::distanceCost(double d, unsigned int a){
    double tf,uf,df,hf;//time, use, distance, hri cost factors
    _mho->getCostsFactors(df,tf,uf,hf);
    return uf * _mho->getAgentUseCost(a) * d / _mho->getAgentSpeed(a) + tf * d / _mho->getAgentSpeed(a) + df * d;
}
double SolutionTools::distanceTime(double d, unsigned int a){
    double tf,uf,df,hf;//time, use, distance, hri cost factors
    _mho->getCostsFactors(df,tf,uf,hf);
    return d / _mho->getAgentSpeed(a);
}
double SolutionTools::timeCost(double t, unsigned int a){
    double tf,uf,df,hf;//time, use, distance, hri cost factors
    _mho->getCostsFactors(df,tf,uf,hf);
    return t * _mho->getAgentUseCost(a) * uf + tf * t;
}
double SolutionTools::objectTimeCost(double t){
    double tf,uf,df,hf;//time, use, distance, hri cost factors
    _mho->getCostsFactors(df,tf,uf,hf);
    return  tf * t;
}

double SolutionTools::handOverCostEstimation(unsigned int i,unsigned int j){
    double tf,uf,df,hf;//time, use, distance, hri cost factors
    _mho->getCostsFactors(df,tf,uf,hf);
    double cost(0),time(0);
    MultiHandOverHO::getOptimalCost(_mho->getAgentStruct(i),_mho->getAgentStruct(j),cost,time,this->_mho);
    return cost * hf + time *  tf + time * uf *(_mho->getAgentUseCost(i) + _mho->getAgentUseCost(j));

    //if(_mho->isHuman(i) && _mho->isHuman(j)){
    //    //only hri, time is short
    //    cost+=hf;
    //}else if(_mho->isHuman(i) != _mho->isHuman(j)){
    //    //both hri and time, as it would be slow
    //    cost+= hf + tf;
    //}else if(!_mho->isHuman(i) && !_mho->isHuman(j)){
    //    //no hri as there is no human, but time
    //    cost+= tf;
    //}
    //add use cost for hand over time
    //cost += uf * (_mho->getAgentUseCost(i) + _mho->getAgentUseCost(j) );
    //return cost;
}

double SolutionTools::handOverCost(FrontierPtr frontier, double *time){
    return handOverCost(frontier->cellA(),frontier->agentA_ind,frontier->cellB(),frontier->agentB_ind,time);
}

double SolutionTools::handOverCost(MultiAgentCell *c1, unsigned int a1, MultiAgentCell *c2, unsigned int a2, double *time){
    double tf,uf,df,hf;//time, use, distance, hri cost factors
    double c,t;
    _mho->getCostsFactors(df,tf,uf,hf);
    MultiHandOverHO ho(_mho->getAgentStruct(a1),_mho->getAgentStruct(a2),c1,c2,this->_mho);
    ho.computeCost();
    //cost= hri_cost * hfactor + time * (a1.useCost + a2.useCost)
    t= ho.getTime();
    c= ho.getCost() * hf + tf * t + uf * ho.getTime()*(_mho->getAgentUseCost(a1) + _mho->getAgentUseCost(a2));
    if(time) *time=t;
    return c;
}

double SolutionTools::handOverMaxDist(unsigned int a1, unsigned int a2){
   return this->_mho->getAgentStruct(a1)->reach() + this->_mho->getAgentStruct(a2)->reach();
}

double SolutionTools::handOverSideCosts(MultiAgentCell *c1, unsigned int a1, MultiAgentCell *c2, unsigned int a2, double *coming_time){
    double c;

    double d1(0);
    if(_mho->getAgentStruct(a1)->have_final_position){
        d1=_mho->getAgentTargetCell(a1)->getDistanceAny(a1,c1);//a1 go to its target position
    }
    double d2=c2->getDistance(a2);//a2 come from its starting position
    //TO DO DONE improve this to be more efficient!
    //c= distanceCost(d1,a1) - timeCost(distanceTime(d1,a1),a1) + distanceCost(d2,a2) - timeCost(distanceTime(d2,a2),a2);
    //DONE:
    c= distanceCostNoTime(d1,a1) + distanceCostNoTime(d2,a2);
    if(coming_time) *coming_time = distanceTime(d2,a2);
    //M3D_DEBUG("sidecosts - dists: "<< d1 << " , " <<d2);
    //assert(c < numeric_limits<double>::infinity());
    assert(! std::isnan(c));//not NaN
    return c;

}

FrontierOptimizerClimber::FrontierOptimizerClimber(MultiAgentGrid* grid,
                                                   ToolSet *tool_set):
    AbstractFrontierOptimizer(),
    _grid(grid), _tool_set(tool_set)//_solution_tools(tool_set), _ho_checker(ho_check)
{ }

MultiHandOverSolution* FrontierOptimizerClimber::optimize(MultiHandOverSolution *solution, unsigned int action_index, double keep_distance, double min_distance)
{
    MultiHandOverSolution best_solution=(*solution), initial_solution=*solution;
    Chronometer chrono;
    vector<unsigned long> times(6);
    bool done(false);
    bool use_alt_distance=true;
    bool find_2d_optimal_before=true;//first find optimal in 2d space, then form this position, searches in 3d space
    bool no_check_3d=false;// when true, do not take into account 3d geometry (HO confs) before cost computation
    bool find_point_2d=true;
    bool best_sol_is_full_solution=true;
    double best_cost=solution->cost,initial_cost=solution->cost;
    unsigned int steps=0;

    Eigen::Vector3d start_pos,end_pos;
    start_pos[2]=solution->actions[action_index].agent(0)->shoulderHeight();
    end_pos[2]=solution->actions[action_index].agent(1)->shoulderHeight();

    //_ho_checker->reach(solution->actions[action_index].agent(0)->reach(),solution->actions[action_index].agent(1)->reach());

    cout << "optimize frontier of cost "<<solution->cost<<endl;
    if(use_alt_distance){
        _grid->computeAltDistancesForAgent(solution->actions[action_index].from(0),solution->actions[action_index].agent(0)->index);
        _grid->computeAltDistancesForAgent(solution->actions[action_index].from(1),solution->actions[action_index].agent(1)->index);
        if(action_index+1<solution->actions.size()){
            _grid->computeAltDistancesForAgent(solution->actions[action_index+1].to(0),solution->actions[action_index+1].agent(0)->index);
        }
//#ifdef DRAW_OPTIMIZE_FRONTIER
//        if(_draw_optimize_frontier){
//            vector<pair<MultiAgentCell*,unsigned int> > to_draw;
//            to_draw.push_back(make_pair(solution->actions[action_index].from(0),solution->actions[action_index].agent(0)->index));
//            to_draw.push_back(make_pair(solution->actions[action_index].from(1),solution->actions[action_index].agent(1)->index));
//            if(action_index+1<solution->actions.size()){
//                to_draw.push_back(make_pair(solution->actions[action_index+1].to(0),solution->actions[action_index+1].agent(0)->index));
//            }
//            _grid->setDrawGridForOptim(true,to_draw);
//            g3d_draw_allwin_active();
//        }
//#endif
    }
    if(find_point_2d){
        find_2d_optimal_before=false;//both can't be set to true simultaneously
        MultiAgentCell* starting_cell=findOptimalCellNullDistanceFrontier(solution,action_index);
        if(starting_cell){
            unsigned int a0,a1;
            a0=solution->actions[action_index].agent(0)->index;
            a1=solution->actions[action_index].agent(1)->index;

            vector<MultiAgentCell*> neigh = starting_cell->findNeighborsInRange(a1,this->_grid->getHandOverReach(a0,a1));
            FrontierPtr ftr;
            random_shuffle(neigh.begin(),neigh.end());
            bool found=false;
            while(!found && neigh.size()){
                MultiAgentCell* other=neigh.back();
                neigh.pop_back();
                abort();//not sure about the above addFrontierComplete() 2nd argument at true
                ftr=starting_cell->addFrontierComplete(other,true,a0,a1,starting_cell->computeDist(other),true);//last arg (true) is to check if frontier exists before adding

                _logger->drawcells(starting_cell,other,false);
                found=_tool_set->frontierTester->test(ftr,FrontierCells::FCS_GRAB_CONFS);
                if(found){
                    solution->frontiers[action_index]=ftr;
                    //reset data of the solution
                    solution->cost=0;
                    solution->handover_times[action_index]=0;
                    this->_tool_set->solutionTools->computeNavigationTrajectories(solution,action_index);
                    this->_tool_set->solutionTools->computeCost(solution);//TODO check for LOCAL cost improvment
                }
                //start_pos[0]=ftr->cellA()->getCenter()[0];
                //start_pos[1]=ftr->cellA()->getCenter()[1];
                //end_pos[0]=ftr->cellB()->getCenter()[0];
                //end_pos[1]=ftr->cellB()->getCenter()[1];
                //_ho_checker->shoulders(start_pos,end_pos);
                //if(_ho_checker->check()){
                //    ftr->object_position=_ho_checker->getQMiddle();
                //    ftr->status_ok[FrontierCells::FCS_OBJECT]=true;
                //    ftr->check_status=std::max(FrontierCells::FCS_OBJECT,ftr->check_status);
                //    ftr->grabA=_ho_checker->grab1();
                //    ftr->grabB=_ho_checker->grab2();
                //    solution->frontiers[action_index]=ftr;
                //    //reset data of the solution
                //    solution->cost=0;
                //    solution->handover_times[action_index]=0;
                //    //new_solution->actions.clear();
                //    this->_solution_tools->computeNavigationTrajectories(solution,action_index);
                //    //check if there is a grab conf ok for this frontier
                //    if(this->_solution_tools->handOverFindConfs(solution,action_index)){
                //        //compute the cost
                //        this->_solution_tools->computeCost(solution);//TODO check for LOCAL cost improvment
                //        found=true;
                //    }
                //}
                _logger->draw();
            }
            if(found){
                best_solution=*solution;
            }
        }
    }
    if(find_2d_optimal_before){
        no_check_3d=true;
    }
    while (!done){
        _logger->drawcells(solution->frontiers[action_index]->cellA(),solution->frontiers[action_index]->cellB(),false);
//#ifdef DRAW_OPTIMIZE_FRONTIER
//        if(_draw_optimize_frontier){
//            //_grid->setDrawAStar(true);
//            resetAgentsConf();
//            _grid->setDrawOptimizeFrontier(solution->frontiers[action_index]->cellA(),solution->frontiers[action_index]->cellB(),false);
//            g3d_draw_allwin_active();
//        }
//#endif
        chrono.reset();
        steps++;
        bool better_found=false;
        //the solution we study is the previous best solution
        *solution=best_solution;
        MultiHandOverSolution::Action *action=&solution->actions[action_index];
        FrontierPtr current_frontier=action->frontier();
        vector<FrontierPtr> candidates=current_frontier->findNeighbourFrontiersJumping(keep_distance,min_distance,50);
        std::random_shuffle(candidates.begin(),candidates.end()); // shuffle the order of the candidates
        //TIME
        times[0]+=chrono.getNanoSec();

        for (vector<FrontierPtr>::iterator it=candidates.begin() ; !better_found && it!=candidates.end() ; ++it){
            //for each candidate frontier we compute its impact on the solution
            //we first check if it's an ok frontier
            chrono.reset();
            //if(no_check_3d || checkFrontierWithObject(**it)){
            FrontierCells::FrontierCheckStatus check_level;
            if(no_check_3d){
                check_level=FrontierCells::FCS_NONE;
            }else{
                check_level=FrontierCells::FCS_GRAB_CONFS;
            }
            bool ok=_tool_set->frontierTester->test(*it,check_level);
            //start_pos[0]=(*it)->cellA()->getCenter()[0];
            //start_pos[1]=(*it)->cellA()->getCenter()[1];
            //end_pos[0]=(*it)->cellB()->getCenter()[0];
            //end_pos[1]=(*it)->cellB()->getCenter()[1];
            //_ho_checker->shoulders(start_pos,end_pos);
            //if(no_check_3d || this->_ho_checker->check()){ //TODO set the hochecker
            //    (*it)->object_position=_ho_checker->getQMiddle();
            //    (*it)->status_ok[FrontierCells::FCS_OBJECT]=true;
            //    (*it)->check_status=std::max(FrontierCells::FCS_OBJECT,(*it)->check_status);
            //    (*it)->grabA=_ho_checker->grab1();
            //    (*it)->grabB=_ho_checker->grab2();
            if(ok){
                //TIME
                times[1]+=chrono.getNanoSec();
                chrono.reset();

                //we set the solution new frontier to this candidate
                solution->frontiers[action_index]=*it;
                //reset data of the solution
                solution->cost=0;
                solution->handover_times[action_index]=0;
                //new_solution->actions.clear();
                this->_tool_set->solutionTools->computeNavigationTrajectories(solution,action_index);//TODO -> DONE? avoid computing ALL trajectories, only 3 need to be changed
                //TIME
                times[2]+=chrono.getNanoSec();
                chrono.reset();

                // //check if there is a grab conf ok for this frontier and add its cost
                // if(no_check_3d || this->_tool_set->solutionTools->handOverFindConfs(solution,action_index)){
                    //TIME
                    times[3]+=chrono.getNanoSec();
                    chrono.reset();

                    //compute the cost
                    this->_tool_set->solutionTools->computeCost(solution);//TODO check for LOCAL cost improvment
                    //TIME
                    times[4]+=chrono.getNanoSec();
                    chrono.reset();

                    if(solution->cost < best_cost){
                        //if its the best solution found
                        best_solution=(*solution);
                        best_cost=best_solution.cost;
                        better_found=true;
                        best_sol_is_full_solution=!no_check_3d;
                        cout<<"optimize solution: better found @"<<steps<<" cost:"<<best_cost<<endl;
                    // }
                    _logger->drawcells(solution->frontiers[action_index]->cellA(),solution->frontiers[action_index]->cellB(),true);
//#ifdef DRAW_OPTIMIZE_FRONTIER
//                    if(_draw_optimize_frontier){
//                        resetAgentsConf();
//                        //_grid->setDrawOptimizeFrontier(solution->frontiers[action_index]->cellA(),solution->frontiers[action_index]->cellB(),false);
//                        _grid->setDrawOptimizeFrontier(solution->frontiers[action_index]->cellA(),solution->frontiers[action_index]->cellB(),true);
//                        g3d_draw_allwin_active();
//                    }
//#endif
                }
            }
        }
        if(!better_found){
            //we fell in a minimum
            if(no_check_3d){
                //now we search with 3d computation
                no_check_3d=false;
                best_cost = initial_cost;
                cout<<"going to full 3d computation"<<endl;
            }else if(best_sol_is_full_solution){
                done=true;
            }else{
                //no full solution (with 3d) found, restart with full 3d from initial solution
                best_solution=initial_solution;
                best_cost=initial_cost;
                best_sol_is_full_solution=true;
                cout<<"restarting with initial solution"<<endl;
            }
        }
    }
    _logger->resetdisp();
//#ifdef DRAW_OPTIMIZE_FRONTIER
//    //_grid->setDrawAStar(false);
//    _grid->unsetDrawOptimizeFrontier();
//    _grid->unsetDrawGridForOptim();
//#endif
    for(unsigned int i=0;i<5;++i){
        cout<<static_cast<double>(times[i])*1e-9<< " , ";
    }
    cout<<endl;
    *solution=best_solution;
    //this->insert_solution_sorted(current_solution);
    //if(current_solution->cost<this->_best_solution->cost){this->_best_solution=current_solution;}
    solution->is_optimized=true;
    return solution;

}

MultiAgentCell* FrontierOptimizerClimber::findOptimalCellNullDistanceFrontier(MultiHandOverSolution *solution, unsigned int action_index){
    MultiHandOverSolution *aux_solution=solution->clone();
    MultiAgentCell* current;
    unsigned int a0,a1;
    MultiAgentCell* from0,*from1,*to1;//correspond to alternative origins for altDistances
    MultiHandOverSolution::Action & act=aux_solution->actions[action_index];
    bool found=false;
    double best_cost=numeric_limits<double>::infinity();
    MultiAgentCell *best_candidate;
    //TODO add for go back costs
    a0=act.agent(0)->index;
    a1=act.agent(1)->index;
    from0=act.from(0);
    from1=act.from(1);
    if(action_index+1<solution->actions.size()){
        to1=solution->actions[action_index+1].to(0);
    }else{
        to1=0;
    }
    current=act.to(0);
    if (!current->isReachable(a0) || !current->isReachable(a1)){
        if(_showText){
            cout<< "MultiHandOver::optimalCellNullDistanceFrontier() : this method is not applicable here as reachable spaces does not overlaps"<<endl;
        }
        return 0;//this method is not applicable here
    }
    aux_solution->frontiers[action_index]=FrontierPtr();//erase the frontier to avoid confusion
    act.path(0).clear();
    act.path(1).clear();
    //aux_solution->navigation_trajs[action_index*2].clear();
    //aux_solution->navigation_trajs[action_index*2+1].clear();
    act.setDist(0,0);
    act.setDist(1,0);
    //aux_solution->nav_dists[action_index*2]=0;
    //aux_solution->nav_dists[action_index*2+1]=0;
    aux_solution->handover_hri_costs[action_index]=0;
    aux_solution->handover_times[action_index]=0;
    best_candidate=current;
    while(!found){
        vector<MultiAgentCell*> neighb=current->getNeighbors(a0);
        for(unsigned int candidate=0;candidate<neighb.size();candidate++){
            MultiAgentCell* c=neighb[candidate];
            if(c->isAccessible(a0) && c->isAccessible(a1)){
                act.setDist(0,c->getAltDistance(a0,from0));
                act.setDist(1,c->getAltDistance(a1,from1));
                //assert(!"you really have to check this...");
                //TODO check this
                //action_index *2+2 is out of range somtimes
                //consider go to target cost
                if(to1){
                    aux_solution->nav_dists[action_index*2+2]=c->getAltDistance(a1,to1);
                }
                //double d0=c->getAltDistance(a0,from0);
                //double d1=c->getAltDistance(a1,from1);
                //double d1_1=c->getAltDistance(a1,to1);
                //double cost= _f_dist*(d0+d1+d1_1)
                //        + _f_time * (d0/getAgentSpeed(a0)+d1_1/getAgentSpeed(a1))
                //        + _f_agent_cost * ( d0/getAgentSpeed(a0)*getAgentUseCost(a0) + (d1+d1_1)/getAgentSpeed(a1)*getAgentUseCost(a1));
                double cost=this->_tool_set->solutionTools->computeCost(aux_solution);
                if(cost<best_cost){
                    best_candidate=c;
                    best_cost=cost;
                }
            }
        }
        if(best_candidate==current){
            found=true;
        }else{
            current=best_candidate;//for next loop
        }
        _logger->drawcells(best_candidate,best_candidate,found);
//#ifdef DRAW_OPTIMIZE_FRONTIER
//        if(_draw_optimize_frontier){
//            resetAgentsConf();
//            //_grid->setDrawOptimizeFrontier(solution->frontiers[action_index]->cellA(),solution->frontiers[action_index]->cellB(),false);
//            _grid->setDrawOptimizeFrontier(best_candidate,best_candidate,found);
//            g3d_draw_allwin_active();
//        }
//#endif
    }
    delete aux_solution;
    return best_candidate;
}

// ////////////////////////////////////////////////////
// ///////////  HANDOVER WINDOW CHECKER   /////////////
// ////////////////////////////////////////////////////

const std::string HandOverDirectChecker::_name="Object Direct check";
const std::string HandOverWindowCheckerRRT::_name="Object RRT check";

// //// DIRECT

HandOverDirectChecker::HandOverDirectChecker(Robot *object,Robot* r1,Robot* r2):
    AbstracHandOverObjectCheck(object,r1,r2)
{
}

bool HandOverDirectChecker::check(){
    bool ret(0);
    p3d_col_deactivate_robot_robot(robots().first->getRobotStruct(),object()->getRobotStruct());
    p3d_col_deactivate_robot_robot(robots().second->getRobotStruct(),object()->getRobotStruct());

    Configuration q_save = *object()->getCurrentPos();

    Eigen::Vector3d direction = shoulder2() - shoulder1();

    this->setOrientationObjectForHO(direction);

    shared_ptr<Configuration> q_temp = object()->getCurrentPos();
    q_middle=confPtr_t(new Configuration(object()));

    double length=direction.norm();
    //double step_size=0.1; //TODO step size for object moves to check frontiers
    double step_size = _object_bb_vects[_object_directions_sorted_for_ho[0]].norm();
    double dist_thresh=0.03;//TODO distance tolerance for object moves to check frontiers
    double middle_reach=(length-reach1()+reach2())/2;
    unsigned int iteration(0);
    bool collision=false;
    direction.normalize();

    q_temp->at(0+6)=shoulder1()[0];
    q_temp->at(1+6)=shoulder1()[1];
    q_temp->at(2+6)=shoulder1()[2];

    while(!collision && iteration*step_size<length+step_size){
        q_temp->at(0+6)+=direction[0]*step_size;
        q_temp->at(1+6)+=direction[1]*step_size;
        q_temp->at(2+6)+=direction[2]*step_size;
        object()->setAndUpdate(*q_temp);
        //if(_draw_object_frontiers_checks){
            //g3d_draw_allwin_active();
        //}
        p3d_vector3 v1,v2;//only for arguments, unused
        //TODO DONE? this is bad as it does not check robots (objects). (disable colision w/ involved robots)
        double dist=p3d_col_robot_environment_distance(object()->getRobotStruct(),v1,v2);
        if(dist < dist_thresh || object()->isInCollisionWithOthersAndEnv()){
            collision=true;
        }
        /*unsigned int i_rob(0);
        while(!collision && i_rob<XYZ_ENV->nr){
            //check the distances with other robots/objects
            if(XYZ_ENV->robot[i_rob] != frontier.agentA->robot()->getRobotStruct() && XYZ_ENV->robot[i_rob] != frontier.agentB->robot()->getRobotStruct()){
                dist=p3d_col_robot_robot_distance(_object->getRobotStruct(),XYZ_ENV->robot[i_rob],v1,v2);
                if(dist<dist_thresh){
                    collision=true;
                }
            }
            ++i_rob;
        }*/
        if(iteration*step_size<=middle_reach && (iteration+1)*step_size>middle_reach){
            *q_middle=*q_temp;
        }
        iteration++;
    }
    //cout<<endl;
    if(collision){
    }else{
        Eigen::Vector3d mid(q_middle->at(6+0),q_middle->at(6+1),q_middle->at(6+2));
        Eigen::Vector3d grabA=mid-direction*step_size/2;
        Eigen::Vector3d grabB=mid+direction*step_size/2;
        this->grabs(grabA,grabB);
    }
    object()->setAndUpdate(q_save);
    ret= !collision;
    p3d_col_activate_robot_robot(robots().first->getRobotStruct(),object()->getRobotStruct());
    p3d_col_activate_robot_robot(robots().second->getRobotStruct(),object()->getRobotStruct());
    return ret;
}

Eigen::Vector3d point_to_Vector3d(p3d_point &p){
    Eigen::Vector3d v;
    v[0]=p.x;
    v[1]=p.y;
    v[2]=p.z;
    return v;
}

Configuration HandOverDirectChecker::confForBBAxisAlign(Eigen::Vector3d const & main_dir, Eigen::Vector3d const & second_dir){
    using namespace Eigen;
    Vector3d dir1=main_dir,dir2=second_dir,dir3=-main_dir.cross(second_dir);
    if(this->_object_bb_vects.size()==0){
        p3d_oriented_BB * bb= p3d_get_oriented_BB_obj(*object()->getRobotStruct()->o); // the bounding box of the object
        //the 3 main vectors describing the bounding box
        _object_bb_vects.resize(3);
        _object_bb_vects[0] = point_to_Vector3d(bb->BottomFrontRight) - point_to_Vector3d(bb->BottomBackRight);//front x
        _object_bb_vects[1] = point_to_Vector3d(bb->BottomFrontLeft) - point_to_Vector3d(bb->BottomFrontRight);//left y
        _object_bb_vects[2] = point_to_Vector3d(bb->UpFrontRight) - point_to_Vector3d(bb->BottomFrontRight);//up z
        vector<double> areas(3);
        areas[0]=_object_bb_vects[2].cross(_object_bb_vects[1]).norm();
        areas[1]=_object_bb_vects[2].cross(_object_bb_vects[0]).norm();
        areas[2]=_object_bb_vects[1].cross(_object_bb_vects[0]).norm();
        _object_directions_sorted_for_ho.clear();
        _object_directions_sorted_for_ho.reserve(3);
        for(unsigned int i=0;i<3;i++){
            for(unsigned int j=0;j<3;j++){
                if(i!=j && areas[i]<=areas[j]){
                    for(unsigned int k=0;k<3;k++){
                        if(j!=k && areas[j]<=areas[k]){
                            _object_directions_sorted_for_ho.push_back(i);
                            _object_directions_sorted_for_ho.push_back(j);
                            _object_directions_sorted_for_ho.push_back(k);
                            break;
                        }
                    }
                    if(_object_directions_sorted_for_ho.size()!=0){break;}
                }
            }
            if(_object_directions_sorted_for_ho.size()!=0){break;}
        }
    }

    p3d_matrix4 xform={{1, 0, 0, 0}, {0, 1, 0, 0}, {0, 0, 1, 0}, {0, 0, 0, 1}};
    dir1.normalize();
    dir2.normalize();
    dir3.normalize();
    xform[0][_object_directions_sorted_for_ho[0]]=dir1.x();
    xform[1][_object_directions_sorted_for_ho[0]]=dir1.y();
    xform[2][_object_directions_sorted_for_ho[0]]=dir1.z();

    xform[0][_object_directions_sorted_for_ho[1]]=dir2.x();
    xform[1][_object_directions_sorted_for_ho[1]]=dir2.y();
    xform[2][_object_directions_sorted_for_ho[1]]=dir2.z();

    xform[0][_object_directions_sorted_for_ho[2]]=dir3.x();
    xform[1][_object_directions_sorted_for_ho[2]]=dir3.y();
    xform[2][_object_directions_sorted_for_ho[2]]=dir3.z();

    Configuration conf = *object()->getCurrentPos();
    double foo[3];
    p3d_mat4ExtractPosReverseOrder(xform,&foo[0],&foo[1],&foo[2],&conf[9],&conf[10],&conf[11]);
    return conf;
}
void HandOverDirectChecker::setOrientationObjectForHO(Eigen::Vector3d & dir1){
    using namespace Eigen;
    vector<Vector3d> dirs(3);
    dirs[0]=Vector3d(1,0,0);
    dirs[1]=Vector3d(0,1,0);
    dirs[2]=Vector3d(0,0,1);

    //Configuration q=this->confForBBAxisAlign(_object,dirs[i],dirs[j]);
    Vector3d dir2;
    dir2=dirs[2]/*z*/.cross(dir1);
    dir2=dir2.cross(dir1);
    if(dir2.dot(dirs[2])<0){
        dir2=-dir2;
    }
    //<=> dir2 = z x d1 x d1
    Configuration q=this->confForBBAxisAlign(dir1,dir2);

    object()->setAndUpdate(q);
}

// //// RRT

HandOverWindowCheckerRRT::HandOverWindowCheckerRRT(Robot *obj,Robot*r1,Robot* r2,bool dummy_smooth):
    AbstracHandOverObjectCheck(obj,r1,r2)
{
    setUseDummySmooth(dummy_smooth);
    saveObjectLimits();
}
HandOverWindowCheckerRRT::HandOverWindowCheckerRRT(Robot* obj,Robot* r1,Robot* r2, bool dummy_smooth,
                                             Eigen::Vector3d shoulder1, Eigen::Vector3d shoulder2,
                                             double reach1, double reach2) :
    AbstracHandOverObjectCheck(obj,r1,r2)
{
    setUseDummySmooth(dummy_smooth);
    this->shoulders(shoulder1,shoulder2);
    this->reach(reach1,reach2);
    //this->object(obj);
    saveObjectLimits();
}

bool HandOverWindowCheckerRRT::check(){
    constructBB();
    setObjectLimits();
    p3d_col_deactivate_robot_robot(robots().first->getRobotStruct(),object()->getRobotStruct());
    p3d_col_deactivate_robot_robot(robots().second->getRobotStruct(),object()->getRobotStruct());
    bool r=computeRRT(this->timeLimit());
    p3d_col_activate_robot_robot(robots().first->getRobotStruct(),object()->getRobotStruct());
    p3d_col_activate_robot_robot(robots().second->getRobotStruct(),object()->getRobotStruct());
    resetObjectLimits();

    return r;
}


bool HandOverWindowCheckerRRT::computeRRT(double time_limit_ms){
    bool is_ok=true;
    Robot* obj=this->object();
    Configuration q_init(*obj->getCurrentPos()),q_goal(*obj->getCurrentPos());
    //StdOutMuter::mute();
    //std::streambuf* cout_sbuf = std::cout.rdbuf(); // save original sbuf
    //std::streambuf* cerr_sbuf = std::cerr.rdbuf(); // save original sbuf
    //std::ofstream   fout("/dev/null");
    //std::cout.rdbuf(fout.rdbuf()); // redirect 'cout' to a 'fout'
    //std::cerr.rdbuf(fout.rdbuf()); // redirect 'cout' to a 'fout'
    ManipulationPlanner *mp=new ManipulationPlanner(obj->getRobotStruct());
    //StdOutMuter::unmute();
    //std::cout.rdbuf(cout_sbuf); // restore the original stream buffer
    //std::cerr.rdbuf(cerr_sbuf); // restore the original stream buffer
    is_ok=(mp!=0);
    if(is_ok){
        global_Project->getActiveScene()->setActiveRobot(obj->getName());
        mp->setPlanningMethod(p3d_planner_function);
        if(this->_use_dummy_smooth){
            mp->setSmoothingMethod(p3d_smoothing_function_dummy);
        }else{
            mp->setSmoothingMethod(p3d_smoothing_function);
        }
        mp->setCleanningRoadmaps(true);
        mp->setPlanningTime(time_limit_ms/1000);
        mp->setOptimizeTime(1./1000);
        setPlanningTimeLimit(time_limit_ms/1000);//takes seconds
        PlanEnv->setBool(PlanParam::planWithTimeLimit,true);
        PlanEnv->setDouble(PlanParam::timeLimitPlanning,time_limit_ms/1000);

        q_init.at(0+6)=shoulder1()[0];
        q_init.at(1+6)=shoulder1()[1];
        q_init.at(2+6)=shoulder1()[2];

        q_goal.at(0+6)=shoulder2()[0];
        q_goal.at(1+6)=shoulder2()[1];
        q_goal.at(2+6)=shoulder2()[2];

        MANIPULATION_TASK_MESSAGE status;

        //std::streambuf* cout_sbuf = std::cout.rdbuf(); // save original sbuf
        //std::streambuf* cerr_sbuf = std::cerr.rdbuf(); // save original sbuf
        //std::ofstream   fout("/dev/null");
        //std::cout.rdbuf(fout.rdbuf()); // redirect 'cout' to a 'fout'
        //std::cerr.rdbuf(fout.rdbuf()); // redirect 'cout' to a 'fout'
        p3d_traj *traj = mp->computeTrajBetweenTwoConfigs(q_init.getConfigStruct(),q_goal.getConfigStruct(),&status);
        //std::cout.rdbuf(cout_sbuf); // restore the original stream buffer
        //std::cerr.rdbuf(cerr_sbuf); // restore the original stream buffer

        is_ok=(status==MANIPULATION_TASK_OK);

        double l;
        if(is_ok){
            l=traj->courbePt->length_lp;
            configPt q_middle_ptr=traj->courbePt->config_at_distance(obj->getRobotStruct(),traj->courbePt,l/2);
            q_middle = shared_ptr<Configuration>(new Configuration(obj));
            q_middle->setConfiguration(q_middle_ptr);
        }
    }
    if(mp){
        delete mp;
    }
    return is_ok;
}


void HandOverWindowCheckerRRT::setObjectLimits(){
    Robot* obj=this->object();
    obj->getRobotStruct()->joints[1]->dof_data[0].vmin=bb.first.x();
    obj->getRobotStruct()->joints[1]->dof_data[1].vmin=bb.first.y();
    obj->getRobotStruct()->joints[1]->dof_data[2].vmin=bb.first.z();

    obj->getRobotStruct()->joints[1]->dof_data[0].vmax=bb.second.x();
    obj->getRobotStruct()->joints[1]->dof_data[1].vmax=bb.second.y();
    obj->getRobotStruct()->joints[1]->dof_data[2].vmax=bb.second.z();
}

void HandOverWindowCheckerRRT::saveObjectLimits(){
    Robot* obj=this->object();
    object_limits_bkp.assign(6,0);
    unsigned int i=0;
    object_limits_bkp[i++]=obj->getRobotStruct()->joints[1]->dof_data[0].vmin;
    object_limits_bkp[i++]=obj->getRobotStruct()->joints[1]->dof_data[1].vmin;
    object_limits_bkp[i++]=obj->getRobotStruct()->joints[1]->dof_data[2].vmin;

    object_limits_bkp[i++]=obj->getRobotStruct()->joints[1]->dof_data[0].vmax;
    object_limits_bkp[i++]=obj->getRobotStruct()->joints[1]->dof_data[1].vmax;
    object_limits_bkp[i++]=obj->getRobotStruct()->joints[1]->dof_data[2].vmax;
}

void HandOverWindowCheckerRRT::resetObjectLimits(){
    Robot* obj=this->object();
    unsigned int i=0;
    obj->getRobotStruct()->joints[1]->dof_data[0].vmin=object_limits_bkp[i++];
    obj->getRobotStruct()->joints[1]->dof_data[1].vmin=object_limits_bkp[i++];
    obj->getRobotStruct()->joints[1]->dof_data[2].vmin=object_limits_bkp[i++];

    obj->getRobotStruct()->joints[1]->dof_data[0].vmax=object_limits_bkp[i++];
    obj->getRobotStruct()->joints[1]->dof_data[1].vmax=object_limits_bkp[i++];
    obj->getRobotStruct()->joints[1]->dof_data[2].vmax=object_limits_bkp[i++];
}

void HandOverWindowCheckerRRT::constructBB(){
    double inf=numeric_limits<double>::infinity();
    Vector3d B(-inf,-inf,-inf),b(inf,inf,inf);
    vector<Eigen::Vector2d> square(4,Eigen::Vector2d(0,0));
    Vector2d diag,diag2;
    square[0]=v3dto2d(shoulder1());
    square[2]=v3dto2d(shoulder2());
    diag=square[2]-square[0];
    diag2.x()=diag.y();
    diag2.y()=-diag.x();
    square[1]=square[0]+diag/2+diag2/2;
    square[3]=square[0]+diag/2-diag2/2;
    //TODO remove this check:
    assert(diag.dot(diag2)< 0.00001);
    for (unsigned int i=0;i<4;i++){
        b.x()=min(b.x(),square[i].x());
        B.x()=max(B.x(),square[i].x());
    }
    for (unsigned int i=0;i<4;i++){
        b.y()=min(b.y(),square[i].y());
        B.y()=max(B.y(),square[i].y());
    }
    b.x()-=0.01;
    b.y()-=0.01;
    B.x()+=0.01;
    B.y()+=0.01;
    b.z()=max(shoulder1().z()-reach1(),shoulder2().z()-reach2());
    B.z()=min(shoulder1().z()+reach1(),shoulder2().z()+reach2());
    this->bb=make_pair(b,B);
}

Eigen::Vector2d HandOverWindowCheckerRRT::v3dto2d(Vector3d v3){
    return Vector2d(v3.x(),v3.y());
}

}
}
