#include "costMapPlot.h"
#include "planner/cost_space.hpp"
#include "API/project.hpp"

INIT_MOVE3D_STATIC_LOGGER(CostMapPlot,"planners.utils.costmapplot");

CostMapPlot::CostMapPlot(const std::string &name):
    _name(name),_defaultSampleNbPerJnt(20)
{
    _name_traj=name;
    _name_traj.append(".traj");
    _socketPlot =new SocketPlot(_name);
    _socketPlotTraj =new SocketPlot(_name_traj);
    _socketPlot->sendEmpty();
    _socketPlotTraj->sendEmpty();
}

CostMapPlot::~CostMapPlot()
{
    if (_socketPlot){
        delete _socketPlot;
        _socketPlot=NULL;
    }
}

void CostMapPlot::start()
{
    std::vector<double> dofs;
    double cost(0);
    _reached_end=false;
    int sample_nb=1;
    for (unsigned int i=0;i<_dofs.size();i++){
        sample_nb*=(_max[i]-_min[i])/_steps[i];
    }
    std::cout<<"sample nb="<<sample_nb<<std::endl;


    saveRobotConfs();

    cost=firstSample(dofs);
    dofs.push_back(cost);
    _socketPlot->addPoint(dofs);
    while(!_reached_end){
        dofs.clear();
        cost=nextSample(dofs);
        if(!_reached_end){
            dofs.push_back(cost);
            //M3D_TRACE("adding point:"<<_socketPlot->vectorToString(dofs));
            _socketPlot->addPoint(dofs);
        }
    }
    _socketPlot->send();

    restoreRobotConfs();
}

void CostMapPlot::plotCurrentTraj()
{
    this->saveRobotConfs();
    API::Trajectory traj = global_Project->getActiveScene()->getActiveRobot()->getCurrentTraj();
    std::vector<LocalPath*> courbe=traj.getCourbe();
    Robot *r=global_Project->getActiveScene()->getActiveRobot();
    for (unsigned int i=0;i<courbe.size();i++){
        LocalPath* lp=courbe[i];
        Configuration q=*lp->getBegin();
        r->setAndUpdate(q);
        double cost=q.cost();
        std::vector<double> point;
        for (std::vector<std::pair<Joint*,int> >::iterator it=_dofs.begin();it!=_dofs.end();++it){
            if (it->first->getJointStruct()->rob == r->getRobotStruct()){
                point.push_back(it->first->getJointDof(it->second));
            }
        }
        point.push_back(cost);
        _socketPlotTraj->addPoint(point);
    }

    _socketPlotTraj->send();
    this->restoreRobotConfs();
}


void CostMapPlot::addDof(Robot *r,Joint *j, int dof_index, double min, double max, double step)
{
    double jmin,jmax;
    _robots_to_update.insert(r);
    _dofs.push_back(std::make_pair(j,dof_index));
    j->getDofBounds(dof_index,jmin,jmax);
    _min.push_back(std::max(jmin,min));
    _max.push_back(std::min(jmax,max));
    assert(_max.back() >= _min.back());
    if(step==0){
        double s=_max.back() - _min.back();
        s=s/defaultSampleNbPerJnt();
        _steps.push_back(s);
    }else{
        _steps.push_back(step);
    }
    assert(_steps.back()!=0);
    std::cout << "addDof "<<(_max.back() -_min.back())/_steps.back()<<" samples"<<std::endl;
}

double CostMapPlot::firstSample(std::vector<double> &dof_values)
{
    double cost(std::numeric_limits<double>::infinity());
    dof_values.clear();
    _curValues.clear();
    _reached_end=false;
    for(unsigned i=0;i<_dofs.size();i++){
        double val(0);
        val=(_steps[i]>=0 ? _min[i] : _max[i]);
        //_dofs[i].first->setJointDof(_dofs[i].second,val);
        this->setAndUpdateJointDof(_dofs[i].first,_dofs[i].second,val);
        dof_values.push_back(val);
    }
    _curValues=dof_values;
    updateRobots();
    confPtr_t q=global_Project->getActiveScene()->getActiveRobot()->getCurrentPos();
    q->setCostAsNotTested();
    cost=q->cost();
    return cost;

}

double CostMapPlot::nextSample(std::vector<double> &dof_values)
{
    double cost(std::numeric_limits<double>::infinity());
    bool movenext=!_reached_end;
    unsigned int i=0;
    dof_values.clear();
    while (movenext && i<_dofs.size()){
        double val=moveJointNextSample(i,movenext);
        dof_values.push_back(val);
        _curValues[i]=val;
        i++;
    }
    if(i==_dofs.size() && movenext){
        _reached_end=true;
    }else{
        //append all DoFs values to dof_values
        for(unsigned ii=i;ii<_dofs.size();ii++){
            dof_values.push_back(this->getDofValue(ii));
        }
    }
    if(!_reached_end){
        updateRobots();
        //g3d_draw_allwin_active();
        confPtr_t q=global_Project->getActiveScene()->getActiveRobot()->getCurrentPos();
        q->setCostAsNotTested();
        cost=q->cost();
        //M3D_TRACE("cost:"<<cost);
        //assert(_curValues.size() == _dofs.size());
        //for(unsigned int i=0;i<_curValues.size();i++){
            //assert( std::fabs(_curValues[i]-this->getDofValue(i)) < 0.00001);
            //M3D_TRACE(i<<": "<<_curValues[i]);
        //}
    }
    return cost;
}

void CostMapPlot::updateRobots()
{
    //for (std::set<Robot*>::iterator i=_robots_to_update.begin();i!=_robots_to_update.end();i++){
        //M3D_TRACE("update robot "<<(*i)->getName());
        //(*i)->setAndUpdate(*(*i)->getCurrentPos());
    //}
    //for(unsigned int i=0;i<_curValues.size();i++){
    //    assert( std::fabs(_curValues[i]-this->getDofValue(i)) < 0.00001);
    //}
}

void CostMapPlot::saveRobotConfs()
{
    _savedConf.clear();
    for (std::set<Robot*>::iterator i=_robots_to_update.begin();i!=_robots_to_update.end();i++){
        _savedConf.push_back(*(*i)->getCurrentPos());
    }
}

void CostMapPlot::restoreRobotConfs()
{
    for (std::vector<Configuration>::iterator i=_savedConf.begin();i!=_savedConf.end();i++){
        i->getRobot()->setAndUpdate(*i);
    }
}
unsigned int CostMapPlot::defaultSampleNbPerJnt() const
{
    return _defaultSampleNbPerJnt;
}

void CostMapPlot::setDefaultSampleNbPerJnt(unsigned int defaultSampleNbPerJnt)
{
    _defaultSampleNbPerJnt = defaultSampleNbPerJnt;
}

void CostMapPlot::setAndUpdateJointDof(Joint *j, int dof, double val)
{
    Robot *r=j->getRobot();
    Configuration q=*r->getCurrentPos();
    q[j->getIndexOfFirstDof() + dof] = val;
    r->setAndUpdate(q);
    double qval=r->getCurrentPos()->at(j->getIndexOfFirstDof()+dof);
    if (std::fabs(qval - val) > 0.00001){
        M3D_WARN("Couldn't set "<<r->getName()<<" joint "<<j->getName()<<" dof "<< dof);
    }
}

void CostMapPlot::test1()
{
    Robot *r=global_Project->getActiveScene()->getActiveRobot();
    Configuration q = *r->getCurrentPos();
    int j0=r->getJoint(2)->getIndexOfFirstDof();
    int dof=0;
    q.at(j0+dof)=0.0;
    j0=r->getJoint(0)->getIndexOfFirstDof();
    dof=0;
    q.at(j0+dof)=0.0;

    r->setAndUpdate(q);


}

void CostMapPlot::test2()
{
    Robot *r=global_Project->getActiveScene()->getActiveRobot();
    r->setAndUpdate(*r->getCurrentPos());
}


double CostMapPlot::moveJointNextSample(unsigned int i,bool &reached_extreme)
{
    Joint* jnt = _dofs[i].first;
    int dof=_dofs[i].second;
    reached_extreme=false;
    double step=_steps[i];
    double max=_max[i];
    double min=_min[i];
    //double val=jnt->getJointDof(dof);
    double val=_curValues[i];
    double new_val=val+step;
    //M3D_TRACE("val"<<val);
    //M3D_TRACE("\tnew_val"<<new_val);
    assert(val<=max && val>=min);
    assert(max>min);
    assert(step!=0);
    if (new_val>max){
        new_val=min;
        reached_extreme=true;
        //M3D_TRACE("\treached max");
    }
    if (new_val<min){
        new_val=max;
        reached_extreme=true;
        //M3D_TRACE("\treached min");
    }
    //M3D_TRACE("\tapplied val"<<new_val);
    //jnt->setJointDof(i,new_val);
    this->setAndUpdateJointDof(jnt,dof,new_val);
    //assert(jnt->getJointDof(i)==new_val);
    assert(val!=new_val);
    return new_val;
}

double CostMapPlot::getDofValue(unsigned int i)
{
    return _dofs[i].first->getJointDof(_dofs[i].second);
}

