#include "SimpleGraph.h"
#include <iostream>
#include <queue>
#include <deque>
#include <limits>
#include <algorithm>

namespace mho{
using namespace std;

SimpleGraph::SimpleGraph(bool directed)
{
    _is_directed=directed;
    _count=0;
}

SimpleGraph::~SimpleGraph(){
    this->clear();
}

SimpleGraph * SimpleGraph::clone(){
    map<NodeInterface*,NodeInterface*> nodes_corresp;
    SimpleGraph * g=new SimpleGraph(this->isDirected());
    for(unsigned int i=0;i<this->size();++i){
        nodes_corresp[this->at(i)] = g->addNode(this->at(i));//does the cloning of the node
    }
    for (unsigned int i=0;i<this->edgesNumber();++i){
        Edge* e=this->_edges[i];
        g->connectNodes(nodes_corresp[e->first()],nodes_corresp[e->second()],e->getCost());
    }
    return g;
}

bool SimpleGraph::isDirected(){return _is_directed;}

NodeInterface * SimpleGraph::addNode(NodeInterface *n){
    return this->addNode(dynamic_cast<Node*>(n));
}

NodeInterface * SimpleGraph::addNode(Node *n){
    Node *cp=dynamic_cast<Node*>(n->clone());
    cp->sons().clear();
    cp->setGraph(this);
    //*cp=*n;
    this->_nodes.push_back(cp);
    _count++;
    return this->_nodes.back();
}

void SimpleGraph::connectNodes(NodeInterface *n1, NodeInterface *n2, cost_t c){
    this->_edges.push_back(new Edge(n1,n2,c));
    n1->connect(_edges.back());
    if (!this->isDirected()){
        n2->connect(_edges.back());
    }
}

//void SimpleGraph::disconnectNodes(NodeInterface *n1, NodeInterface *n2,cost_t c){
//    vector<EdgeInterface*> sons= dynamic_cast<Node*>(n1)->sons();
//    for(unsigned int i=0;i<sons.size();++i){
//        if(sons[i]->oppositeSideOf(n1)==n2){
//            for(unsigned int j=0;j<_edges.size();++j){
//                if(_edges[j]==sons[i]){
//                    delete _edges[j];
//                    _edges[j]=_edges.back();
//                    _edges.pop_back();
//                }
//            }
//        }
//    }
//}

void SimpleGraph::updateCost(NodeInterface* n1,NodeInterface* n2,cost_t newcost,cost_t oldcost, bool onlySpecifiedCost){
    vector<EdgeInterface*> sons= this->sons(n1);
    for(unsigned int i=0;i<sons.size();++i){
        if(sons[i]->oppositeSideOf(n1)==n2 && (oldcost==sons[i]->getCost() || !onlySpecifiedCost)){
            sons[i]->setCost(newcost);
            break;
        }
    }
}
void SimpleGraph::updateCost(NodeInterface *n1, NodeInterface *n2, cost_t newcost){
    this->updateCost(n1,n2,newcost,0,false);
}

void SimpleGraph::updateCost(EdgeInterface *edge,cost_t cost){
    edge->setCost(cost);
}

std::vector<EdgeInterface*> SimpleGraph::sons(NodeInterface *i){
    std::vector<EdgeInterface*> s(0);
    if(i && i->getGraph()==this){
        s=dynamic_cast<Node*>(i)->sons();
    }
    return s;
}

std::vector<NodeInterface*> SimpleGraph::sonsNodes(NodeInterface *i){
    std::vector<NodeInterface*> nodes;
    std::vector<EdgeInterface*> sons=this->sons(i);
    //for(std::vector<EdgeInterface*>::iterator it=sons.begin();it!=sons.end();++it){
    //nodes.push_back((*it)->oppositeSideOf(i));
    for(unsigned int k = 0;k<sons.size();++k){
        nodes.push_back(sons[k]->oppositeSideOf(i));
    }
    return nodes;
}


std::vector<NodeInterface*> SimpleGraph::getNodes(void){
    std::vector<NodeInterface*> r;
    r.reserve(this->_nodes.size());
    for(std::vector<Node*>::iterator it=this->_nodes.begin();it!=this->_nodes.end();++it){
        r.push_back(*it);
    }
    return r;
}


unsigned int SimpleGraph::edgesNumber(void){return this->_edges.size();}

NodeInterface* SimpleGraph::at(unsigned int i){
    return this->_nodes.at(i);
}

EdgeInterface* SimpleGraph::edgeAt(unsigned int i){
    return this->_edges.at(i);
}

void SimpleGraph::clear(){
    for (vector<Node*>::iterator it=this->_nodes.begin();it!=this->_nodes.end();++it){
        delete *it;
    }
    //this->vector::clear();
    for (vector<Edge*>::iterator it=this->_edges.begin();it!=this->_edges.end();++it){
        delete *it;
    }
    this->_edges.clear();
    this->_nodes.clear();
    this->_count=0;
}

unsigned int SimpleGraph::size(void){return /* 0*/this->_nodes.size();}

#ifdef _DEBUG_SIMPLEGRAPH_H
ostream& operator<<( ostream& os, SimpleGraph& g){
    os<<"Graphe"<<endl<<"Nodes ("<<g.size() << ") :"<<endl;
    for(vector<Node*>::iterator i=g._nodes.begin();i!=g._nodes.end();++i){
        os<<(*i)->getId()<<":"<<endl;
        std::vector<EdgeInterface*> sons=g.sons(*i);
        cout<<"#sons:"<<sons.size()<<endl;
        for(std::vector<EdgeInterface*>::iterator ei=sons.begin();ei!=sons.end();++ei){
            os << "\t" << "->"<<(*ei)->oppositeSideOf(*i)<<endl;
            os << "\t" << "->"<<(*ei)->oppositeSideOf(*i)->getId()<<"("<<(*ei)->getCost()<<')' <<endl;
        }
    }
    os<<"---"<<endl;
    return os;
}

#endif

}
